package datavisualizer;

import javawrapper.ModelException;

import javax.swing.JPanel;
import java.awt.Color;

import javax.swing.*;
import java.awt.*;

import org.jfree.chart.*;
import org.jfree.chart.axis.CategoryAxis;
import org.jfree.chart.axis.NumberAxis;
import org.jfree.chart.axis.ValueAxis;
import org.jfree.chart.axis.LogarithmicAxis;
import org.jfree.chart.plot.*;
import org.jfree.chart.renderer.*;
import org.jfree.data.DefaultCategoryDataset;
import org.jfree.ui.TextAnchor;
import org.jfree.data.XYZDataset;
import org.jfree.chart.plot.XYPlot;
import org.jfree.chart.renderer.XYBarRenderer;
import org.jfree.chart.axis.NumberTickUnit;

/**
 * Objects of this class can create graphs of various kinds, given data.  The
 * functions you are most likely to call are the Draw[X] and Update[X].
 * <p>Copyright: Copyright (c) Charles D. Canham 2003</p>
 * <p>Company: Institute of Ecosystem Studies</p>
 * @author Lora E. Murphy
 * @version 1.0
 *
 * <br>Edit history:
 * <br>------------------
 * <br>April 28, 2004: Submitted in beta version (LEM)
 * <br>May 11, 2004: Made all the functions static (LEM)
 * <br>July 19, 2004:  Preserved user resizing of charts (LEM)
 */
public class DataGrapher {

  /**
   * Updates a line chart in a window with a fresh dataset.  The old content
   * pane is replaced with a new one, so do not count on existing content pane
   * references to the window working.
   * @param oDataset The new dataset.
   * @param jFrame The window in which to update the chart.
   * @param oLegend The legend for this chart.
   * @param iNumSpecies The number of total species there are.
   * @throws ModelException Passing through underlying exceptions.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public static void UpdateLineChart(DefaultCategoryDataset oDataset,
                                     JInternalFrame jFrame, Legend oLegend,
                                     int iNumSpecies) throws ModelException {

    //Get the X and Y axis labels from the existing chart
    ChartPanel jPanel = (ChartPanel) jFrame.getContentPane();
    JFreeChart oChart = jPanel.getChart();
    CategoryPlot oPlot = oChart.getCategoryPlot();
    CategoryAxis sXAxis = oPlot.getDomainAxis();
    ValueAxis sYAxis = oPlot.getRangeAxis();
    String sXAxisLabel = sXAxis.getLabel();
    String sYAxisLabel = sYAxis.getLabel();

    //Make a new chart
    jPanel = MakeLineChart(oDataset, sXAxisLabel, sYAxisLabel,
                           oLegend, iNumSpecies, jPanel.getSize());

    //jPanel.setPreferredSize(new java.awt.Dimension(500, 270));

    //Replace the old chart with the new one in our window
    jFrame.setContentPane(jPanel);
  }

  /**
   * Updates a histogram with a fresh dataset.
   * @param oDataset The new dataset.
   * @param jFrame The window in which to update the chart.
   * @param oLegend The legend for this chart.
   * @param oRequest The HistogramDataRequest object for this chart.
   * @throws ModelException Passing through underlying exceptions.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public static void UpdateHistogram(ModelHistogramDataset oDataset,
                                     JInternalFrame jFrame,
                                     Legend oLegend,
                                     HistogramDataRequest oRequest
                                     ) throws ModelException {

    ChartPanel jPanel = null;
    java.awt.Dimension jPreferredSize = new java.awt.Dimension(750, 270);
    JPanel jContentPanel = (JPanel) jFrame.getContentPane();

    //Remove the existing chart
    for (int i = 0; i < jContentPanel.getComponentCount(); i++) {
      if (jContentPanel.getComponent(i) instanceof ChartPanel) {
        jPanel = (ChartPanel) jContentPanel.getComponent(i);
        jPreferredSize.setSize(jPanel.getSize());
        jContentPanel.remove(i);
      }
    }

    //Get the X and Y axis labels from the existing chart
    JFreeChart oChart = jPanel.getChart();
    XYPlot oPlot = oChart.getXYPlot();
    ValueAxis sXAxis = oPlot.getDomainAxis();
    ValueAxis sYAxis = oPlot.getRangeAxis();
    String sXAxisLabel = sXAxis.getLabel();
    String sYAxisLabel = sYAxis.getLabel();

    jPanel = MakeHistogram(oDataset, sXAxisLabel, sYAxisLabel, oLegend,
                           oRequest, jPreferredSize);

    //Replace the old chart with the new one in our window
    jContentPanel.add(jPanel);
    jFrame.setContentPane(jContentPanel);
  }

  /**
   * Updates a histogram with a fresh dataset.
   * @param oDataset The new dataset.
   * @param jFrame The window in which to update the chart.
   * @param oLegend The legend for this chart.
   * @param oRequest The GridHistogramDataRequest object for this chart.
   * @throws ModelException Passing through underlying exceptions.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public static void UpdateHistogram(ModelHistogramDataset oDataset,
                                     JInternalFrame jFrame,
                                     Legend oLegend,
                                     GridHistogramDataRequest oRequest
                                     ) throws ModelException {

    ChartPanel jPanel = null;
    java.awt.Dimension jPreferredSize = new java.awt.Dimension(300, 270);
    JPanel jContentPanel = (JPanel) jFrame.getContentPane();

    //Remove the existing chart
    for (int i = 0; i < jContentPanel.getComponentCount(); i++) {
      if (jContentPanel.getComponent(i) instanceof ChartPanel) {
        jPanel = (ChartPanel) jContentPanel.getComponent(i);
        jPreferredSize.setSize(jPanel.getSize());
        jContentPanel.remove(i);
      }
    }

    //Get the X and Y axis labels from the existing chart
    JFreeChart oChart = jPanel.getChart();
    XYPlot oPlot = oChart.getXYPlot();
    ValueAxis sXAxis = oPlot.getDomainAxis();
    ValueAxis sYAxis = oPlot.getRangeAxis();
    String sXAxisLabel = sXAxis.getLabel();
    String sYAxisLabel = sYAxis.getLabel();

    jPanel = MakeHistogram(oDataset, sXAxisLabel, sYAxisLabel, oLegend,
                           oRequest, jPreferredSize);

    //Replace the old chart with the new one in our window
    jContentPanel.add(jPanel);
    jFrame.setContentPane(jContentPanel);
  }

  /**
   * Updates a tree map with a fresh dataset.  The old content pane is replaced
   * with a new one, so do not count on existing content pane references to the
   * window working.
   * @param oDataset The new dataset.
   * @param jFrame The window in which to update the chart.
   * @param oLegend The legend for this map.
   * @param fDbhScaleFactor The scale factor which controls how large trees
   * appear
   * @throws ModelException Passing through underlying exceptions.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   * <br>November 15, 2004:  Made this smarter about replacing the chart (LEM)
   */
  public static void UpdateTreeMap(XYZDataset oDataset, JInternalFrame jFrame,
                                   Legend oLegend, float fDbhScaleFactor) throws
      ModelException {

    //Get the X and Y axis labels from the existing chart
    ChartPanel jPanel = (ChartPanel) DataGrapher.FindNamedComponent(jFrame.
        getContentPane(), "tree_map_panel");
    JFreeChart oChart = jPanel.getChart();
    XYPlot oPlot = oChart.getXYPlot();
    ValueAxis sXAxis = oPlot.getDomainAxis();
    ValueAxis sYAxis = oPlot.getRangeAxis();
    String sXAxisLabel = sXAxis.getLabel();
    String sYAxisLabel = sYAxis.getLabel();

    //Get the axis start/ends
    float fXStart = (float) sXAxis.getLowerBound(),
        fXEnd = (float) sXAxis.getUpperBound(),
        fYStart = (float) sYAxis.getLowerBound(),
        fYEnd = (float) sYAxis.getUpperBound();

    JPanel jContentPanel = (JPanel) jFrame.getContentPane();

    //Draw the chart
    JPanel jChartPanel = MakeTreeMap(oDataset, sXAxisLabel, sYAxisLabel,
                                     fXStart, fXEnd, fYStart, fYEnd,
                                     fDbhScaleFactor, oLegend,
                                     jPanel.getSize());

    //Replace the chart
    ReplaceNamedComponent(jContentPanel, jChartPanel, "tree_map_panel");

    //Put the chart in the window
    jFrame.setContentPane(jContentPanel);

  }

  /**
   * Updates a grid map with a fresh dataset.
   * @param oDataset The new dataset.
   * @param jFrame The window in which to update the chart.
   * @param oLegend The legend for this map.
   * @param oRenderer The XYCellRenderer, or null if the defaults are OK.
   * If you provided it to draw the original chart, you must provide it again
   * to keep the grayscale values the same.
   * @throws ModelException Passing through underlying exceptions.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public static void UpdateGridMap(XYZDataset oDataset, JInternalFrame jFrame,
                                   Legend oLegend, XYCellRenderer oRenderer) throws
      ModelException {

    java.awt.Dimension jPreferredSize = new java.awt.Dimension(0, 0);
    java.text.NumberFormat oFormat = java.text.NumberFormat.getInstance();
    oFormat.setMaximumFractionDigits(3);
    String sName;
    int i;

    //Remove the existing chart
    JPanel jContentPane = (JPanel) jFrame.getContentPane();
    ChartPanel jPanel = null;
    for (i = 0; i < jContentPane.getComponentCount(); i++) {
      if (jContentPane.getComponent(i) instanceof ChartPanel) {
        jPanel = (ChartPanel) jContentPane.getComponent(i);
        jPreferredSize.setSize(jPanel.getSize());
        jContentPane.remove(i);
      }
    }

    //Get the X and Y axis labels from the existing chart
    JFreeChart oChart = jPanel.getChart();
    XYPlot oPlot = oChart.getXYPlot();
    ValueAxis sXAxis = oPlot.getDomainAxis();
    ValueAxis sYAxis = oPlot.getRangeAxis();
    String sXAxisLabel = sXAxis.getLabel();
    String sYAxisLabel = sYAxis.getLabel();

    //Get the plot lengths
    double fXPlotLength = sXAxis.getUpperBound();
    double fYPlotLength = sYAxis.getUpperBound();

    if (jPreferredSize.getHeight() == 0) {
      jPreferredSize.setSize(new java.awt.Dimension(Math.min(600,
          (int) fXPlotLength * 3), Math.min(600, (int) fYPlotLength * 3)));
    }

    //Draw the chart
    jPanel = MakeGridMap(oDataset, sXAxisLabel, sYAxisLabel, (int) fXPlotLength,
                         (int) fYPlotLength, oLegend, oRenderer, jPreferredSize);

    //Put the chart in the window
    jContentPane.add(jPanel, BorderLayout.CENTER);

    //Update the grayscale values
    JPanel jGrayscalePanel = (JPanel) FindNamedComponent(jContentPane,
        "jGrayscalePanel");
    JTextField jField = null;
    for (i = 0; i < jGrayscalePanel.getComponentCount(); i++) {
      sName = jGrayscalePanel.getComponent(i).getName();
      if (sName != null) {
        if (sName.equals("jMinValue")) {
          jField = (JTextField) jGrayscalePanel.getComponent(i);
          jField.setText(oFormat.format(oRenderer.GetMinimumValue()));
        }
        else if (sName.equals("jMinPix")) {
          jField = (JTextField) jGrayscalePanel.getComponent(i);
          jField.setText(String.valueOf(oRenderer.GetMinimumColor(0).getBlue()));
        }
        else if (sName.equals("jKneeValue")) {
          jField = (JTextField) jGrayscalePanel.getComponent(i);
          jField.setText(oFormat.format(oRenderer.GetKneeValue()));
        }
        else if (sName.equals("jKneePix")) {
          jField = (JTextField) jGrayscalePanel.getComponent(i);
          jField.setText(String.valueOf(oRenderer.GetKneeColor(0).getBlue()));
        }
        else if (sName.equals("jMaxValue")) {
          jField = (JTextField) jGrayscalePanel.getComponent(i);
          jField.setText(oFormat.format(oRenderer.GetMaximumValue()));
        }
        else if (sName.equals("jMaxPix")) {
          jField = (JTextField) jGrayscalePanel.getComponent(i);
          jField.setText(String.valueOf(oRenderer.GetMaximumColor(0).getBlue()));
        }
      }
    }
    jFrame.setContentPane(jContentPane);
  }

  /**
   * Creates a histogram and returns it in a ChartPanel.
   * @param oDataset The dataset from which to create a histogram.
   * @param sXAxisLabel The X axis label
   * @param sYAxisLabel The Y axis label
   * @param sTitle Chart title.
   * @param oLegend The legend for this chart.
   * @param oRequest The HistogramDataRequest object for this chart.
   * @return The chart in a ChartPanel.
   * @throws ModelException if something goes wrong with the chart.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public static JInternalFrame DrawHistogram(ModelHistogramDataset oDataset,
                                             String sXAxisLabel,
                                             String sYAxisLabel, String sTitle,
                                             Legend oLegend,
                                             HistogramDataRequest oRequest) throws
      ModelException {

    JInternalFrame jFrame = new JInternalFrame(sTitle, //window title
                                               true, //resizable
                                               true, //closeable
                                               true, //maximizable
                                               true); //iconifiable

    ChartPanel jPanel = MakeHistogram(oDataset, sXAxisLabel,
                                      sYAxisLabel,
                                      oLegend, oRequest,
                                      new java.awt.Dimension(750, 270));


    jFrame = AddFileMenu(jFrame, oRequest, oRequest.m_bShowOneTimestep);
    jFrame.getContentPane().setBackground(Color.WHITE);
    jFrame.getContentPane().add(jPanel);
    return jFrame;
  }

  /**
   * Creates a histogram and returns it in a ChartPanel.
   * @param oDataset The dataset from which to create a histogram.
   * @param sXAxisLabel The X axis label
   * @param sYAxisLabel The Y axis label
   * @param sTitle Chart title.
   * @param oLegend The legend for this chart.
   * @param oRequest The HistogramDataRequest object for this chart.
   * @return The chart in a ChartPanel.
   * @throws ModelException if something goes wrong with the chart.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public static JInternalFrame DrawHistogram(ModelHistogramDataset oDataset,
                                             String sXAxisLabel,
                                             String sYAxisLabel, String sTitle,
                                             Legend oLegend,
                                             GridHistogramDataRequest oRequest) throws
      ModelException {

    JInternalFrame jFrame = new JInternalFrame(sTitle, //window title
                                               true, //resizable
                                               true, //closeable
                                               true, //maximizable
                                               true); //iconifiable

    ChartPanel jPanel = MakeHistogram(oDataset, sXAxisLabel,
                                      sYAxisLabel,
                                      oLegend, oRequest,
                                      new java.awt.Dimension(300, 270));


    jFrame = AddFileMenu(jFrame, oRequest, oRequest.m_bShowOneTimestep);
    jFrame.getContentPane().setBackground(Color.WHITE);
    jFrame.getContentPane().add(jPanel);
    return jFrame;
  }

  /**
   * Adds a file menu with a save command to a frame.  The command action
   * string is "WriteChartData".
   * @param jFrame JInternalFrame The frame to add the menu to.
   * @param oListener The listener to add to the file save command.
   * @param bAllTimesteps Whether to add an option for saving for all
   * timesteps.
   * @return JInternalFrame The frame, with the menu added.
   */
  public static JInternalFrame AddFileMenu(JInternalFrame jFrame, java.awt.event.ActionListener oListener, boolean bAllTimesteps) {
    //Create our menu
    JMenuBar jMenuBar = new JMenuBar();
    jMenuBar.setBackground(java.awt.SystemColor.control);
    JMenu jMenuFile = new JMenu();
    jMenuFile.setText("File");
    jMenuFile.setFont(new javawrapper.ModelFont());
    jMenuFile.setMnemonic(java.awt.event.KeyEvent.VK_F);
    jMenuFile.setBackground(java.awt.SystemColor.control);
    JMenuItem jSave = new JMenuItem("Save chart data to file",
                                    java.awt.event.KeyEvent.VK_S);
    jSave.setAccelerator(KeyStroke.getKeyStroke(java.awt.event.KeyEvent.VK_S,
                                                java.awt.event.ActionEvent.
                                                CTRL_MASK));
    jSave.setFont(new javawrapper.ModelFont());
    jSave.setActionCommand("WriteChartData");
    jSave.addActionListener(oListener);
    jSave.setBackground(java.awt.SystemColor.control);
    jMenuFile.add(jSave);

    if (true == bAllTimesteps) {
      jSave = new javax.swing.JMenuItem(
          "Save chart data to file for all timesteps",
          java.awt.event.KeyEvent.VK_T);
      jSave.setAccelerator(javax.swing.KeyStroke.getKeyStroke(java.awt.event.
          KeyEvent.VK_T,
          java.awt.event.ActionEvent.
          CTRL_MASK));
      jSave.setFont(new javawrapper.ModelFont());
      jSave.setActionCommand("WriteWholeRunChartData");
      jSave.addActionListener(oListener);
      jSave.setBackground(java.awt.SystemColor.control);
      jMenuFile.add(jSave);
    }

    jMenuBar.add(jMenuFile);
    jFrame.setJMenuBar(jMenuBar);
    return jFrame;
  }

  /**
   * Draws a line graph using the given dataset and places it in a
   * JInternalFrame as the content pane.  The JInternalFrame is not sized or
   * made visible.
   * <p>
   * The chart comes without a legend or title.
   * @param oDataset The dataset to draw.
   * @param oListener The ActionListener that will respond to the file save
   * command string "WriteChartData"
   * @param sXAxisLabel The X axis label
   * @param sYAxisLabel The Y axis label
   * @param sTitle The window title for the chart
   * @param oLegend The legend for this chart
   * @param iNumSpecies Number of species, total.
   * @return The chart.
   * @throws ModelException if anything goes wrong with drawing the chart.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public static JInternalFrame DrawLineChart(DefaultCategoryDataset oDataset,
                                             java.awt.event.ActionListener oListener,
                                             String sXAxisLabel,
                                             String sYAxisLabel, String sTitle,
                                             Legend oLegend, int iNumSpecies) throws
      ModelException {

    JInternalFrame jFrame = new JInternalFrame(sTitle, //window title
                                               true, //resizable
                                               true, //closeable
                                               true, //maximizable
                                               true); //iconifiable

    ChartPanel jPanel = MakeLineChart(oDataset, sXAxisLabel,
                                      sYAxisLabel,
                                      oLegend, iNumSpecies,
                                      new java.awt.Dimension(500, 270));

//    jPanel.setPreferredSize(new java.awt.Dimension(500, 270));
    jFrame.setContentPane(jPanel);
    jFrame = AddFileMenu(jFrame, oListener, false);
    return jFrame;
  }

  /**
   * Creates a tree map and places it as a content pane in a new JInternalFrame
   * object.  The JInternalFrame object is not sized or made visible.
   * <p>
   * The tree map is created with a variation on the XYBubbleRenderer in
   * JFreeChart called XYTreeRenderer, which draws the trees as circles.  The
   * chart comes without a legend or title.
   * @param oDataset The data to make into a map.
   * @param sXAxisLabel The X axis label.
   * @param sYAxisLabel The Y axis label.
   * @param sTitle The window title for the chart.
   * @param iXLength Length of the plot in the X direction, in meters
   * @param iYLength Length of the plot in the Y direction, in meters
   * @param fDbhScaleFactor The scale factor which controls how large trees
   * appear
   * @param oListener The ActionListener that will respond to the file save
   * command string "WriteChartData"
   * @param oLegend The legend for this chart.
   * @return The chart in a JInternalFrame.
   * @throws ModelException if something goes wrong with the chart.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public static JInternalFrame DrawTreeMap(XYZDataset oDataset,
                                           String sXAxisLabel,
                                           String sYAxisLabel, String sTitle,
                                           int iXLength, int iYLength,
                                           float fDbhScaleFactor,
                                           java.awt.event.ActionListener oListener,
                                           Legend oLegend) throws
      ModelException {

    JInternalFrame jFrame = new JInternalFrame(sTitle, //window title
                                               true, //resizable
                                               true, //closeable
                                               true, //maximizable
                                               true); //iconifiable
    float fAxisStart = 0;
    jFrame.setContentPane(MakeTreeMap(oDataset, sXAxisLabel, sYAxisLabel,
                                      fAxisStart, (float) iXLength, fAxisStart,
                                      (float) iYLength, fDbhScaleFactor,
                                      oLegend,
                                      new java.awt.Dimension(Math.min(600,
        iXLength * 3), Math.min(600, iYLength * 3))));
    jFrame = AddFileMenu(jFrame, oListener, false);
    return jFrame;
  }

  /**
   * Creates a grayscale grid map and places it as a content pane in a new
   * JInternalFrame object.  The JInternalFrame object is not sized or made
   * visible.
   * <p>
   * The grid map is drawn by an XYCellRenderer object.  This object can be
   * pre-created and passed to control the grayscale and where it is applied
   * to various grid values.
   * @param oDataset The data to make into a map.
   * @param sXAxisLabel The X axis label.
   * @param sYAxisLabel The Y axis label.
   * @param sTitle The window title for the chart.
   * @param iXLength Length of the plot in the X direction, in meters
   * @param iYLength Length of the plot in the Y direction, in meters
   * @param oLegend The legend for this chart.
   * @param oRenderer The XYCellRenderer, or null if the defaults are OK.
   * @param oRequest The GridDataRequest object that owns the frame.
   * @return The chart in a JInternalFrame.
   * @throws ModelException if something goes wrong with the chart.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public static JInternalFrame DrawGridMap(XYZDataset oDataset,
                                           String sXAxisLabel,
                                           String sYAxisLabel, String sTitle,
                                           int iXLength, int iYLength,
                                           Legend oLegend,
                                           XYCellRenderer oRenderer,
                                           GridDataRequest oRequest) throws
      ModelException {

    JInternalFrame jFrame = new JInternalFrame(sTitle, //window title
                                               true, //resizable
                                               true, //closeable
                                               true, //maximizable
                                               true); //iconifiable

    //Create the panel that contains the grayscale control values - we'll put
    //it down the left side
    java.text.NumberFormat oFormat = java.text.NumberFormat.getInstance();
    oFormat.setMaximumFractionDigits(3);
    JPanel jGrayscalePanel = new JPanel();
    jGrayscalePanel.setName("jGrayscalePanel");
    jGrayscalePanel.setLayout(new GridLayout(4, 3));
    JLabel jLabel = new JLabel("");
    jLabel.setFont(new javawrapper.ModelFont());
    jGrayscalePanel.add(jLabel);
    jLabel = new JLabel("Value");
    jLabel.setFont(new javawrapper.ModelFont());
    jGrayscalePanel.add(jLabel);
    jLabel = new JLabel("Brightness");
    jLabel.setFont(new javawrapper.ModelFont());
    jGrayscalePanel.add(jLabel);
    jLabel = new JLabel("Min");
    jLabel.setFont(new javawrapper.ModelFont());
    jGrayscalePanel.add(jLabel);
    JTextField jMinValue = new JTextField(oFormat.format(oRenderer.GetMinimumValue()));
    jMinValue.setName("jMinValue");
    jMinValue.setFont(new javawrapper.ModelFont());
    jGrayscalePanel.add(jMinValue);
    JTextField jMinPix = new JTextField(String.valueOf(oRenderer.
        GetMinimumColor(0).getBlue()));
    jMinPix.setFont(new javawrapper.ModelFont());
    jMinPix.setName("jMinPix");
    jGrayscalePanel.add(jMinPix);
    jLabel = new JLabel("Knee");
    jLabel.setFont(new javawrapper.ModelFont());
    jGrayscalePanel.add(jLabel);
    JTextField jKneeValue = new JTextField(oFormat.format(oRenderer.
        GetKneeValue()));
    jKneeValue.setFont(new javawrapper.ModelFont());
    jKneeValue.setName("jKneeValue");
    jGrayscalePanel.add(jKneeValue);
    JTextField jKneePix = new JTextField(String.valueOf(oRenderer.GetKneeColor(
        0).getBlue()));
    jKneePix.setFont(new javawrapper.ModelFont());
    jKneePix.setName("jKneePix");
    jGrayscalePanel.add(jKneePix);
    jLabel = new JLabel("Max");
    jLabel.setFont(new javawrapper.ModelFont());
    jGrayscalePanel.add(jLabel);
    JTextField jMaxValue = new JTextField(oFormat.format(oRenderer.
        GetMaximumValue()));
    jMaxValue.setName("jMaxValue");
    jMaxValue.setFont(new javawrapper.ModelFont());
    jGrayscalePanel.add(jMaxValue);
    JTextField jMaxPix = new JTextField(String.valueOf(oRenderer.
        GetMaximumColor(0).getBlue()));
    jMaxPix.setName("jMaxPix");
    jMaxPix.setFont(new javawrapper.ModelFont());
    jGrayscalePanel.add(jMaxPix);
    jGrayscalePanel.setBorder(BorderFactory.createEmptyBorder(10, 10, 10, 10));

    //Create a panel for an update button and instructions - put the button at
    //NORTH so it stays right under the brightness values
    JButton jUpdateButton = new JButton("Update");
    jUpdateButton.setFont(new javawrapper.ModelFont());
    jUpdateButton.addActionListener(new GrayscaleListener(oRequest));
    JTextArea jInstructions = new JTextArea("Color lightens linearly with "+
                                            "map value from min to knee, and "+
                                            "from knee to max. If map looks "+
                                            "too dark, lower knee.  If too "+
                                            "light, raise knee.");
    jInstructions.setEditable(false);
    jInstructions.setLineWrap(true);
    jInstructions.setWrapStyleWord(true);
    jInstructions.setFont(new javawrapper.ModelFont());
    jInstructions.setBackground(SystemColor.control);

    JPanel jButtonPanel = new JPanel(new BorderLayout());
    jButtonPanel.add(jUpdateButton, BorderLayout.NORTH);
    jButtonPanel.add(jInstructions, BorderLayout.CENTER);
    jButtonPanel.setBorder(BorderFactory.createEmptyBorder(0, 10, 0, 10));

    //Encapsulate the grayscale and the update button together in one panel.
    //Put the jGrayscalePanel as NORTH so that the individual text boxes
    //stay the right size
    JPanel jControlPanel = new JPanel(new BorderLayout());
    jControlPanel.setName("jControlPanel");
    jControlPanel.add(jGrayscalePanel, BorderLayout.NORTH);
    jControlPanel.add(jButtonPanel, BorderLayout.CENTER);

    //Create the chart panel
    JPanel jPanel = MakeGridMap(oDataset, sXAxisLabel, sYAxisLabel,
                                iXLength, iYLength,
                                oLegend, oRenderer,
                                new java.awt.Dimension(Math.min(600,
        iXLength * 3), Math.min(600, iYLength * 3)));

    //Add everything
    jFrame.getContentPane().setLayout(new BorderLayout());
    jFrame.getContentPane().add(jControlPanel, BorderLayout.WEST);
    jFrame.getContentPane().add(jPanel, BorderLayout.CENTER);
    jFrame = AddFileMenu(jFrame, oRequest, oRequest.m_bShowOneTimestep);

    return jFrame;
  }

  /**
   * Draws a line graph using the given dataset and returns a ChartPanel
   * containing the desired chart.
   * <p>
   * The chart comes without a legend or title.
   * @param dataset The dataset to draw.
   * @param sXAxisLabel The X axis label
   * @param sYAxisLabel The Y axis label
   * @param oLegend The legend for this chart
   * @param iNumSpecies Number of species, total.
   * @param jPreferredSize Chart's size.
   * @return The chart.
   * @throws ModelException if anything goes wrong with drawing the chart.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  private static ChartPanel MakeLineChart(DefaultCategoryDataset dataset,
                                         String sXAxisLabel,
                                         String sYAxisLabel,
                                         Legend oLegend, int iNumSpecies,
                                         Dimension jPreferredSize) throws
      ModelException {
    //I based this largely from the LineChartDemo1 class that came with
    //JFreeChart.

    // create the chart...
    JFreeChart oChart = ChartFactory.createLineChart(
        "", // chart title
        sXAxisLabel, // X axis label
        sYAxisLabel, // Y axis label
        dataset, // data
        PlotOrientation.VERTICAL, // orientation
        false, // include legend
        true, // tooltips
        false // urls
        );

    oChart.setBackgroundPaint(Color.WHITE);

    CategoryPlot plot = oChart.getCategoryPlot();

    //Customize the X axis...
    CategoryAxis domainAxis = plot.getDomainAxis();
    domainAxis.setVerticalCategoryLabels(true);
    domainAxis.setSkipCategoryLabelsToFit(true);

    //Customize the Y axis...
    NumberAxis rangeAxis = (NumberAxis) plot.getRangeAxis();
    rangeAxis.setAutoRangeIncludesZero(true);
    // rangeAxis.setLabelAngle(Math.PI);

    //Customize the renderer...
    LineAndShapeRenderer oRenderer = (LineAndShapeRenderer) plot.getRenderer();
    oRenderer.setDrawShapes(true);

    //Set the appropriate color for each species
    for (int j = 0; j < dataset.getRowCount(); j++) {
      String sRowKey = (String) dataset.getRowKey(j);
      for (int i = 0; i < iNumSpecies; i++) {
        if (sRowKey.compareTo(oLegend.GetSpeciesCodeName(i)) == 0) {
          oRenderer.setSeriesPaint(j, oLegend.GetSpeciesColor(i));
        }
        else if (sRowKey.compareTo("Total") == 0) {
          oRenderer.setSeriesPaint(j, Color.BLACK);
        }
      }
    }

    oRenderer.setItemLabelTextAnchor(TextAnchor.BOTTOM_CENTER);
    oRenderer.setItemLabelAnchor(ItemLabelAnchor.OUTSIDE12);

    ChartPanel oPanel = new ChartPanel(oChart);
    oPanel.setPreferredSize(jPreferredSize);
//    return new ChartPanel(oChart);
    return oPanel;
  }

  /**
   * Creates a tree map and returns it in a ChartPanel.
   * @param oDataset The data to make into a map.
   * @param sXAxisLabel The X axis label
   * @param sYAxisLabel The Y axis label
   * @param fXStart Start coordinate of the X axis, in meters
   * @param fXEnd End coordinate of the X axis, in meters
   * @param fYStart Start coordinate of the Y axis, in meters
   * @param fYEnd End coordinate of the Y axis, in meters
   * @param fDbhScaleFactor The scale factor which controls how large trees
   * appear
   * @param oLegend The legend for this chart.
   * @param jPreferredSize Chart's size.
   * @return The chart in a ChartPanel.
   * @throws ModelException if something goes wrong with the chart.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   * <br>November 15, 2004:  Added setting of both plot end points
   */
  private static ChartPanel MakeTreeMap(XYZDataset oDataset, String sXAxisLabel,
                                       String sYAxisLabel,
                                       float fXStart, float fXEnd,
                                       float fYStart, float fYEnd,
                                       float fDbhScaleFactor,
                                       Legend oLegend,
                                       Dimension jPreferredSize) throws
      ModelException {
    JFreeChart oChart = ChartFactory.createScatterPlot(
        "",
        sXAxisLabel, sYAxisLabel,
        oDataset,
        PlotOrientation.VERTICAL,
        false,
        true,
        false
        );

    oChart.setBackgroundPaint(Color.WHITE);

    //Get the plot and hook it to an XYTreeRenderer
    XYPlot oPlot = (XYPlot) oChart.getPlot();
    XYTreeRenderer oRenderer = new XYTreeRenderer();

    if (fDbhScaleFactor > 0) {
      oRenderer.SetScaleFactor(fDbhScaleFactor);
    }

    //Set the appropriate color for each species
    for (int i = 0; i < oDataset.getSeriesCount(); i++) {
      String sName = oDataset.getSeriesName(i);
      oRenderer.setSeriesPaint(i, oLegend.GetSpeciesColor(sName));
    }

    oPlot.setRenderer(oRenderer);

    //Force the axes to include the entire plot length
    NumberAxis oXAxis = (NumberAxis) oChart.getXYPlot().getDomainAxis();
    oXAxis.setLowerBound(fXStart);
    oXAxis.setUpperBound(fXEnd);

    NumberAxis oYAxis = (NumberAxis) oChart.getXYPlot().getRangeAxis();
    oYAxis.setLowerBound(fYStart);
    oYAxis.setUpperBound(fYEnd);

    ChartPanel oChartPanel = new ChartPanel(oChart);
    oChartPanel.setName("tree_map_panel");
    //   oChartPanel.setPreferredSize(new java.awt.Dimension(Math.min(600,
    //       iXLength * 3), Math.min(600, iYLength * 3)));
    oChartPanel.setPreferredSize(jPreferredSize);
    //oChartPanel.setVerticalAxisTrace(true);
    //oChartPanel.setHorizontalAxisTrace(true);
    oChartPanel.setVerticalZoom(true);
    oChartPanel.setHorizontalZoom(true);
    return oChartPanel;
  }

  /**
   * Creates a grid map and returns it in a ChartPanel.
   * @param oDataset The data to make into a map.
   * @param sXAxisLabel The X axis label
   * @param sYAxisLabel The Y axis label
   * @param iXLength Length of the plot in the X direction, in meters
   * @param iYLength Length of the plot in the Y direction, in meters
   * @param oLegend The legend for this chart.
   * @param oRenderer XYCellRenderer object.  If NULL, this will create one.
   * @param jPreferredSize Chart's size.
   * @return The chart in a ChartPanel.
   * @throws ModelException if something goes wrong with the chart.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  private static ChartPanel MakeGridMap(XYZDataset oDataset, String sXAxisLabel,
                                       String sYAxisLabel, int iXLength,
                                       int iYLength,
                                       Legend oLegend, XYCellRenderer oRenderer,
                                       Dimension jPreferredSize
                                       ) throws ModelException {
    JFreeChart oChart = ChartFactory.createScatterPlot(
        "",
        sXAxisLabel, sYAxisLabel,
        oDataset,
        PlotOrientation.VERTICAL,
        false,
        true,
        false
        );

    oChart.setBackgroundPaint(Color.WHITE);

    //Get the plot and hook it to the XYCellRenderer
    if (null == oRenderer) {
      oRenderer = new XYCellRenderer();
    }
    XYPlot oPlot = (XYPlot) oChart.getPlot();
    oPlot.setRenderer(oRenderer);

    //Force the axes to include the entire plot length
    NumberAxis oXAxis = (NumberAxis) oChart.getXYPlot().getDomainAxis();
    oXAxis.setLowerBound(0);
    oXAxis.setUpperBound( (double) iXLength);

    NumberAxis oYAxis = (NumberAxis) oChart.getXYPlot().getRangeAxis();
    oYAxis.setLowerBound(0);
    oYAxis.setUpperBound( (double) iYLength);

    ChartPanel oChartPanel = new ChartPanel(oChart);
//    oChartPanel.setPreferredSize(new java.awt.Dimension(Math.min(600,
//        iXLength * 3), Math.min(600, iYLength * 3)));
    oChartPanel.setPreferredSize(jPreferredSize);

    return oChartPanel;
  }

  /**
   * Makes an overlaid map.
   * @param oPrimaryDataset XYZDataset First dataset to chart
   * @param oSecondaryDataset XYZDataset Second dataset to chart
   * @param sAxisLabel String Label of X axis
   * @param sYAxisLabel String Label of Y axis
   * @param iXLength int Length of plot in X direction in meters
   * @param iYLength int Length of plot in Y direction in meters
   * @param oPrimaryRenderer XYItemRenderer Renderer for primary dataset
   * @param oSecondaryRenderer XYItemRenderer Renderer for secondary dataset
   * @param jPreferredSize Dimension Preferred chart size
   * @return ChartPanel Overlaid map chart panel
   */
  public static ChartPanel MakeOverlaidMap(XYZDataset oPrimaryDataset,
                                           XYZDataset oSecondaryDataset,
                                           String sAxisLabel,
                                           String sYAxisLabel,
                                           int iXLength, int iYLength,
                                           XYItemRenderer oPrimaryRenderer,
                                           XYItemRenderer oSecondaryRenderer,
                                           Dimension jPreferredSize) {
    //Create the plot ...
    NumberAxis oXAxis = new NumberAxis(sAxisLabel);
    oXAxis.setAutoRangeIncludesZero(true);
    oXAxis.setLowerBound(0);
    oXAxis.setUpperBound( (double) iXLength);
    NumberAxis oYAxis = new NumberAxis(sYAxisLabel);
    oYAxis.setAutoRangeIncludesZero(true);
    oYAxis.setLowerBound(0);
    oYAxis.setUpperBound( (double) iYLength);

    XYPlot oPlot = new XYPlot(oPrimaryDataset, oXAxis, oYAxis, oPrimaryRenderer);

    //Add the second dataset and renderer...
    if (oSecondaryDataset != null) {
      oPlot.setSecondaryDataset(0, oSecondaryDataset);
      oPlot.setSecondaryRenderer(0, oSecondaryRenderer);
    }

    //Return a new chart containing the overlaid plot...
    JFreeChart oChart = new JFreeChart("", JFreeChart.DEFAULT_TITLE_FONT, oPlot, false);

    ChartPanel oChartPanel = new ChartPanel(oChart);
    oChartPanel.setPreferredSize(jPreferredSize);
    return oChartPanel;
  }

  /**
   * Creates a set of overview tables packaged into a JPanel.  The overview
   * tables contain information on density and basal area for each tree type.
   * The four tables appear in two rows:  adults and saplings on top, and
   * seedlings and snags on the bottom.  This method expects no null arrays,
   * although null values within the array are all right.  They will appear
   * as blanks.  A column of species names is added to each dataset as the
   * first column in the table.
   * <p>
   * The tables themselves are named, so they can be found later and updated
   * without rebuilding the entire panel.  The seedling table is named
   * "seedling_table", the sapling table is named "sapling_table", the
   * adult table is named "adult_table", and the snag table is named
   * "snag_table".
   *
   * @param sChartTitle Chart window table.
   * @param p_oSeedlingData The seedling dataset.  The first array index is
   * rows, one per species, with the last row showing totals.  The single
   * column is for absolute density.
   * @param p_oSaplingData The sapling dataset.  The first array index is
   * rows, one per species, with the last row showing totals.  The four
   * columns are 1) absolute density, 2) percent of density, 3) absolute basal
   * area, and 4) percent basal area.
   * @param p_oAdultData The adult dataset.  The first array index is
   * rows, one per species, with the last row showing totals.  The four
   * columns are 1) absolute density, 2) percent of density, 3) absolute basal
   * area, and 4) percent basal area.
   * @param p_oSnagData The snag dataset.  The first array index is
   * rows, one per species, with the last row showing totals.  The four
   * columns are 1) absolute density, 2) percent of density, 3) absolute basal
   * area, and 4) percent basal area.
   * @param oLegend The legend that goes with these tables, for extracting species
   * display names.
   * @param oListener The ActionListener that will respond to the file save
   * command string "WriteChartData"
   * @return The panel with the four tables within it plus appropriate labels.
   * @throws ModelException passed through from calling methods; should never
   * be thrown.
   */
  public static JInternalFrame DrawOverviewTables(String sChartTitle,
                                          Object[][] p_oSeedlingData,
                                          Object[][] p_oSaplingData,
                                          Object[][] p_oAdultData,
                                          Object[][] p_oSnagData,
                                          Legend oLegend,
                                          java.awt.event.ActionListener oListener) throws javawrapper.
      ModelException {
    JInternalFrame jFrame = new JInternalFrame(sChartTitle, //window title
                                               true, //resizable
                                               true, //closeable
                                               true, //maximizable
                                               true); //iconifiable
    jFrame.setContentPane(MakeOverviewTables(p_oSeedlingData, p_oSaplingData,
                                             p_oAdultData, p_oSnagData, oLegend));
    jFrame = AddFileMenu(jFrame, oListener, false);
    return jFrame;

  }

  /**
   * Creates a set of overview tables packaged into a JPanel.  The overview
   * tables contain information on density and basal area for each tree type.
   * The four tables appear in two rows:  adults and saplings on top, and
   * seedlings and snags on the bottom.  This method expects no null arrays,
   * although null values within the array are all right.  They will appear
   * as blanks.  A column of species names is added to each dataset as the
   * first column in the table.
   * <p>
   * The tables themselves are named, so they can be found later and updated
   * without rebuilding the entire panel.  The seedling table is named
   * "seedling_table", the sapling table is named "sapling_table", the
   * adult table is named "adult_table", and the snag table is named
   * "snag_table".
   *
   * @param p_oSeedlingData The seedling dataset.  The first array index is
   * rows, one per species, with the last row showing totals.  The single
   * column is for absolute density.
   * @param p_oSaplingData The sapling dataset.  The first array index is
   * rows, one per species, with the last row showing totals.  The four
   * columns are 1) absolute density, 2) percent of density, 3) absolute basal
   * area, and 4) percent basal area.
   * @param p_oAdultData The adult dataset.  The first array index is
   * rows, one per species, with the last row showing totals.  The four
   * columns are 1) absolute density, 2) percent of density, 3) absolute basal
   * area, and 4) percent basal area.
   * @param p_oSnagData The snag dataset.  The first array index is
   * rows, one per species, with the last row showing totals.  The four
   * columns are 1) absolute density, 2) percent of density, 3) absolute basal
   * area, and 4) percent basal area.
   * @param oLegend The legend that goes with these tables, for extracting species
   * display names.
   * @return The panel with the four tables within it plus appropriate labels.
   * @throws ModelException passed through from calling methods; should never
   * be thrown.
   */
  private static JPanel MakeOverviewTables(Object[][] p_oSeedlingData,
                                          Object[][] p_oSaplingData,
                                          Object[][] p_oAdultData,
                                          Object[][] p_oSnagData,
                                          Legend oLegend) throws javawrapper.
      ModelException {

    String[][] p_sColumnNames = new String[4][];
    String[] p_sLabels = new String[4],
        p_sNames = new String[4];
    Object[][][] p_oAllData = new Object[4][][];
    JPanel[] p_jTablePanels = new JPanel[4]; //each table packaged with an identifying label
    int iSpeciesColumnWidth, i, j, k,
        iNumSpecies = p_oSeedlingData.length - 1;

    //Figure out how big to make our species column
    iSpeciesColumnWidth = 0;
    JLabel jTempLabel = new JLabel();
    jTempLabel.setFont(new javawrapper.ModelFont());
    for (i = 0; i < iNumSpecies; i++) {
      jTempLabel.setText(oLegend.GetSpeciesDisplayName(i));
      if (iSpeciesColumnWidth < jTempLabel.getPreferredSize().getWidth()) {
        iSpeciesColumnWidth = (int) jTempLabel.getPreferredSize().getWidth();
      }
    }
    iSpeciesColumnWidth *= 1.25;

    //Set up our arrays so we can just loop on everything to build the tables
    //Array indexes are the types in order - seedlings = 0, saplings = 1, etc.
    p_sColumnNames[0] = new String[] {
        "Species", "Abs Den"};
    for (i = 1; i < 4; i++) {
      p_sColumnNames[i] = new String[] {
          "Species", "Abs Den", "Rel Den", "Abs BA", "Rel BA"};
    }
    p_sLabels[0] = "Seedlings";
    p_sLabels[1] = "Saplings";
    p_sLabels[2] = "Adults";
    p_sLabels[3] = "Snags";
    p_sNames[0] = "seedling_table";
    p_sNames[1] = "sapling_table";
    p_sNames[2] = "adult_table";
    p_sNames[3] = "snag_table";
    p_oAllData[0] = new Object[p_oSeedlingData.length][p_oSeedlingData[0].
        length + 1];
    for (i = 0; i < p_oSeedlingData.length; i++) {
      for (j = 0; j < p_oSeedlingData[i].length; j++) {
        p_oAllData[0][i][j + 1] = p_oSeedlingData[i][j];
      }
    }
    p_oAllData[1] = new Object[p_oSaplingData.length][p_oSaplingData[0].length +
        1];
    for (i = 0; i < p_oSaplingData.length; i++) {
      for (j = 0; j < p_oSaplingData[i].length; j++) {
        p_oAllData[1][i][j + 1] = p_oSaplingData[i][j];
      }
    }
    p_oAllData[2] = new Object[p_oAdultData.length][p_oAdultData[0].length + 1];
    for (i = 0; i < p_oAdultData.length; i++) {
      for (j = 0; j < p_oAdultData[i].length; j++) {
        p_oAllData[2][i][j + 1] = p_oAdultData[i][j];
      }
    }
    p_oAllData[3] = new Object[p_oSnagData.length][p_oSnagData[0].length + 1];
    for (i = 0; i < p_oSnagData.length; i++) {
      for (j = 0; j < p_oSnagData[i].length; j++) {
        p_oAllData[3][i][j + 1] = p_oSnagData[i][j];
      }
    }

    //Create all our tables
    for (i = 0; i < 4; i++) {
      //Add the species names to the dataset
      for (j = 0; j < p_oAllData[i].length - 1; j++) {
        p_oAllData[i][j][0] = oLegend.GetSpeciesDisplayName(j);
      }
      p_oAllData[i][p_oAllData[i].length - 1][0] = "Total";
      JTable jTable = new JTable(p_oAllData[i], p_sColumnNames[i]);
      jTable.setGridColor(Color.WHITE);
      jTable.setBackground(Color.WHITE);
      jTable.setFont(new javawrapper.ModelFont());
      jTable.getColumnModel().getColumn(0).setPreferredWidth(
          iSpeciesColumnWidth);
      jTable.setName(p_sNames[i]);
      jTable.getTableHeader().setFont(new javawrapper.ModelFont());
      //Set the cell renderer to right-justify
      javax.swing.table.DefaultTableCellRenderer jRenderer = new javax.swing.table.DefaultTableCellRenderer();
      jRenderer.setHorizontalAlignment(javax.swing.SwingConstants.RIGHT);
      for (k = 1; k < jTable.getColumnCount(); k++) {
        jTable.setDefaultRenderer(jTable.getColumnClass(k), jRenderer);
      }

      //Put the table into the larger panel
      jTempLabel = new JLabel(p_sLabels[i]);
      jTempLabel.setFont(new javawrapper.ModelFont());
      p_jTablePanels[i] = new JPanel(new java.awt.BorderLayout());
      p_jTablePanels[i].setName("table_panel_" + i);
      p_jTablePanels[i].add(jTempLabel, java.awt.BorderLayout.NORTH);
      JPanel jMiniPanel = new JPanel(new java.awt.BorderLayout());
      jMiniPanel.add(jTable.getTableHeader(), java.awt.BorderLayout.NORTH);
      jMiniPanel.add(jTable, java.awt.BorderLayout.CENTER);
      p_jTablePanels[i].add(jMiniPanel);
    }

    //Create the panels for the two rows
    JPanel jTopRowPanel = new JPanel(new java.awt.FlowLayout(java.awt.
        FlowLayout.
        LEFT));
    JPanel jBottomRowPanel = new JPanel(new java.awt.FlowLayout(java.awt.
        FlowLayout.
        LEFT));
    jTopRowPanel.add(p_jTablePanels[2]);
    jTopRowPanel.add(p_jTablePanels[1]);
    jBottomRowPanel.add(p_jTablePanels[3]);
    jBottomRowPanel.add(p_jTablePanels[0]);

    //Create a panel for all the tables together
    JPanel jAllTablePanel = new JPanel();
    jAllTablePanel.setLayout(new javax.swing.BoxLayout(jAllTablePanel,
        javax.swing.BoxLayout.PAGE_AXIS));

    jAllTablePanel.add(jTopRowPanel);
    jAllTablePanel.add(jBottomRowPanel);
    JScrollPane jScroller = new JScrollPane(jAllTablePanel);

    //Wrap with a legend
    JPanel jTempPanel = new JPanel(new java.awt.BorderLayout());
    jTempLabel = new JLabel("Density is in #/ha, basal area in sq. m./ha");
    jTempLabel.setFont(new javawrapper.ModelFont());
    jTempPanel.add(jTempLabel, java.awt.BorderLayout.NORTH);
    jTempPanel.add(jScroller);

    return jTempPanel;

  }

  /*  public static JPanel MakeOverviewTables(Object[][] p_oSeedlingData,
                                            Object[][] p_oSaplingData,
                                            Object[][] p_oAdultData,
                                            Object[][] p_oSnagData,
                                            Legend oLegend) throws javawrapper.
        ModelException {

      String[][] p_sColumnNames = new String[4][];
      String[] p_sLabels = new String[4],
          p_sNames = new String[4];
      Object[][][] p_oAllData = new Object[4][][];
      JPanel[] p_jTablePanels = new JPanel[4]; //each table packaged with an identifying label
      int iSpeciesColumnWidth, iTableHeight, i, j,
          iNumSpecies = p_oSeedlingData.length - 1;

      //Figure out how big to make our species column and our table height
      iSpeciesColumnWidth = 0;
      JLabel jTempLabel = new JLabel();
      jTempLabel.setFont(new javawrapper.ModelFont());
      for (i = 0; i < iNumSpecies; i++) {
        jTempLabel.setText(oLegend.GetSpeciesDisplayName(i));
        if (iSpeciesColumnWidth < jTempLabel.getPreferredSize().getWidth()) {
          iSpeciesColumnWidth = (int) jTempLabel.getPreferredSize().getWidth();
        }
      }
      iSpeciesColumnWidth *= 1.25;
      iTableHeight = (int) (jTempLabel.getPreferredSize().getHeight() *
                            iNumSpecies * 1.05);

      //Set up our arrays so we can just loop on everything to build the tables
      //Array indexes are the types in order - seedlings = 0, saplings = 1, etc.
      p_sColumnNames[0] = new String[] {
          "Species", "Abs Den"};
      for (i = 1; i < 4; i++) {
        p_sColumnNames[i] = new String[] {
            "Species", "Abs Den", "Rel Den", "Abs BA", "Rel BA"};
      }
      p_sLabels[0] = "Seedlings";
      p_sLabels[1] = "Saplings";
      p_sLabels[2] = "Adults";
      p_sLabels[3] = "Snags";
      p_sNames[0] = "seedling_table";
      p_sNames[1] = "sapling_table";
      p_sNames[2] = "adult_table";
      p_sNames[3] = "snag_table";
      p_oAllData[0] = new Object[p_oSeedlingData.length][p_oSeedlingData[0].
          length + 1];
      for (i = 0; i < p_oSeedlingData.length; i++) {
        for (j = 0; j < p_oSeedlingData[i].length; j++) {
          p_oAllData[0][i][j + 1] = p_oSeedlingData[i][j];
        }
      }
   p_oAllData[1] = new Object[p_oSaplingData.length][p_oSaplingData[0].length +
          1];
      for (i = 0; i < p_oSaplingData.length; i++) {
        for (j = 0; j < p_oSaplingData[i].length; j++) {
          p_oAllData[1][i][j + 1] = p_oSaplingData[i][j];
        }
      }
   p_oAllData[2] = new Object[p_oAdultData.length][p_oAdultData[0].length + 1];
      for (i = 0; i < p_oAdultData.length; i++) {
        for (j = 0; j < p_oAdultData[i].length; j++) {
          p_oAllData[2][i][j + 1] = p_oAdultData[i][j];
        }
      }
   p_oAllData[3] = new Object[p_oSnagData.length][p_oSnagData[0].length + 1];
      for (i = 0; i < p_oSnagData.length; i++) {
        for (j = 0; j < p_oSnagData[i].length; j++) {
          p_oAllData[3][i][j + 1] = p_oSnagData[i][j];
        }
      }

      //Create all our tables
      for (i = 0; i < 4; i++) {
        //Add the species names to the dataset
        for (j = 0; j < p_oAllData[i].length - 1; j++) {
          p_oAllData[i][j][0] = oLegend.GetSpeciesDisplayName(j);
        }
        p_oAllData[i][p_oAllData[i].length - 1][0] = "Total";
        JTable jTable = new JTable(p_oAllData[i], p_sColumnNames[i]);
        jTable.setGridColor(Color.WHITE);
        jTable.setBackground(Color.WHITE);
        jTable.setFont(new javawrapper.ModelFont());
        jTable.getColumnModel().getColumn(0).setPreferredWidth(
            iSpeciesColumnWidth);
        //Size the table - the first column, plus allow 65 pix for each other column
        jTable.setPreferredScrollableViewportSize(new java.awt.Dimension(
            iSpeciesColumnWidth + (65 * (p_oAllData[i][0].length - 1)),
            iTableHeight));
        //Name the table
        jTable.setName(p_sNames[i]);

        //Create the scroll pane and add the table to it.
        JScrollPane jScrollPane = new JScrollPane(jTable);
        jScrollPane.setBackground(Color.WHITE);
        jScrollPane.getViewport().setBackground(Color.WHITE);
        jTable.getTableHeader().setFont(new javawrapper.ModelFont());

        //Put the table into the larger panel
        jTempLabel = new JLabel(p_sLabels[i]);
        jTempLabel.setFont(new javawrapper.ModelFont());
        p_jTablePanels[i] = new JPanel(new java.awt.BorderLayout());
        p_jTablePanels[i].setName("table_panel_" + i);
        p_jTablePanels[i].add(jTempLabel, java.awt.BorderLayout.NORTH);
        p_jTablePanels[i].add(jScrollPane, java.awt.BorderLayout.CENTER);
      }

      //Create the panels for the two rows
      JPanel jTopRowPanel = new JPanel(new java.awt.FlowLayout(java.awt.
          FlowLayout.
          LEFT));
      JPanel jBottomRowPanel = new JPanel(new java.awt.FlowLayout(java.awt.
          FlowLayout.
          LEFT));
      jTopRowPanel.add(p_jTablePanels[2]);
      jTopRowPanel.add(p_jTablePanels[1]);
      jBottomRowPanel.add(p_jTablePanels[3]);
      jBottomRowPanel.add(p_jTablePanels[0]);

      //Create a panel for all the tables together
      JPanel jAllTablePanel = new JPanel();
      jAllTablePanel.setLayout(new javax.swing.BoxLayout(jAllTablePanel,
          javax.swing.BoxLayout.PAGE_AXIS));

      jAllTablePanel.add(jTopRowPanel);
      jAllTablePanel.add(jBottomRowPanel);

      //Wrap with a legend
      JPanel jTempPanel = new JPanel(new java.awt.BorderLayout());
      jTempLabel = new JLabel("Density is in #/ha, basal area in sq. m./ha");
      jTempLabel.setFont(new javawrapper.ModelFont());
      jTempPanel.add(jTempLabel, java.awt.BorderLayout.NORTH);
      jTempPanel.add(jAllTablePanel, java.awt.BorderLayout.CENTER);

      return jTempPanel;

    } */


  /**
   * Updates the overview tables with new data without rebuilding the whole
   * window.  This finds the tables within a panel and sets the values
   * cell-by-cell from the datasets passed in.  The tables can be children at
   * any level of the panel being passed - it will be recursively searched to
   * find them.  The species names in column one are not touched, only the data
   * values.
   * @param jTablePanel The component containing the tables within it.
   * @param p_oSeedlingData The seedling dataset.  The first array index is
   * rows, one per species, with the last row showing totals.  The single
   * column is for absolute density.
   * @param p_oSaplingData The sapling dataset.  The first array index is
   * rows, one per species, with the last row showing totals.  The four
   * columns are 1) absolute density, 2) percent of density, 3) absolute basal
   * area, and 4) percent basal area.
   * @param p_oAdultData The adult dataset.  The first array index is
   * rows, one per species, with the last row showing totals.  The four
   * columns are 1) absolute density, 2) percent of density, 3) absolute basal
   * area, and 4) percent basal area.
   * @param p_oSnagData The snag dataset.  The first array index is
   * rows, one per species, with the last row showing totals.  The four
   * columns are 1) absolute density, 2) percent of density, 3) absolute basal
   * area, and 4) percent basal area.
   */
  public static void UpdateOverviewTables(JPanel jTablePanel,
                                          Object[][] p_oSeedlingData,
                                          Object[][] p_oSaplingData,
                                          Object[][] p_oAdultData,
                                          Object[][] p_oSnagData) {
    JTable jTable;
    java.awt.Component jComponent;
    int i, j;

    //Seedling data
    //Find the seedling table
    jComponent = FindNamedComponent(jTablePanel, "seedling_table");
    if (jComponent != null) {

      jTable = (JTable) jComponent;

      //Update it with the new data
      for (i = 0; i < p_oSeedlingData.length; i++) {
        for (j = 0; j < p_oSeedlingData[i].length; j++) {
          jTable.setValueAt(p_oSeedlingData[i][j], i, j + 1);
        }
      }
    }

    //Sapling data
    //Find the sapling table
    jComponent = FindNamedComponent(jTablePanel, "sapling_table");
    if (jComponent != null) {

      jTable = (JTable) jComponent;

      //Update it with the new data
      for (i = 0; i < p_oSaplingData.length; i++) {
        for (j = 0; j < p_oSaplingData[i].length; j++) {
          jTable.setValueAt(p_oSaplingData[i][j], i, j + 1);
        }
      }
    }

    //Adult data
    //Find the adult table
    jComponent = FindNamedComponent(jTablePanel, "adult_table");
    if (jComponent != null) {

      jTable = (JTable) jComponent;

      //Update it with the new data
      for (i = 0; i < p_oAdultData.length; i++) {
        for (j = 0; j < p_oAdultData[i].length; j++) {
          jTable.setValueAt(p_oAdultData[i][j], i, j + 1);
        }
      }
    }

    //Snag data
    //Find the snag table
    jComponent = FindNamedComponent(jTablePanel, "snag_table");
    if (jComponent != null) {

      jTable = (JTable) jComponent;

      //Update it with the new data
      for (i = 0; i < p_oSnagData.length; i++) {
        for (j = 0; j < p_oSnagData[i].length; j++) {
          jTable.setValueAt(p_oSnagData[i][j], i, j + 1);
        }
      }
    }
  }

  /**
   * Finds a named component within another component.  This recursively
   * searches all children of the parent component to find the named
   * component.
   * @param jParent The parent component in which to search.
   * @param sName The name of the component to find.
   * @return The component, or null if it could not be found.
   */
  public static java.awt.Component FindNamedComponent(java.awt.Component
      jParent, String sName) {

    java.awt.Component jFound = null;

    //Is the parent the component in question?
    String sChildName = jParent.getName();
    if (sChildName != null && sChildName.equals(sName)) {
      return jParent;
    }

    //Only search if the parent component is a container of some kind
    if (jParent instanceof java.awt.Container) {
      java.awt.Container jContainer = (java.awt.Container) jParent;
      java.awt.Component jChild;
      int i, iNumComponents = jContainer.getComponentCount();

      for (i = 0; i < iNumComponents; i++) {
        jChild = jContainer.getComponent(i);
        sChildName = jChild.getName();
        if (sChildName != null && sChildName.equals(sName)) {
          jFound = jChild;
          break;
        }
        else {
          jFound = FindNamedComponent(jChild, sName);
          if (jFound != null) {
            break;
          }
        }
      }
    }
    return jFound;
  }

  /**
   * Replaces a named component with another component.  This recursively
   * searches all children of the parent component to find the named
   * component, and then places the new component in its place.
   * @param jParent The parent component in which to search.
   * @param jToReplaceWith The component with which to replace the named component.
   * @param sName The name of the component to find.
   * @return The component, or null if it could not be found.
   */
  public static boolean ReplaceNamedComponent(java.awt.Component
                                              jParent,
                                              java.awt.Component jToReplaceWith,
                                              String sName) {

    boolean bReplaced = false;

    //Only search if the parent component is a container of some kind
    if (jParent instanceof java.awt.Container) {
      java.awt.Container jContainer = (java.awt.Container) jParent;
      java.awt.Component jChild;
      int i, iNumComponents = jContainer.getComponentCount();

      for (i = 0; i < iNumComponents; i++) {
        jChild = jContainer.getComponent(i);
        String sChildName = jChild.getName();
        if (sChildName != null && sChildName.equals(sName)) {
          jContainer.remove(i);
          jContainer.add(jToReplaceWith, i);
          bReplaced = true;
          break;
        }
        else {
          bReplaced = ReplaceNamedComponent(jChild, jToReplaceWith, sName);
          if (bReplaced) {
            break;
          }
        }
      }
    }
    return bReplaced;
  }

  /**
   * Creates a histogram and returns it in a ChartPanel.
   * @param oDataset The dataset from which to create a histogram.
   * @param sXAxisLabel The X axis label
   * @param sYAxisLabel The Y axis label
   * @param oLegend The legend for this chart.
   * @param oRequest The HistogramDataRequest object for this chart.
   * @param jPreferredSize Chart's size.
   * @return The chart in a ChartPanel.
   * @throws ModelException if something goes wrong with the chart.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   * <br>December 6, 2004:  Added total bar (LEM)
   */
  private static ChartPanel MakeHistogram(ModelHistogramDataset oDataset,
                                         String sXAxisLabel,
                                         String sYAxisLabel, Legend oLegend,
                                         HistogramDataRequest oRequest,
                                         Dimension jPreferredSize) throws
      ModelException {

    JFreeChart oChart = ChartFactory.createHistogram("", sXAxisLabel,
        sYAxisLabel, oDataset, PlotOrientation.VERTICAL, false, false, false);
    oChart.setBackgroundPaint(Color.WHITE);

    XYPlot oPlot = (XYPlot) oChart.getPlot();

    //Set a logarithmic axis, if desired
    if (oRequest.GetUseLogarithmicAxis()) {
      LogarithmicAxis oAxis = new LogarithmicAxis(sYAxisLabel);
      oAxis.setAllowNegativesFlag(true);
      oPlot.setRangeAxis(oAxis);
    }

    //Set the X axis tick marks to match the bin size
    NumberAxis oXAxis = (NumberAxis) oPlot.getDomainAxis();
    oXAxis.setTickUnit(new NumberTickUnit(oRequest.GetBinSize()));

    //Set the appropriate color for each species
    XYBarRenderer oRenderer = (XYBarRenderer) oPlot.getRenderer();
    for (int i = 0; i < oDataset.getSeriesCount(); i++) {
      String sName = oDataset.getSeriesName(i);
      if (sName.equalsIgnoreCase("total")) {
        oRenderer.setSeriesPaint(i, Color.black);
      }
      else {
        oRenderer.setSeriesPaint(i, oLegend.GetSpeciesColor(sName));
      }
    }

    ChartPanel oChartPanel = new ChartPanel(oChart);
    oChartPanel.setPreferredSize(jPreferredSize);
    oChartPanel.setMouseZoomable(true, false);
    return oChartPanel;
  }

  /**
   * Creates a histogram and returns it in a ChartPanel.
   * @param oDataset The dataset from which to create a histogram.
   * @param sXAxisLabel The X axis label
   * @param sYAxisLabel The Y axis label
   * @param oLegend The legend for this chart.
   * @param oRequest The HistogramDataRequest object for this chart.
   * @param jPreferredSize Chart's size.
   * @return The chart in a ChartPanel.
   * @throws ModelException if something goes wrong with the chart.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   * <br>December 6, 2004:  Added total bar (LEM)
   */
  private static ChartPanel MakeHistogram(ModelHistogramDataset oDataset,
                                         String sXAxisLabel,
                                         String sYAxisLabel, Legend oLegend,
                                         GridHistogramDataRequest oRequest,
                                         Dimension jPreferredSize) throws
      ModelException {

    JFreeChart oChart = ChartFactory.createHistogram("", sXAxisLabel,
        sYAxisLabel, oDataset, PlotOrientation.VERTICAL, false, false, false);
    oChart.setBackgroundPaint(Color.WHITE);

    XYPlot oPlot = (XYPlot) oChart.getPlot();

    //Set a logarithmic axis, if desired
    if (oRequest.GetUseLogarithmicAxis()) {
      LogarithmicAxis oAxis = new LogarithmicAxis(sYAxisLabel);
      oAxis.setAllowNegativesFlag(true);
      oPlot.setRangeAxis(oAxis);
    }

    //Set the X axis tick marks to match the bin size
    NumberAxis oXAxis = (NumberAxis) oPlot.getDomainAxis();
    oXAxis.setTickUnit(new NumberTickUnit(oRequest.GetBinSize()));

    //Set the appropriate color - black
    XYBarRenderer oRenderer = (XYBarRenderer) oPlot.getRenderer();
    oRenderer.setSeriesPaint(0, Color.black);

    ChartPanel oChartPanel = new ChartPanel(oChart);
    oChartPanel.setPreferredSize(jPreferredSize);
    oChartPanel.setMouseZoomable(true, false);
    return oChartPanel;
  }
}
