package datavisualizer;

import java.io.FileReader;
import java.util.Vector;
import javax.swing.*;
import javax.swing.JInternalFrame;
import org.jfree.data.DefaultCategoryDataset;
import javawrapper.ModelFileFunctions;
import javawrapper.ModelException;
import javawrapper.ErrorGUI;

/**
 * Manages all the data visualization for a single short output file.
 * <p>Copyright: Copyright (c) Charles D. Canham 2003</p>
 * <p>Company: Institute of Ecosystem Studies</p>
 * @author Lora E. Murphy
 * @version 1.0
 *
 * <br>Edit history:
 * <br>------------------
 * <br>April 28, 2004: Submitted in beta version (LEM)
 * <br>July 15, 2004:  Added subplot support (LEM)
 * <br>July 27, 2004:  Added table (LEM)
 */

public class ShortOutputFileManager
    extends DataFileManager {
  private static final int SEEDLING = 0; /**<Seedling*/
  private static final int SAPLING = 1; /**<Sapling*/
  private static final int ADULT = 2; /**<Adult*/
  private static final int SNAG = 3; /**<Snag*/

  private final static int RELATIVE_BASAL_AREA = 4; /**<Relative basal area chart type*/
  private final static int RELATIVE_DENSITY = 5; /**<Relative density chart type*/
  private final static int ABSOLUTE_BASAL_AREA = 6; /**<Absolute basal area chart type*/
  private final static int ABSOLUTE_DENSITY = 7; /**<Absolute density chart type*/

  private Vector mp_oDatasets = new Vector(0); /**<Datasets for each chart*/

  /**Relative basal area data.  Array indexes are #1 - subplot, #2 - type, #3 - species, and #4 - timestep.*/
  private Float[][][][] mp_fRBA;

  /**Absolute basal area data.  Array indexes are #1 - subplot, #2 - type, #3 - species, and #4 - timestep.*/
  private Float[][][][] mp_fABA;

  /**Relative density data.  Array indexes are #1 - subplot, #2 - type, #3 - species, and #4 - timestep.*/
  private Float[][][][] mp_fRDN;

  /**Absolute density data.  Array indexes are #1 - subplot, #2 - type, #3 - species, and #4 - timestep.*/
  private Float[][][][] mp_fADN;

  /**Absolute basal area totals.  Array indexes are #1 - subplot, #2 - type, and #3 - timestep.*/
  private Float[][][] mp_fABT;

  /**Absolute density totals.  Array indexes are #1 - subplot, #2 - type, and #3 - timestep.*/
  private Float[][][] mp_fADT;

  /**List of species names.*/
  private String[] mp_sSpeciesNames;

  /**List of subplot names.*/
  private String[] mp_sSubplotNames = null;

  /**Display string of filename - should have the last 20 chars.*/
  private String m_sFileDisplayName;

  /**Number of tree types that this object is willing to work with.*/
  private int m_iNumTypes = 4;

  /**Number of subplots in this file.*/
  private int m_iNumSubplots = 1;

  /**Number of species in the file.*/
  private int m_iNumSpecies = 0;

  /**Number of timesteps in the file.*/
  private int m_iNumTimesteps = 0;

  /**
   * Draws a requested chart.
   * @param sGraphName Name of the chart to draw.
   * @return Chart, or NULL if it could not be drawn.
   * @throws ModelException if there's a problem drawing the chart.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   * <br>June 25, 2004:  Added snag support (LEM)
   * <br>July 15, 2004:  Added subplot support (LEM)
   * <br>July 27, 2004:  Added the table (LEM)
   */
  protected JInternalFrame DrawChart(String sGraphName) throws ModelException {
    DefaultCategoryDataset oDataset = null;
    String sXAxisLabel = "Timesteps", sYAxisLabel = "",
        sSubplotName = sGraphName.substring(0, sGraphName.indexOf(":")),
        sGraph = sGraphName.substring(sGraphName.indexOf(":") + 2);
    int iSubplotIndex = 0, i, iTreeType, iChartType;

    for (i = 0; i < mp_sSubplotNames.length; i++) {
      if (mp_sSubplotNames[i].equals(sSubplotName)) {
        iSubplotIndex = i;
      }
    }

    sGraphName += " - " + m_sFileDisplayName;

    //Check to see if it's the table
    if (sGraph.equalsIgnoreCase("table")) {
      return CreateTable(0, iSubplotIndex, m_oLegend, sGraphName);
    }

    if (sGraph.equalsIgnoreCase("adult absolute basal area")) {
      oDataset = CreateDataset(mp_fABA[iSubplotIndex][ADULT],
                               mp_fABT[iSubplotIndex][ADULT]);
      sYAxisLabel = "sq. m/ha";
      iTreeType = ADULT;
      iChartType = ABSOLUTE_BASAL_AREA;
    }
    else if (sGraph.equalsIgnoreCase("adult relative basal area")) {
      oDataset = CreateDataset(mp_fRBA[iSubplotIndex][ADULT], null);
      iTreeType = ADULT;
      iChartType = RELATIVE_BASAL_AREA;
      sYAxisLabel = "Proportion";
    }
    else if (sGraph.equalsIgnoreCase("adult absolute density")) {
      oDataset = CreateDataset(mp_fADN[iSubplotIndex][ADULT],
                               mp_fADT[iSubplotIndex][ADULT]);
      iTreeType = ADULT;
      iChartType = ABSOLUTE_DENSITY;
      sYAxisLabel = "#/ha";
    }
    else if (sGraph.equalsIgnoreCase("adult relative density")) {
      oDataset = CreateDataset(mp_fRDN[iSubplotIndex][ADULT], null);
      iTreeType = ADULT;
      iChartType = RELATIVE_DENSITY;
      sYAxisLabel = "Proportion";
    }
    else if (sGraph.equalsIgnoreCase("snag absolute basal area")) {
      oDataset = CreateDataset(mp_fABA[iSubplotIndex][SNAG],
                               mp_fABT[iSubplotIndex][SNAG]);
      iTreeType = SNAG;
      iChartType = ABSOLUTE_BASAL_AREA;
      sYAxisLabel = "sq. m/ha";
    }
    else if (sGraph.equalsIgnoreCase("snag relative basal area")) {
      oDataset = CreateDataset(mp_fRBA[iSubplotIndex][SNAG], null);
      iTreeType = SNAG;
      iChartType = RELATIVE_BASAL_AREA;
      sYAxisLabel = "Proportion";
    }
    else if (sGraph.equalsIgnoreCase("snag absolute density")) {
      oDataset = CreateDataset(mp_fADN[iSubplotIndex][SNAG],
                               mp_fADT[iSubplotIndex][SNAG]);
      iTreeType = SNAG;
      iChartType = ABSOLUTE_DENSITY;
      sYAxisLabel = "#/ha";
    }
    else if (sGraph.equalsIgnoreCase("snag relative density")) {
      oDataset = CreateDataset(mp_fRDN[iSubplotIndex][SNAG], null);
      iTreeType = SNAG;
      iChartType = RELATIVE_DENSITY;
      sYAxisLabel = "Proportion";
    }
    else if (sGraph.equalsIgnoreCase("sapling absolute basal area")) {
      oDataset = CreateDataset(mp_fABA[iSubplotIndex][SAPLING],
                               mp_fABT[iSubplotIndex][SAPLING]);
      iTreeType = SAPLING;
      iChartType = ABSOLUTE_BASAL_AREA;
      sYAxisLabel = "sq. m/ha";
    }
    else if (sGraph.equalsIgnoreCase("sapling relative basal area")) {
      oDataset = CreateDataset(mp_fRBA[iSubplotIndex][SAPLING], null);
      iTreeType = SAPLING;
      iChartType = RELATIVE_BASAL_AREA;
      sYAxisLabel = "Proportion";
    }
    else if (sGraph.equalsIgnoreCase("sapling absolute density")) {
      oDataset = CreateDataset(mp_fADN[iSubplotIndex][SAPLING],
                               mp_fADT[iSubplotIndex][SAPLING]);
      iTreeType = SAPLING;
      iChartType = ABSOLUTE_DENSITY;
      sYAxisLabel = "#/ha";
    }
    else if (sGraph.equalsIgnoreCase("sapling relative density")) {
      oDataset = CreateDataset(mp_fRDN[iSubplotIndex][SAPLING], null);
      iTreeType = SAPLING;
      iChartType = RELATIVE_DENSITY;
      sYAxisLabel = "Proportion";
    }
    else if (sGraph.equalsIgnoreCase("seedling absolute density")) {
      oDataset = CreateDataset(mp_fADN[iSubplotIndex][SEEDLING],
                               mp_fADT[iSubplotIndex][SEEDLING]);
      iTreeType = SEEDLING;
      iChartType = ABSOLUTE_DENSITY;
      sYAxisLabel = "#/ha";
    }
    else if (sGraph.equalsIgnoreCase("seedling relative density")) {
      oDataset = CreateDataset(mp_fRDN[iSubplotIndex][SEEDLING], null);
      iTreeType = SEEDLING;
      iChartType = RELATIVE_DENSITY;
      sYAxisLabel = "Proportion";
    }
    else {
      return null;
    }

    mp_oDatasets.add(new ModelDataset(oDataset, sGraphName));

    //Adjust the dataset to make sure only desired species are visible
    DefaultCategoryDataset oAdjustedDataset = AdjustVisibleSpecies(oDataset,
        m_oLegend);

    LineGraphFileWriter oGraphWriter = new LineGraphFileWriter(iTreeType,
        iChartType, iSubplotIndex);
    JInternalFrame jFrame = DataGrapher.DrawLineChart(oAdjustedDataset,
        oGraphWriter, sXAxisLabel,
        sYAxisLabel,
        sGraphName,
        m_oLegend, mp_sSpeciesNames.length);
    oGraphWriter.SetFrame(jFrame);
    return jFrame;

  }

  /**
   * Creates the datasets for the overview table.  This creates four table
   * datasets, one each for seedlings, saplings, adults, and snags.  All
   * numbers are placed in as formatted strings with a max of 2 decimal
   * places.
   * @param iTimestep Timestep for which to draw the table.
   * @param iSubplot Index of subplot, or 0 if it's the whole plot.
   * @return Array of datasets for tables, ready to pass to DataGrapher.
   * Index 0 is the seedling dataset, 1 is the sapling dataset, 2 is the
   * adult dataset, and 3 is the snag dataset.
   */
  private Object[][][] CreateTableDataset(int iTimestep, int iSubplot) {
    Object[][][] p_oAllDatasets = new Object[4][][];

    java.text.NumberFormat oFormat = java.text.NumberFormat.getInstance();
    oFormat.setMaximumFractionDigits(3);
    oFormat.setMinimumFractionDigits(3);

    java.text.NumberFormat oRelFormat = java.text.NumberFormat.getPercentInstance();
    oRelFormat.setMaximumFractionDigits(2);
    oRelFormat.setMinimumFractionDigits(2);

    int i;

    //*****************************
     // Adults table dataset
     //*****************************
      //Create our table data array
      p_oAllDatasets[2] = new Object[m_iNumSpecies + 1][4]; //rows by columns

      //Create a row for each species
    for (i = 0; i < m_iNumSpecies; i++) {

      //Column 1: absolute density
      if (mp_fADN != null && mp_fADN[iSubplot][ADULT][i] != null) {
        p_oAllDatasets[2][i][0] = oFormat.format(mp_fADN[iSubplot][ADULT][i]
                                                 [iTimestep].doubleValue());
      }
      else {
        p_oAllDatasets[2][i][0] = null;
      }

      //Column 2: percent of density
      if (mp_fRDN != null && mp_fRDN[iSubplot][ADULT][i] != null) {
        p_oAllDatasets[2][i][1] = oRelFormat.format(mp_fRDN[iSubplot][ADULT][i][
                                                 iTimestep].doubleValue());
      }
      else {
        p_oAllDatasets[2][i][1] = null;
      }

      //Column 3: absolute basal area
      if (mp_fABA != null && mp_fABA[iSubplot][ADULT][i] != null) {
        p_oAllDatasets[2][i][2] = oFormat.format(mp_fABA[iSubplot][ADULT][i][
                                                 iTimestep].doubleValue());
      }
      else {
        p_oAllDatasets[2][i][2] = null;
      }

      //Column 4: percent of basal area
      if (mp_fRBA != null && mp_fRBA[iSubplot][ADULT][i] != null) {
        p_oAllDatasets[2][i][3] = oRelFormat.format(mp_fRBA[iSubplot][ADULT][i][
                                                 iTimestep].doubleValue());
      }
      else {
        p_oAllDatasets[2][i][3] = null;
      }
    }

    //Row for totals
    //Column 1: absolute density
    if (mp_fADT != null && mp_fADT[iSubplot][ADULT] != null) {
      p_oAllDatasets[2][m_iNumSpecies][0] = oFormat.format(mp_fADT[iSubplot][
          ADULT][iTimestep].doubleValue());
    }
    else {
      p_oAllDatasets[2][m_iNumSpecies][0] = null;
    }

    //Column 3: absolute basal area
    if (mp_fABT != null && mp_fABT[iSubplot][ADULT] != null) {
      p_oAllDatasets[2][m_iNumSpecies][2] = oFormat.format(mp_fABT[iSubplot][
          ADULT][iTimestep].doubleValue());
    }
    else {
      p_oAllDatasets[2][m_iNumSpecies][2] = null;
    }

    //*****************************
     // Saplings table dataset
     //*****************************
      //Create our table data array
      p_oAllDatasets[1] = new Object[m_iNumSpecies + 1][4]; //rows by columns

      //Create a row for each species
    for (i = 0; i < m_iNumSpecies; i++) {

      //Column 1: absolute density
      if (mp_fADN != null && mp_fADN[iSubplot][SAPLING][i] != null) {
        p_oAllDatasets[1][i][0] = oFormat.format(mp_fADN[iSubplot][SAPLING][i][
                                                 iTimestep].doubleValue());
      }
      else {
        p_oAllDatasets[1][i][0] = null;
      }

      //Column 2: percent of density
      if (mp_fRDN != null && mp_fRDN[iSubplot][SAPLING][i] != null) {
        p_oAllDatasets[1][i][1] = oRelFormat.format(mp_fRDN[iSubplot][SAPLING][i][
                                                 iTimestep].doubleValue());
      }
      else {
        p_oAllDatasets[1][i][1] = null;
      }

      //Column 3: absolute basal area
      if (mp_fABA != null && mp_fABA[iSubplot][SAPLING][i] != null) {
        p_oAllDatasets[1][i][2] = oFormat.format(mp_fABA[iSubplot][SAPLING][i][
                                                 iTimestep].doubleValue());
      }
      else {
        p_oAllDatasets[1][i][2] = null;
      }

      //Column 4: percent of basal area
      if (mp_fRBA != null && mp_fRBA[iSubplot][SAPLING][i] != null) {
        p_oAllDatasets[1][i][3] = oRelFormat.format(mp_fRBA[iSubplot][SAPLING][i][
                                                 iTimestep].doubleValue());
      }
      else {
        p_oAllDatasets[1][i][3] = null;
      }
    }

    //Row for totals
    //Column 1: absolute density
    if (mp_fADT != null && mp_fADT[iSubplot][SAPLING] != null) {
      p_oAllDatasets[1][m_iNumSpecies][0] = oFormat.format(mp_fADT[iSubplot][
          SAPLING][iTimestep].doubleValue());
    }
    else {
      p_oAllDatasets[1][m_iNumSpecies][0] = null;
    }

    //Column 3: absolute basal area
    if (mp_fABT != null && mp_fABT[iSubplot][SAPLING] != null) {
      p_oAllDatasets[1][m_iNumSpecies][2] = oFormat.format(mp_fABT[iSubplot][
          SAPLING][iTimestep].doubleValue());
    }
    else {
      p_oAllDatasets[1][m_iNumSpecies][2] = null;
    }

    //*****************************
     // Snags table dataset
     //*****************************
      //Create our table data array
      p_oAllDatasets[3] = new Object[m_iNumSpecies + 1][4]; //rows by columns

      //Create a row for each species
    for (i = 0; i < m_iNumSpecies; i++) {

      //Column 1: absolute density
      if (mp_fADN != null && mp_fADN[iSubplot][SNAG][i] != null) {
        p_oAllDatasets[3][i][0] = oFormat.format(mp_fADN[iSubplot][SNAG][i][
                                                 iTimestep].doubleValue());
      }
      else {
        p_oAllDatasets[3][i][0] = null;
      }

      //Column 2: percent of density
      if (mp_fRDN != null && mp_fRDN[iSubplot][SNAG][i] != null) {
        p_oAllDatasets[3][i][1] = oRelFormat.format(mp_fRDN[iSubplot][SNAG][i][
                                                 iTimestep].doubleValue());
      }
      else {
        p_oAllDatasets[3][i][1] = null;
      }

      //Column 3: absolute basal area
      if (mp_fABA != null && mp_fABA[iSubplot][SNAG][i] != null) {
        p_oAllDatasets[3][i][2] = oFormat.format(mp_fABA[iSubplot][SNAG][i][
                                                 iTimestep].doubleValue());
      }
      else {
        p_oAllDatasets[3][i][2] = null;
      }

      //Column 4: percent of basal area
      if (mp_fRBA != null && mp_fRBA[iSubplot][SNAG][i] != null) {
        p_oAllDatasets[3][i][3] = oRelFormat.format(mp_fRBA[iSubplot][SNAG][i][
                                                 iTimestep].doubleValue());
      }
      else {
        p_oAllDatasets[3][i][3] = null;
      }
    }

    //Row for totals
    //Column 1: absolute density
    if (mp_fADT != null && mp_fADT[iSubplot][SNAG] != null) {
      p_oAllDatasets[3][m_iNumSpecies][0] = oFormat.format(mp_fADT[iSubplot][
          SNAG][iTimestep].doubleValue());
    }
    else {
      p_oAllDatasets[3][m_iNumSpecies][0] = null;
    }

    //Column 3: absolute basal area
    if (mp_fABT != null && mp_fABT[iSubplot][SNAG] != null) {
      p_oAllDatasets[3][m_iNumSpecies][2] = oFormat.format(mp_fABT[iSubplot][
          SNAG][iTimestep].doubleValue());
    }
    else {
      p_oAllDatasets[3][m_iNumSpecies][2] = null;
    }

    //*****************************
     // Seedlings table dataset
     //*****************************
      //Create our table data array
      p_oAllDatasets[0] = new Object[m_iNumSpecies + 1][1]; //rows by columns

      //Create a row for each species
    for (i = 0; i < m_iNumSpecies; i++) {

      //Column 1: absolute density
      if (mp_fADN != null && mp_fADN[iSubplot][SEEDLING][i] != null) {
        p_oAllDatasets[0][i][0] = oFormat.format(mp_fADN[iSubplot][SEEDLING][i][
                                                 iTimestep].doubleValue());
      }
      else {
        p_oAllDatasets[0][i][0] = null;
      }
    }

    //Row for totals
    //Column 1: absolute density
    if (mp_fADT != null && mp_fADT[iSubplot][SEEDLING] != null) {
      p_oAllDatasets[0][m_iNumSpecies][0] = oFormat.format(mp_fADT[iSubplot][
          SEEDLING][iTimestep].doubleValue());
    }
    else {
      p_oAllDatasets[0][m_iNumSpecies][0] = null;
    }

    return p_oAllDatasets;
  }

  /**
   * Creates the table.  There are four sub-tables, one each for seedlings,
   * saplings, adults, and snags.  The tables are arranged in two rows,
   * with adults and saplings in the top row and snags and seedlings in the
   * bottom row.
   *
   * A pair of buttons is added to advance the table forward and back.
   * @param oLegend Legend for this table.
   * @param iTimestep Timestep for which to draw the table.
   * @param iSubplot Index of subplot, or 0 if it's the whole plot.
   * @param sChartTitle Title to display in the chart window.
   * @return Panel containing the table.
   * @throws ModelException If anything goes wrong with the drawing.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>July 27, 2004: Created (LEM)
   */
  private JInternalFrame CreateTable(int iTimestep, int iSubplot,
                                     Legend oLegend, String sChartTitle) throws
      ModelException {

    //Get the datasets
    Object[][][] p_oAllDatasets = CreateTableDataset(iTimestep, iSubplot);

    //Pass the datasets to create the table
    TableFileWriter oFileWriter = new TableFileWriter();
    JInternalFrame jFrame = DataGrapher.DrawOverviewTables(sChartTitle,
        p_oAllDatasets[0],
        p_oAllDatasets[1], p_oAllDatasets[2], p_oAllDatasets[3], m_oLegend,
        oFileWriter);
    oFileWriter.SetFrame(jFrame);

    //Create a panel with buttons for scrolling through timesteps
    JPanel jButtonsPanel = new JPanel(new java.awt.FlowLayout(java.awt.
        FlowLayout.LEFT));
    JButton jBackButton = new JButton(new javawrapper.ModelIcon(15, 15,
        javawrapper.ModelIcon.LEFT_TRIANGLE));
    jBackButton.setToolTipText("Show previous timestep");
    jBackButton.setActionCommand("StepBack");
    jBackButton.setName("StepBack");
    jBackButton.setFont(new javawrapper.ModelFont());
    jButtonsPanel.add(jBackButton);

    JLabel jTempLabel = new JLabel("Showing timestep ");
    jTempLabel.setFont(new javawrapper.ModelFont());
    jButtonsPanel.add(jTempLabel);

    jTempLabel = new JLabel(String.valueOf(iTimestep));
    jTempLabel.setFont(new javawrapper.ModelFont());
    jTempLabel.setName("current_timestep");
    jButtonsPanel.add(jTempLabel);

    jTempLabel = new JLabel(" of ");
    jTempLabel.setFont(new javawrapper.ModelFont());
    jButtonsPanel.add(jTempLabel);

    jTempLabel = new JLabel(String.valueOf(m_iNumTimesteps));
    jTempLabel.setFont(new javawrapper.ModelFont());
    jTempLabel.setName("number_timesteps");
    jButtonsPanel.add(jTempLabel);

    JButton jForwardButton = new JButton(new javawrapper.ModelIcon(15, 15,
        javawrapper.ModelIcon.RIGHT_TRIANGLE));
    jForwardButton.setActionCommand("StepForward");
    jForwardButton.setName("StepForward");
    jForwardButton.setToolTipText("Show next timestep");
    jForwardButton.setFont(new javawrapper.ModelFont());
    jButtonsPanel.add(jForwardButton);

    JPanel jWholePanel = new JPanel(new java.awt.BorderLayout());
    jWholePanel.add(jFrame.getContentPane(), java.awt.BorderLayout.CENTER);
    jWholePanel.add(jButtonsPanel, java.awt.BorderLayout.NORTH);

    TableUpdater oUpdater = new TableUpdater(jWholePanel, this, iSubplot);
    jBackButton.addActionListener(oUpdater);
    jForwardButton.addActionListener(oUpdater);

    jFrame.setContentPane(jWholePanel);

    return jFrame;
  }

  /**
   * Updates a table window with new data.
   * @param jTableWindowPanel The content pane of the window containing
   * the table to update.
   * @param iTimestep The timestep of the data to display.
   * @param iSubplot The subplot of the data to display.
   */
  protected void UpdateTable(JPanel jTableWindowPanel, int iTimestep,
                             int iSubplot) {
    //Create the new dataset
    Object[][][] p_oAllDatasets = CreateTableDataset(iTimestep, iSubplot);
    //Update the table
    DataGrapher.UpdateOverviewTables(jTableWindowPanel, p_oAllDatasets[0],
                                     p_oAllDatasets[1], p_oAllDatasets[2],
                                     p_oAllDatasets[3]);

    //Update the timestep label
    JLabel jLabel = (JLabel) DataGrapher.FindNamedComponent(jTableWindowPanel,
        "current_timestep");
    if (jLabel == null) {
      return;
    }
    jLabel.setText(String.valueOf(iTimestep));
  }

  /**
   * Updates charts for a current run.  This reparses and redraws the charts
   * from scratch, since they show the cumulative history of a run.
   * @throws ModelException if there is a problem reading the file or
   * drawing the charts.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>November 18, 2004: Created (LEM)
   */
  public void UpdateCurrentRunCharts() throws ModelException {

    if (mp_oCharts.size() == 0) {
      return;
    }

    //Parse the output file again to refresh the dataset
    ParseFile();

    //Redraw all the charts
    JInternalFrame jFrame;
    ModelDataset oModelDataset = null;
    String sKey;
    int i, j, k;
    boolean bFound;

    //Loop through our current set of windows
    for (i = 0; i < mp_oCharts.size(); i++) {

      //Get the window title - that's the key to finding this chart's dataset
      jFrame = (JInternalFrame) mp_oCharts.elementAt(i);
      sKey = jFrame.getTitle();

      //Do line charts and tables separately
      if (sKey.indexOf("Table") == -1) {

        //Now find the corresponding dataset
        bFound = false;
        for (j = 0; j < mp_oDatasets.size(); j++) {
          oModelDataset = (ModelDataset) mp_oDatasets.elementAt(j);
          if (sKey.equalsIgnoreCase(oModelDataset.GetName())) {
            bFound = true;

            DefaultCategoryDataset oNewDataset = null;

            //Create a new dataset
            String sGraph = sKey.substring(sKey.indexOf(":") + 2).toLowerCase(),
                sSubplotName = sKey.substring(0, sKey.indexOf(":"));
            int iSubplotIndex = 0;

            for (k = 0; k < mp_sSubplotNames.length; k++) {
              if (mp_sSubplotNames[k].equals(sSubplotName)) {
                iSubplotIndex = k;
              }
            }

            if (sGraph.indexOf("adult absolute basal area") > -1) {
              oNewDataset = CreateDataset(mp_fABA[iSubplotIndex][ADULT],
                                          mp_fABT[iSubplotIndex][ADULT]);
            }
            else if (sGraph.indexOf("adult relative basal area") > -1) {
              oNewDataset = CreateDataset(mp_fRBA[iSubplotIndex][ADULT], null);
            }
            else if (sGraph.indexOf("adult absolute density") > -1) {
              oNewDataset = CreateDataset(mp_fADN[iSubplotIndex][ADULT],
                                          mp_fADT[iSubplotIndex][ADULT]);
            }
            else if (sGraph.indexOf("adult relative density") > -1) {
              oNewDataset = CreateDataset(mp_fRDN[iSubplotIndex][ADULT], null);
            }
            else if (sGraph.indexOf("snag absolute basal area") > -1) {
              oNewDataset = CreateDataset(mp_fABA[iSubplotIndex][SNAG],
                                          mp_fABT[iSubplotIndex][SNAG]);
            }
            else if (sGraph.indexOf("snag relative basal area") > -1) {
              oNewDataset = CreateDataset(mp_fRBA[iSubplotIndex][SNAG], null);
            }
            else if (sGraph.indexOf("snag absolute density") > -1) {
              oNewDataset = CreateDataset(mp_fADN[iSubplotIndex][SNAG],
                                          mp_fADT[iSubplotIndex][SNAG]);
            }
            else if (sGraph.indexOf("snag relative density") > -1) {
              oNewDataset = CreateDataset(mp_fRDN[iSubplotIndex][SNAG], null);
            }
            else if (sGraph.indexOf("sapling absolute basal area") > -1) {
              oNewDataset = CreateDataset(mp_fABA[iSubplotIndex][SAPLING],
                                          mp_fABT[iSubplotIndex][SAPLING]);
            }
            else if (sGraph.indexOf("sapling relative basal area") > -1) {
              oNewDataset = CreateDataset(mp_fRBA[iSubplotIndex][SAPLING], null);
            }
            else if (sGraph.indexOf("sapling absolute density") > -1) {
              oNewDataset = CreateDataset(mp_fADN[iSubplotIndex][SAPLING],
                                          mp_fADT[iSubplotIndex][SAPLING]);
            }
            else if (sGraph.indexOf("sapling relative density") > -1) {
              oNewDataset = CreateDataset(mp_fRDN[iSubplotIndex][SAPLING], null);
            }
            else if (sGraph.indexOf("seedling absolute density") > -1) {
              oNewDataset = CreateDataset(mp_fADN[iSubplotIndex][SEEDLING],
                                          mp_fADT[iSubplotIndex][SEEDLING]);
            }
            else if (sGraph.indexOf("seedling relative density") > -1) {
              oNewDataset = CreateDataset(mp_fRDN[iSubplotIndex][SEEDLING], null);
            }
            else {
              break;
            }

            //Set the new dataset back in the list
            ModelDataset oReplacementDataset = new ModelDataset(oNewDataset,
                oModelDataset.GetName());
            mp_oDatasets.remove(oModelDataset);
            mp_oDatasets.add(oReplacementDataset);

            //Adjust the dataset to make sure only desired species are visible
            DefaultCategoryDataset oAdjustedDataset = AdjustVisibleSpecies(
                oNewDataset, m_oLegend);

            DataGrapher.UpdateLineChart(oAdjustedDataset, jFrame, m_oLegend,
                                        mp_sSpeciesNames.length);

          }
        }
        if (!bFound) {
          //Panic
          ModelException oErr = new ModelException();
          oErr.iErrorCode = ErrorGUI.CANT_FIND_OBJECT;
          oErr.sCallingFunction = "JAVA";
          oErr.sMessage = "Can't find dataset for " + sKey;
          throw (oErr);
        }
      }
      else {
        //This is the table - update it

        //Get the number of timesteps written so far to the file - do it by
        //finding a non-null bunch of data
        Float[][][][][] p_fDatasets = new Float[4][][][][];
        p_fDatasets[0] = mp_fABA;
        p_fDatasets[1] = mp_fRBA;
        p_fDatasets[2] = mp_fADN;
        p_fDatasets[3] = mp_fRDN;
        int i1, i2, iNumTimesteps = 0;
        for (i1 = 0; i1 < p_fDatasets.length; i1++) {
          if (p_fDatasets[i1] != null) {
            for (i2 = 0; i2 < p_fDatasets[i1][0].length; i2++) {
              if (p_fDatasets[i1][0][i2][0] != null) {
                iNumTimesteps = p_fDatasets[i1][0][i2][0].length - 1;
                break;
              }
            }
          }
          if (iNumTimesteps > 0) {
            break;
          }
        }

        //Set the number of timesteps we counted
        JLabel jLabel = (JLabel) DataGrapher.FindNamedComponent(jFrame.
            getContentPane(),
            "number_timesteps");
        if (jLabel == null) {
          return;
        }
        jLabel.setText(String.valueOf(iNumTimesteps));

        //Get the table's subplot by finding a button listener
        JButton jButton = (JButton) DataGrapher.FindNamedComponent(jFrame.
            getContentPane(), "StepForward");
        Object[] p_oListeners = jButton.getActionListeners();
        boolean bFoundListener = false;
        for (int m = 0; m < p_oListeners.length; m++) {
          if (p_oListeners[m] instanceof TableUpdater) {
            int iSubplot = ( (TableUpdater) p_oListeners[m]).m_iSubplot;
            UpdateTable( (JPanel) jFrame.getContentPane(), iNumTimesteps,
                        iSubplot);
            bFoundListener = true;
            break;
          }
        }
        //Throw an error if we didn't find the table updater
        if (bFoundListener == false) {
          throw (new javawrapper.ModelException(ErrorGUI.CANT_FIND_OBJECT,
                                                "JAVA",
                                                "There was a problem updating the table.\nThe subplot was not found."));
        }
      }
    }
  }

  /**
   * Redraws all open chart objects, except tables.  Tables are not affected
   * by any incidence that affects the other charts, such as clicking species
   * on the legend.
   * @throws ModelException passing through from other methods.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void UpdateCharts() throws ModelException {
    JInternalFrame jFrame;
    ModelDataset oModelDataset = null;
    String sKey;
    int i, j;
    boolean bFound;

    //Loop through our current set of windows
    for (i = 0; i < mp_oCharts.size(); i++) {

      //Get the window title - that's the key to finding this chart's dataset
      jFrame = (JInternalFrame) mp_oCharts.elementAt(i);
      sKey = jFrame.getTitle();

      //If this is a table, ignore
      if (sKey.indexOf("Table") == -1) {

        //Now find the corresponding dataset
        bFound = false;
        for (j = 0; j < mp_oDatasets.size(); j++) {
          oModelDataset = (ModelDataset) mp_oDatasets.elementAt(j);
          if (sKey.equalsIgnoreCase(oModelDataset.GetName())) {
            bFound = true;

            //Adjust the dataset to make sure only desired species are visible
            DefaultCategoryDataset oAdjustedDataset = AdjustVisibleSpecies( (
                DefaultCategoryDataset) oModelDataset.GetDataset(),
                m_oLegend);

            DataGrapher.UpdateLineChart(oAdjustedDataset, jFrame, m_oLegend,
                                        mp_sSpeciesNames.length);

          }
        }
        if (!bFound) {
          //Panic
          ModelException oErr = new ModelException();
          oErr.iErrorCode = ErrorGUI.CANT_FIND_OBJECT;
          oErr.sCallingFunction = "JAVA";
          oErr.sMessage = "Can't find dataset for " + sKey;
          throw (oErr);
        }
      }
    }
  }

  /**
   * Gets table options.  A summary table can be created from this file type.
   * @return List of table options
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   * <br>July 27, 2004: Implemented (LEM)
   */
  public String[] GetTableOptions() {
    String[] p_sOptions = new String[m_iNumSubplots];
    for (int i = 0; i < m_iNumSubplots; i++) {
      p_sOptions[i] = mp_sSubplotNames[i] + ": Table";
    }
    return p_sOptions;
  }

  /**
   * No histogram options for this type of chart.
   * @return null
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public String[] GetHistogramOptions() {
    return null;
  }

  /**
   * No map options for this type of chart.
   * @return null.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public String[] GetMapOptions() {
    return null;
  }

  /**
   * Gets the line graph options.
   * @return List of line graphs.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)]
   * <br>June 25, 2004:  Added snags (LEM)
   * <br>July 15, 2004: Added subplots (LEM)
   */
  public String[] GetLineGraphOptions() {
    //Count the options we can do
    int iOptions = 0, i, j, k;
    if (mp_fRBA != null) {
      for (i = 0; i < m_iNumSubplots; i++) {
        for (j = 0; j < m_iNumTypes; j++) {
          if (mp_fRBA[i][j][0] != null && mp_fRBA[i][j][0].length > 0) {
            iOptions++;
          }
        }
      }
    }

    if (mp_fABA != null) {
      for (i = 0; i < m_iNumSubplots; i++) {
        for (j = 0; j < m_iNumTypes; j++) {
          if (mp_fABA[i][j][0] != null && mp_fABA[i][j][0].length > 0) {
            iOptions++;
          }
        }
      }
    }

    if (mp_fRDN != null) {
      for (i = 0; i < m_iNumSubplots; i++) {
        for (j = 0; j < m_iNumTypes; j++) {
          if (mp_fRDN[i][j][0] != null && mp_fRDN[i][j][0].length > 0) {
            iOptions++;
          }
        }
      }
    }

    if (mp_fADN != null) {
      for (i = 0; i < m_iNumSubplots; i++) {
        for (j = 0; j < m_iNumTypes; j++) {
          if (mp_fADN[i][j][0] != null && mp_fADN[i][j][0].length > 0) {
            iOptions++;
          }
        }
      }
    }

    if (iOptions == 0) {
      return null;
    }

    String[] sOptions = new String[iOptions];

    iOptions = 0;
    if (mp_fRBA != null) {
      for (i = 0; i < m_iNumSubplots; i++) {
        if (mp_fRBA[i][ADULT][0] != null &&
            mp_fRBA[i][ADULT][0].length > 0) {
          sOptions[iOptions] = mp_sSubplotNames[i] +
              ": Adult relative basal area";
          iOptions++;
        }
        if (mp_fRBA[i][SAPLING][0] != null &&
            mp_fRBA[i][SAPLING][0].length > 0) {
          sOptions[iOptions] = mp_sSubplotNames[i] +
              ": Sapling relative basal area";
          iOptions++;
        }
        if (mp_fRBA[i][SNAG][0] != null &&
            mp_fRBA[i][SNAG][0].length > 0) {
          sOptions[iOptions] = mp_sSubplotNames[i] +
              ": Snag relative basal area";
          iOptions++;
        }
      }
    }

    if (mp_fABA != null) {
      for (i = 0; i < m_iNumSubplots; i++) {
        if (mp_fABA[i][ADULT][0] != null &&
            mp_fABA[i][ADULT][0].length > 0) {
          sOptions[iOptions] = mp_sSubplotNames[i] +
              ": Adult absolute basal area";
          iOptions++;
        }
        if (mp_fABA[i][SAPLING][0] != null &&
            mp_fABA[i][SAPLING][0].length > 0) {
          sOptions[iOptions] = mp_sSubplotNames[i] +
              ": Sapling absolute basal area";
          iOptions++;
        }
        if (mp_fABA[i][SNAG][0] != null &&
            mp_fABA[i][SNAG][0].length > 0) {
          sOptions[iOptions] = mp_sSubplotNames[i] +
              ": Snag absolute basal area";
          iOptions++;
        }
      }
    }

    if (mp_fRDN != null) {
      for (i = 0; i < m_iNumSubplots; i++) {
        if (mp_fRDN[i][ADULT][0] != null &&
            mp_fRDN[i][ADULT][0].length > 0) {
          sOptions[iOptions] = mp_sSubplotNames[i] + ": Adult relative density";
          iOptions++;
        }
        if (mp_fRDN[i][SAPLING][0] != null &&
            mp_fRDN[i][SAPLING][0].length > 0) {
          sOptions[iOptions] = mp_sSubplotNames[i] +
              ": Sapling relative density";
          iOptions++;
        }
        if (mp_fRDN[i][SEEDLING][0] != null &&
            mp_fRDN[i][SEEDLING][0].length > 0) {
          sOptions[iOptions] = mp_sSubplotNames[i] +
              ": Seedling relative density";
          iOptions++;
        }
        if (mp_fRDN[i][SNAG][0] != null &&
            mp_fRDN[i][SNAG][0].length > 0) {
          sOptions[iOptions] = mp_sSubplotNames[i] + ": Snag relative density";
          iOptions++;
        }
      }
    }

    if (mp_fADN != null) {
      for (i = 0; i < m_iNumSubplots; i++) {
        if (mp_fADN[i][ADULT][0] != null &&
            mp_fADN[i][ADULT][0].length > 0) {
          sOptions[iOptions] = mp_sSubplotNames[i] + ": Adult absolute density";
          iOptions++;
        }
        if (mp_fADN[i][SAPLING][0] != null &&
            mp_fADN[i][SAPLING][0].length > 0) {
          sOptions[iOptions] = mp_sSubplotNames[i] +
              ": Sapling absolute density";
          iOptions++;
        }
        if (mp_fADN[i][SEEDLING][0] != null &&
            mp_fADN[i][SEEDLING][0].length > 0) {
          sOptions[iOptions] = mp_sSubplotNames[i] +
              ": Seedling absolute density";
          iOptions++;
        }
        if (mp_fADN[i][SNAG][0] != null &&
            mp_fADN[i][SNAG][0].length > 0) {
          sOptions[iOptions] = mp_sSubplotNames[i] + ": Snag absolute density";
          iOptions++;
        }
      }
    }

    return sOptions;

  }

  /**
   * Constructor.  Analyzes the output file for the data it contains.
   * @param sFileName  Filename to manage.
   * @throws ModelException If file cannot be read.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   * <br>June 25, 2004:  Added snag support (LEM)
   * <br>July 15, 2004:  Added subplot support (LEM)
   */
  public ShortOutputFileManager(String sFileName) throws ModelException {
    super(sFileName);

    //Create a display name
    if (sFileName.length() > 20) {
      //Take the LAST 20
      String sText = sFileName.substring(sFileName.length() - 20);
      m_sFileDisplayName = "..." + sText;
    }
    else {
      m_sFileDisplayName = sFileName;
    }

    ParseFile();

    //Create our legend
    String sFileTitle = new java.io.File(sFileName).getName();
    m_oLegend = new Legend(this, sFileTitle, mp_sSpeciesNames);
  }

  /**
   * Reads the file data into the internal arrays.
   * @throws ModelException if the file cannot be opened or read.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>June 29, 2005:  Put the relative values as percentages (LEM)
   */
  private void ParseFile() throws ModelException {
    FileReader in = null;
    try {

      in = new FileReader(m_sFilename);
      Vector p_sSpeciesNames = new Vector(0),
          p_oLine;
      //Data arrays - # subplots by # types by # species
      Vector[][][] p_fABA, p_fRBA, p_fADN, p_fRDN;
      //# subplots by # types
      Vector[][] p_fABT, p_fADT;
      String sLocation = new String(), //file location
          sData = new String(),
          sLabel = new String(), sOldLabel = new String();
      int iTestCh, //character
          iSubplotIndex, //index in the array of the subplot currently being read
          //These indicate which column each data piece starts in a row of values
          iAABAIndex = -1, //adult absolute basal area
          iARBAIndex = -1, //adult relative basal area
          iAADNIndex = -1, //adult absolute density
          iARDNIndex = -1, //adult relative density
          iJABAIndex = -1, //sapling absolute basal area
          iJRBAIndex = -1, //sapling relative basal area
          iJADNIndex = -1, //sapling absolute density
          iJRDNIndex = -1, //sapling relative density
          iSADNIndex = -1, //seedling absolute density
          iSRDNIndex = -1, //seedling relative density
          iSNABAIndex = -1, //snag absolute basal area
          iSNRBAIndex = -1, //snag relative basal area
          iSNADNIndex = -1, //snag absolute density
          iSNRDNIndex = -1, //snag relative density
          iLineSize, //expected number of tabbed cells per line
          iOffSet, //where the column headers start
          i, j, k, m;
      boolean bSubplotsExist = false;

      //Read in the location
      iTestCh = in.read();
      while (iTestCh != '\t' && iTestCh != '\n'
             && iTestCh != '\r') {
        sLocation = sLocation.concat(String.valueOf( (char) iTestCh));

        iTestCh = in.read();
      }
      //If there were extra tabs throw them away
      if (iTestCh != '\n') {
        while (iTestCh != '\n') {
          iTestCh = in.read();
        }
      }

      //Throw away the next line
      p_oLine = ModelFileFunctions.ReadLine(in);

      //And the next one after that
      p_oLine = ModelFileFunctions.ReadLine(in);

      //Now read in the next line
      p_oLine = ModelFileFunctions.ReadLine(in);

      //Read until the line says "subplots:" or "step"
      sData = (String) p_oLine.elementAt(0);
      sData = sData.toLowerCase();
      while (sData.indexOf("step") == -1 &&
             sData.indexOf("subplots") == -1) {
        p_oLine = ModelFileFunctions.ReadLine(in);
        sData = (String) p_oLine.elementAt(0);
        sData = sData.toLowerCase();
      }

      //Does the first bucket say "Subplots:" or "Step"?
      sData = (String) p_oLine.elementAt(0);
      if (sData.toLowerCase().indexOf("subplots") >= 0) {

        Vector p_oTempNames = new Vector(0);

        bSubplotsExist = true;
        iOffSet = 2;

        //Figure out how many subplots we have and capture their names -
        //we're done when we get to the line that starts "Step"
        while (true) {
          p_oLine = ModelFileFunctions.ReadLine(in);
          if (p_oLine.size() == 0) {
            ModelException oExp = new ModelException();
            oExp.iErrorCode = ErrorGUI.BAD_FILE;
            throw (oExp);
          }
          sData = (String) p_oLine.elementAt(0);
          if (sData.toLowerCase().startsWith("step")) {
            break;
          }

          //The subplot name is just after the " = "
          p_oTempNames.add(sData.substring(sData.indexOf("=") + 2));
        }

        m_iNumSubplots = p_oTempNames.size();
        mp_sSubplotNames = new String[m_iNumSubplots];
        for (i = 0; i < m_iNumSubplots; i++) {
          mp_sSubplotNames[i] = (String) p_oTempNames.elementAt(i);
        }
      }
      else {
        iOffSet = 1;
        mp_sSubplotNames = new String[m_iNumSubplots];
        mp_sSubplotNames[0] = "Whole plot";
      }

      //******************************
       //Now we're at the header row - parse out the header columns
       //******************************

        iLineSize = p_oLine.size(); //all data lines must match the number of
        //columns in the header row

        //First - collect all the species names - go through one type of data
        //Find out the data type of the first column
      sData = (String) p_oLine.elementAt(iOffSet);
      if (sData.indexOf(":") > -1) {
        sLabel = sData.substring(0, sData.indexOf(":"));
      }
      else if (sData.length() >= 4) {
        sLabel = sData.substring(0, 4);
      }
      else {
        sLabel = sData;
      }
      sOldLabel = sLabel;
      i = iOffSet + 1;
      while (sLabel.equals(sOldLabel) && i < iLineSize) {

        //Get species name part of the column heading
        if (sData.indexOf(":") > -1) {
          sData = sData.substring(sData.indexOf(":") + 2);
        }
        else {
          sData = sData.substring(5).trim();
        }
        p_sSpeciesNames.add(sData);

        //Get the data label part of the column heading - so we know when
        //we've reached a new data type
        sData = (String) p_oLine.elementAt(i);
        if (sData.indexOf(":") > -1) {
          sLabel = sData.substring(0, sData.indexOf(":"));
        }
        else if (sData.length() >= 4) {
          sLabel = sData.substring(0, 4);
        }
        else {
          sLabel = sData;
        }
        i++;
      }
      m_iNumSpecies = p_sSpeciesNames.size();

      //Now determine all the different data types that are in this file
      sOldLabel = "";
      for (i = iOffSet; i < p_oLine.size(); i++) {

        //Read each column
        sData = (String) p_oLine.elementAt(i);

        //What is this data type?
        if (sData.indexOf(":") > -1) {
          sLabel = sData.substring(0, sData.indexOf(":"));
        }
        else if (sData.length() >= 4) {
          sLabel = sData.substring(0, 4);
        }
        else {
          sLabel = sData;
        }
        if (!sLabel.equals(sOldLabel)) { //a new set of labels
          if (sLabel.equals("Adult Rel BA") || sLabel.equals("ARBA")) {
            iARBAIndex = i;
          }
          else if (sLabel.equals("Adult Abs BA") || sLabel.equals("AABA")) {
            iAABAIndex = i;
          }
          else if (sLabel.equals("Adult Abs Den") || sLabel.equals("AADN")) {
            iAADNIndex = i;
          }
          else if (sLabel.equals("Adult Rel Den") || sLabel.equals("ARDN")) {
            iARDNIndex = i;
          }
          else if (sLabel.equals("Sapl Abs BA") || sLabel.equals("JABA")) {
            iJABAIndex = i;
          }
          else if (sLabel.equals("Sapl Rel BA") || sLabel.equals("JRBA")) {
            iJRBAIndex = i;
          }
          else if (sLabel.equals("Sapl Abs Den") || sLabel.equals("JADN")) {
            iJADNIndex = i;
          }
          else if (sLabel.equals("Sapl Rel Den") || sLabel.equals("JRDN")) {
            iJRDNIndex = i;
          }
          else if (sLabel.equals("Sdl Abs Den") || sLabel.equals("SADN")) {
            iSADNIndex = i;
          }
          else if (sLabel.equals("Sdl Rel Den") || sLabel.equals("SRDN")) {
            iSRDNIndex = i;
          }
          else if (sLabel.equals("Snag Abs BA")) {
            iSNABAIndex = i;
          }
          else if (sLabel.equals("Snag Rel BA")) {
            iSNRBAIndex = i;
          }
          else if (sLabel.equals("Snag Abs Den")) {
            iSNADNIndex = i;
          }
          else if (sLabel.equals("Snag Rel Den")) {
            iSNRDNIndex = i;
          }

        }
        sOldLabel = sLabel;
      }

      //Declare our vector arrays and initialize
      //Absolute basal area
      p_fABA = null;
      p_fABA = new Vector[m_iNumSubplots][m_iNumTypes][p_sSpeciesNames.size()];
      for (i = 0; i < m_iNumSubplots; i++) {
        for (j = 0; j < m_iNumTypes; j++) {
          for (k = 0; k < p_sSpeciesNames.size(); k++) {
            p_fABA[i][j][k] = new Vector(0);
          }
        }
      }

      //Basal area totals
      p_fABT = null;
      p_fABT = new Vector[m_iNumSubplots][m_iNumTypes];
      for (i = 0; i < m_iNumSubplots; i++) {
        for (j = 0; j < m_iNumTypes; j++) {
          p_fABT[i][j] = new Vector(0);
        }
      }

      //Relative basal area
      p_fRBA = null;
      p_fRBA = new Vector[m_iNumSubplots][m_iNumTypes][p_sSpeciesNames.size()];
      for (i = 0; i < m_iNumSubplots; i++) {
        for (j = 0; j < m_iNumTypes; j++) {
          for (k = 0; k < p_sSpeciesNames.size(); k++) {
            p_fRBA[i][j][k] = new Vector(0);
          }
        }
      }

      //Absolute density
      p_fADN = null;
      p_fADN = new Vector[m_iNumSubplots][m_iNumTypes][p_sSpeciesNames.size()];
      for (i = 0; i < m_iNumSubplots; i++) {
        for (j = 0; j < m_iNumTypes; j++) {
          for (k = 0; k < p_sSpeciesNames.size(); k++) {
            p_fADN[i][j][k] = new Vector(0);
          }
        }
      }

      //Density area totals
      p_fADT = null;
      p_fADT = new Vector[m_iNumSubplots][m_iNumTypes];
      for (i = 0; i < m_iNumSubplots; i++) {
        for (j = 0; j < m_iNumTypes; j++) {
          p_fADT[i][j] = new Vector(0);
        }
      }

      //Relative density
      p_fRDN = null;
      p_fRDN = new Vector[m_iNumSubplots][m_iNumTypes][p_sSpeciesNames.size()];
      for (i = 0; i < m_iNumSubplots; i++) {
        for (j = 0; j < m_iNumTypes; j++) {
          for (k = 0; k < p_sSpeciesNames.size(); k++) {
            p_fRDN[i][j][k] = new Vector(0);
          }
        }
      }

      //Read to the end of the file
      p_oLine = ModelFileFunctions.ReadLine(in);
      while (p_oLine.size() > 0) {

        if (iLineSize != p_oLine.size()) {
          //Premature file trunction
          ModelException oErr = new ModelException();
          oErr.iErrorCode = ErrorGUI.TAB_FILE_PREMATURE_END;
          oErr.sMessage = m_sFilename;
          throw (oErr);
        }

        //Get the subplot index
        if (bSubplotsExist) {
          sData = (String) p_oLine.elementAt(1);
          try {
            iSubplotIndex = new Integer(sData.trim()).intValue();
          }
          catch (java.lang.NumberFormatException oE) {
            ModelException oErr = new ModelException();
            oErr.iErrorCode = ErrorGUI.BAD_DATA;
            oErr.sMessage = "I cannot read the data in the file.";
            oErr.sCallingFunction = "JAVA";
            throw (oErr);
          }
        }
        else {
          iSubplotIndex = 0;
        }

        //Read each kind of data

        //Adults
        if (iAABAIndex != -1) {
          for (i = 0; i < p_sSpeciesNames.size(); i++) {
            sData = (String) p_oLine.elementAt(i + iAABAIndex);
            p_fABA[iSubplotIndex][ADULT][i].add(new Float(sData));
          }
          //And the total
          sData = (String) p_oLine.elementAt(i + iAABAIndex);
          p_fABT[iSubplotIndex][ADULT].add(new Float(sData));
        }

        if (iARBAIndex != -1) {
          for (i = 0; i < p_sSpeciesNames.size(); i++) {
            sData = (String) p_oLine.elementAt(i + iARBAIndex);
            p_fRBA[iSubplotIndex][ADULT][i].add(new Float(sData));
          }
        }

        if (iAADNIndex != -1) {
          for (i = 0; i < p_sSpeciesNames.size(); i++) {
            sData = (String) p_oLine.elementAt(i + iAADNIndex);
            p_fADN[iSubplotIndex][ADULT][i].add(new Float(sData));
          }
          //And the total
          sData = (String) p_oLine.elementAt(i + iAADNIndex);
          p_fADT[iSubplotIndex][ADULT].add(new Float(sData));
        }

        if (iARDNIndex != -1) {
          for (i = 0; i < p_sSpeciesNames.size(); i++) {
            sData = (String) p_oLine.elementAt(i + iARDNIndex);
            p_fRDN[iSubplotIndex][ADULT][i].add(new Float(sData));
          }
        }

        //Snags
        if (iSNRDNIndex != -1) {
          for (i = 0; i < p_sSpeciesNames.size(); i++) {
            sData = (String) p_oLine.elementAt(i + iSNRDNIndex);
            p_fRDN[iSubplotIndex][SNAG][i].add(new Float(sData));
          }
        }

        if (iSNABAIndex != -1) {
          for (i = 0; i < p_sSpeciesNames.size(); i++) {
            sData = (String) p_oLine.elementAt(i + iSNABAIndex);
            p_fABA[iSubplotIndex][SNAG][i].add(new Float(sData));
          }
          //And the total
          sData = (String) p_oLine.elementAt(i + iSNABAIndex);
          p_fABT[iSubplotIndex][SNAG].add(new Float(sData));
        }

        if (iSNRBAIndex != -1) {
          for (i = 0; i < p_sSpeciesNames.size(); i++) {
            sData = (String) p_oLine.elementAt(i + iSNRBAIndex);
            p_fRBA[iSubplotIndex][SNAG][i].add(new Float(sData));
          }
        }

        if (iSNADNIndex != -1) {
          for (i = 0; i < p_sSpeciesNames.size(); i++) {
            sData = (String) p_oLine.elementAt(i + iSNADNIndex);
            p_fADN[iSubplotIndex][SNAG][i].add(new Float(sData));
          }
          //And the total
          sData = (String) p_oLine.elementAt(i + iSNADNIndex);
          p_fADT[iSubplotIndex][SNAG].add(new Float(sData));
        }

        //Saplings
        if (iJABAIndex != -1) {
          for (i = 0; i < p_sSpeciesNames.size(); i++) {
            sData = (String) p_oLine.elementAt(i + iJABAIndex);
            p_fABA[iSubplotIndex][SAPLING][i].add(new Float(sData));
          }
          //And the total
          sData = (String) p_oLine.elementAt(i + iJABAIndex);
          p_fABT[iSubplotIndex][SAPLING].add(new Float(sData));
        }

        if (iJRBAIndex != -1) {
          for (i = 0; i < p_sSpeciesNames.size(); i++) {
            sData = (String) p_oLine.elementAt(i + iJRBAIndex);
            p_fRBA[iSubplotIndex][SAPLING][i].add(new Float(sData));
          }
        }

        if (iJADNIndex != -1) {
          for (i = 0; i < p_sSpeciesNames.size(); i++) {
            sData = (String) p_oLine.elementAt(i + iJADNIndex);
            p_fADN[iSubplotIndex][SAPLING][i].add(new Float(sData));
          }
          //And the total
          sData = (String) p_oLine.elementAt(i + iJADNIndex);
          p_fADT[iSubplotIndex][SAPLING].add(new Float(sData));
        }

        if (iJRDNIndex != -1) {
          for (i = 0; i < p_sSpeciesNames.size(); i++) {
            sData = (String) p_oLine.elementAt(i + iJRDNIndex);
            p_fRDN[iSubplotIndex][SAPLING][i].add(new Float(sData));
          }
        }

        //Seedlings
        if (iSADNIndex != -1) {
          for (i = 0; i < p_sSpeciesNames.size(); i++) {
            sData = (String) p_oLine.elementAt(i + iSADNIndex);
            p_fADN[iSubplotIndex][SEEDLING][i].add(new Float(sData));
          }
          //And the total
          sData = (String) p_oLine.elementAt(i + iSADNIndex);
          p_fADT[iSubplotIndex][SEEDLING].add(new Float(sData));
        }

        if (iSRDNIndex != -1) {
          for (i = 0; i < p_sSpeciesNames.size(); i++) {
            sData = (String) p_oLine.elementAt(i + iSRDNIndex);
            p_fRDN[iSubplotIndex][SEEDLING][i].add(new Float(sData));
          }
        }

        p_oLine = ModelFileFunctions.ReadLine(in);
      }

      //Transfer the data over to our fixed arrays
      mp_sSpeciesNames = null;
      mp_sSpeciesNames = new String[p_sSpeciesNames.size()];
      for (i = 0; i < p_sSpeciesNames.size(); i++) {
        //The XML will have replaced any spaces in species names with
        //underscores - so change them back
        String sName = (String) p_sSpeciesNames.elementAt(i);
        mp_sSpeciesNames[i] = sName.replace('_', ' ');

      }
      mp_fABA = null;
      mp_fABA = new Float[m_iNumSubplots][m_iNumTypes][p_sSpeciesNames.size()][];
      for (i = 0; i < m_iNumSubplots; i++) {
        for (j = 0; j < m_iNumTypes; j++) {
          for (k = 0; k < p_sSpeciesNames.size(); k++) {
            if (p_fABA[i][j][k].size() > 0) {
              m_iNumTimesteps = p_fABA[i][j][k].size() - 1;
              mp_fABA[i][j][k] = new Float[p_fABA[i][j][k].size()];
              for (m = 0; m < p_fABA[i][j][k].size(); m++) {
                mp_fABA[i][j][k][m] = (Float) p_fABA[i][j][k].elementAt(m);
              }
            }
          }
        }
      }

      mp_fRBA = null;
      mp_fRBA = new Float[m_iNumSubplots][m_iNumTypes][p_sSpeciesNames.size()][];
      for (i = 0; i < m_iNumSubplots; i++) {
        for (j = 0; j < m_iNumTypes; j++) {
          for (k = 0; k < p_sSpeciesNames.size(); k++) {
            if (p_fRBA[i][j][k].size() > 0) {
              m_iNumTimesteps = p_fRBA[i][j][k].size() - 1;
              mp_fRBA[i][j][k] = new Float[p_fRBA[i][j][k].size()];
              for (m = 0; m < p_fRBA[i][j][k].size(); m++) {
                mp_fRBA[i][j][k][m] = (Float) p_fRBA[i][j][k].elementAt(m);
              }
            }
          }
        }
      }

      mp_fADN = null;
      mp_fADN = new Float[m_iNumSubplots][m_iNumTypes][p_sSpeciesNames.size()][];
      for (i = 0; i < m_iNumSubplots; i++) {
        for (j = 0; j < m_iNumTypes; j++) {
          for (k = 0; k < p_sSpeciesNames.size(); k++) {
            if (p_fADN[i][j][k].size() > 0) {
              m_iNumTimesteps = p_fADN[i][j][k].size() - 1;
              mp_fADN[i][j][k] = new Float[p_fADN[i][j][k].size()];
              for (m = 0; m < p_fADN[i][j][k].size(); m++) {
                mp_fADN[i][j][k][m] = (Float) p_fADN[i][j][k].elementAt(m);
              }
            }
          }
        }
      }

      mp_fRDN = null;
      mp_fRDN = new Float[m_iNumSubplots][m_iNumTypes][p_sSpeciesNames.size()][];
      for (i = 0; i < m_iNumSubplots; i++) {
        for (j = 0; j < m_iNumTypes; j++) {
          for (k = 0; k < p_sSpeciesNames.size(); k++) {
            if (p_fRDN[i][j][k].size() > 0) {
              m_iNumTimesteps = p_fRDN[i][j][k].size() - 1;
              mp_fRDN[i][j][k] = new Float[p_fRDN[i][j][k].size()];
              for (m = 0; m < p_fRDN[i][j][k].size(); m++) {
                mp_fRDN[i][j][k][m] = (Float) p_fRDN[i][j][k].elementAt(m);
              }
            }
          }
        }
      }

      mp_fABT = null;
      mp_fABT = new Float[m_iNumSubplots][m_iNumTypes][];
      for (i = 0; i < m_iNumSubplots; i++) {
        for (j = 0; j < m_iNumTypes; j++) {
          if (p_fABT[i][j].size() > 0) {
            mp_fABT[i][j] = new Float[p_fABT[i][j].size()];
            for (k = 0; k < p_fABT[i][j].size(); k++) {
              mp_fABT[i][j][k] = (Float) p_fABT[i][j].elementAt(k);
            }
          }
        }
      }

      mp_fADT = null;
      mp_fADT = new Float[m_iNumSubplots][m_iNumTypes][];
      for (i = 0; i < m_iNumSubplots; i++) {
        for (j = 0; j < m_iNumTypes; j++) {
          if (p_fADT[i][j].size() > 0) {
            mp_fADT[i][j] = new Float[p_fADT[i][j].size()];
            for (k = 0; k < p_fADT[i][j].size(); k++) {
              mp_fADT[i][j][k] = (Float) p_fADT[i][j].elementAt(k);
            }
          }
        }
      }
    }
    catch (java.io.IOException oErr) {
      ModelException oExp = new ModelException();
      oExp.iErrorCode = ErrorGUI.BAD_FILE;
      throw (oExp);
    }
    finally {
      if (in != null) {
        try {
          in.close();
        }
        catch (java.io.IOException e) {
          ; //do nothing
        }
      }
    }
  }

  /**
   * Translates data into the CategoryDataset format, which can then be fed
   * to a chart for display.
   * @param p_fDataToGraph The data to graph as a 2D Float array, with the
   * first index being number of species and the second number of timesteps.
   * @param p_fTotals The totals array, sized number of timesteps, if
   * applicable; if there are no totals, pass null.
   * @return The oDataset.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  private DefaultCategoryDataset CreateDataset(Float[][] p_fDataToGraph,
                                               Float[] p_fTotals) {
    int i, j;

    //Create the row keys.  Rows equal timesteps.
    String[] p_sRowKeys = new String[p_fDataToGraph[0].length];
    for (i = 0; i < p_fDataToGraph[0].length; i++) {
      p_sRowKeys[i] = String.valueOf(i);

      //Now the column keys.
    }
    String[] p_sColumnKeys = new String[mp_sSpeciesNames.length + 1];
    for (i = 0; i < mp_sSpeciesNames.length; i++) {
      p_sColumnKeys[i] = mp_sSpeciesNames[i];
    }
    //Last one for total
    if (null != p_fTotals) {
      p_sColumnKeys[mp_sSpeciesNames.length] = "Total";
    }

    // create the oDataset...
    DefaultCategoryDataset oDataset = new DefaultCategoryDataset();

    //Add each value, keyed to the proper row and column
    for (i = 0; i < mp_sSpeciesNames.length; i++) {
      for (j = 0; j < p_fDataToGraph[i].length; j++) {
        oDataset.addValue(p_fDataToGraph[i][j], p_sColumnKeys[i], p_sRowKeys[j]);
      }
    }

    //Now add the total
    if (null != p_fTotals) {
      for (i = 0; i < p_fTotals.length; i++) {
        oDataset.addValue(p_fTotals[i], p_sColumnKeys[mp_sSpeciesNames.length],
                          p_sRowKeys[i]);
      }
    }

    return oDataset;
  }

  /**
   * This will take a dataset and make a copy with only series for species
   * which are supposed to be visible.
   * @param oDataset The dataset.
   * @param oLegend The legend which controls which species are visible.
   * @return The dataset copy.
   * @throws ModelException wrapping another exception.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  protected DefaultCategoryDataset AdjustVisibleSpecies(DefaultCategoryDataset
      oDataset, Legend oLegend) throws ModelException {

    Comparable oRowKey, oColKey;
    int iSp, iNumSpecies = oLegend.GetNumberOfSpecies(), iRow, iCol, iNumRows,
        iNumCols, i;

    //Make a copy of the new dataset - cloning doesn't work
    DefaultCategoryDataset oNewDataset = new DefaultCategoryDataset();
    iNumRows = oDataset.getRowCount();
    iNumCols = oDataset.getColumnCount();
    for (iCol = 0; iCol < iNumCols; iCol++) {
      oColKey = oDataset.getColumnKey(iCol);
      for (iRow = 0; iRow < iNumRows; iRow++) {
        oRowKey = oDataset.getRowKey(iRow);
        oNewDataset.setValue(oDataset.getValue(oRowKey, oColKey), oRowKey,
                             oColKey);
      }
    }

    for (iSp = 0; iSp < iNumSpecies; iSp++) {
      //The "rows" are what contain the species
      if (!oLegend.GetIsSpeciesSelected(iSp)) {
        for (i = 0; i < oNewDataset.getRowCount(); i++) {
          if (oNewDataset.getRowKey(i).compareTo(oLegend.GetSpeciesCodeName(iSp)) ==
              0) {
            oNewDataset.removeRow(i);
          }
        }
      }
    }

    return oNewDataset;
  }

  /**
   * Takes care of writing out a line graph's data.
   * <p>Copyright: Copyright (c) Charles D. Canham 2003</p>
   * <p>Company: Institute of Ecosystem Studies</p>
   * @author Lora E. Murphy
   * @version 1.0
   */
  class LineGraphFileWriter
      implements java.awt.event.ActionListener {
    JInternalFrame m_jFrame = null; /**<Frame for a chart*/
    int m_iTreeType, /**<Tree type - SEEDLING, SAPLING, etc.*/
        m_iChartType,  /**<Chart type - RELATIVE_DENSITY, etc.*/
        m_iSubplot; /**<Subplot - 0 if whole plot*/

    /**
     * Constructor.
     * @param iTreeType int Tree type - SEEDLING, SAPLING, etc.
     * @param iChartType int Chart type - ABSOLUTE_DENSITY, etc.
     * @param iSubplot Subplot number
     */
    public LineGraphFileWriter(int iTreeType, int iChartType, int iSubplot) {
      m_iTreeType = iTreeType;
      m_iChartType = iChartType;
      m_iSubplot = iSubplot;
    }

    /**
     * Sets the frame for this object to post its dialogs on.
     * @param jFrame JInternalFrame Frame to set.
     */
    public void SetFrame(JInternalFrame jFrame) {
      m_jFrame = jFrame;
    }

    /**
     * Writes the chart to file.
     * @param oEvent ActionEvent Event which triggered this method.
     */
    public void actionPerformed(java.awt.event.ActionEvent oEvent) {
      java.io.FileWriter jOut = null;
      String sFileName = "";
      try {

        //Allow the user to save a text file
        javax.swing.JFileChooser jChooser = new javax.swing.JFileChooser();
        jChooser.setFileFilter(new javawrapper.TextFileFilter());

        int iReturnVal = jChooser.showSaveDialog(m_jFrame);
        if (iReturnVal == javax.swing.JFileChooser.APPROVE_OPTION) {
          //User chose a file - trigger the save
          java.io.File oFile = jChooser.getSelectedFile();
          if (oFile.exists()) {
            iReturnVal = javax.swing.JOptionPane.showConfirmDialog(
                m_jFrame,
                "Do you wish to overwrite the existing file?",
                "Model",
                javax.swing.JOptionPane.YES_NO_OPTION);
          }
          if (!oFile.exists() ||
              iReturnVal == javax.swing.JOptionPane.YES_OPTION) {
            sFileName = oFile.getAbsolutePath();
            if (sFileName.indexOf(".") == -1) {
              sFileName += ".txt";
            }

            jOut = new java.io.FileWriter(sFileName);
            int i, j;
            boolean bUseTotals = (m_iChartType ==
                                  ShortOutputFileManager.ABSOLUTE_BASAL_AREA ||
                                  m_iChartType ==
                                  ShortOutputFileManager.ABSOLUTE_DENSITY);

            //Write the chart header
            jOut.write(m_jFrame.getTitle() + "\n");
            if (m_iChartType == ShortOutputFileManager.ABSOLUTE_BASAL_AREA) {
              jOut.write("Values in square meters of basal area per hectare.\n");
            }
            else if (m_iChartType == ShortOutputFileManager.ABSOLUTE_DENSITY) {
              jOut.write("Values in number of trees per hectare.\n");
            }

            //Get the appropriate data array
            Object[][][][] p_fData = null;
            Object[][][] p_fTotals = null;
            if (m_iChartType == ShortOutputFileManager.ABSOLUTE_BASAL_AREA) {
              p_fData = mp_fABA;
              p_fTotals = mp_fABT;
            }
            else if (m_iChartType == ShortOutputFileManager.ABSOLUTE_DENSITY) {
              p_fData = mp_fADN;
              p_fTotals = mp_fADT;
            }
            else if (m_iChartType == ShortOutputFileManager.RELATIVE_BASAL_AREA) {
              p_fData = mp_fRBA;
            }
            else if (m_iChartType == ShortOutputFileManager.RELATIVE_DENSITY) {
              p_fData = mp_fRDN;
            }
            else {
              throw (new ModelException(javawrapper.ErrorGUI.BAD_ARGUMENT,
                                        "JAVA",
                                        "I don't recognize the data type."));
            }

            //Make sure there's data: Array indexes are #1 - subplot, #2 -
            //type, #3 - species, and #4 - timestep
            if (p_fData[m_iSubplot][m_iTreeType] == null) {
              jOut.write("This chart has no data.");
              return;
            }

            //Write column headers
            jOut.write("Timestep");
            for (i = 0; i < p_fData[m_iSubplot][m_iTreeType].length; i++) {
              jOut.write("\t" + m_oLegend.GetSpeciesDisplayName(i));
            }
            if (bUseTotals) {
              jOut.write("\tTotal");
            }
            jOut.write("\n");

            //Write data
            for (i = 0; i < p_fData[m_iSubplot][m_iTreeType][0].length; i++) {
              jOut.write(String.valueOf(i));
              for (j = 0; j < p_fData[m_iSubplot][m_iTreeType].length; j++) {
                jOut.write("\t" +
                           p_fData[m_iSubplot][m_iTreeType][j][i].toString());
              }
              if (bUseTotals) {
                jOut.write("\t" + p_fTotals[m_iSubplot][m_iTreeType][i]);
              }
              jOut.write("\n");
            }
            //Tell the user
            javax.swing.JOptionPane.showMessageDialog(m_jFrame,
                "File has been saved.");

          }
        }
      }
      catch (javawrapper.ModelException oErr) {
        javawrapper.ErrorGUI oHandler = new javawrapper.ErrorGUI(m_jFrame);
        oHandler.WriteErrorMessage(oErr);
      }
      catch (java.io.IOException oErr) {
        javawrapper.ModelException oExp = new javawrapper.ModelException();
        oExp.sMessage = "There was a problem writing the file " + sFileName +
            ".";
        oExp.sCallingFunction = "JAVA";
        javawrapper.ErrorGUI oHandler = new javawrapper.ErrorGUI(m_jFrame);
        oHandler.WriteErrorMessage(oExp);
      }
      finally {
        try {
          if (jOut != null) {
            jOut.close();
          }
        }
        catch (java.io.IOException oErr) {
          ; //do nothing - not much can be done at this point
        }
      }
    }
  }

  /**
   * Writes out table data to file.
   * <p>Copyright: Copyright (c) Charles D. Canham 2003</p>
   * <p>Company: Institute of Ecosystem Studies</p>
   * @author Lora E. Murphy
   * @version 1.0
   */
  class TableFileWriter
      implements java.awt.event.ActionListener {

    private JInternalFrame m_jFrame; /**<For getting the data and creating dialogs*/

    /**
     * Sets the frame for this object to post its dialogs on.
     * @param jFrame JInternalFrame Frame to set.
     */
    public void SetFrame(JInternalFrame jFrame) {
      m_jFrame = jFrame;
    }

    /**
     * Writes the table.
     * @param oEvent ActionEvent Event triggering the table writing.
     */
    public void actionPerformed(java.awt.event.ActionEvent oEvent) {
      java.io.FileWriter jOut = null;
      String sFileName = "";
      try {

        //Allow the user to save a text file
        javax.swing.JFileChooser jChooser = new javax.swing.JFileChooser();
        jChooser.setFileFilter(new javawrapper.TextFileFilter());

        int iReturnVal = jChooser.showSaveDialog(m_jFrame);
        if (iReturnVal == javax.swing.JFileChooser.APPROVE_OPTION) {
          //User chose a file - trigger the save
          java.io.File oFile = jChooser.getSelectedFile();
          if (oFile.exists()) {
            iReturnVal = javax.swing.JOptionPane.showConfirmDialog(
                m_jFrame,
                "Do you wish to overwrite the existing file?",
                "Model",
                javax.swing.JOptionPane.YES_NO_OPTION);
          }
          if (!oFile.exists() ||
              iReturnVal == javax.swing.JOptionPane.YES_OPTION) {
            sFileName = oFile.getAbsolutePath();
            if (sFileName.indexOf(".") == -1) {
              sFileName += ".txt";
            }
            String[] p_sNames = new String[] {
                "seedling_table", "sapling_table",
                "adult_table", "snag_table"};

            jOut = new java.io.FileWriter(sFileName);
            int i, j, iTimestep;

            //Get the current timestep
            JLabel jTempLabel = (JLabel) DataGrapher.FindNamedComponent(m_jFrame.getContentPane(), "current_timestep");
            iTimestep = new Integer(jTempLabel.getText()).intValue();

            //Write the chart header
            jOut.write(m_jFrame.getTitle() + "\n");
            jOut.write("Timestep: " + String.valueOf( iTimestep ) + "\n");

            //Find each table
            java.awt.Component jContentPane = m_jFrame.getContentPane();
            for (i = 0; i < 4; i++) {
              JPanel jTablePanel = (JPanel) DataGrapher.FindNamedComponent(
                  jContentPane, "table_panel_" + i);

              //First find the label heading the panel
              for (j = 0; j < jTablePanel.getComponentCount(); j++) {
                if (jTablePanel.getComponent(j) instanceof JLabel) {
                  jOut.write("\n" +
                             ( (JLabel) jTablePanel.getComponent(j)).getText() +
                             "\n");
                  break;
                }
              }

              //Now find the table
              JTable jTable = (JTable) DataGrapher.FindNamedComponent(
                  jTablePanel,
                  p_sNames[i]);
              int iNumRows = jTable.getRowCount();
              int iNumCols = jTable.getColumnCount();
              int iRow, iCol;

              //Write header row
              jOut.write(String.valueOf(jTable.getTableHeader().getColumnModel().
                                        getColumn(0).getHeaderValue()));
              for (iCol = 1; iCol < iNumCols; iCol++) {
                jOut.write("\t" +
                           String.valueOf(jTable.getTableHeader().
                                          getColumnModel().
                                          getColumn(iCol).getHeaderValue()));
              }
              jOut.write("\n");

              for (iRow = 0; iRow < iNumRows; iRow++) {
                jOut.write(String.valueOf(jTable.getValueAt(iRow, 0)));
                for (iCol = 1; iCol < iNumCols; iCol++) {
                  jOut.write("\t" +
                             String.valueOf(jTable.getValueAt(iRow, iCol)));
                }
                jOut.write("\n");
              }
            }

            //Tell the user
            javax.swing.JOptionPane.showMessageDialog(m_jFrame,
                "File has been saved.");

          }
        }
      }
      catch (java.io.IOException oErr) {
        javawrapper.ModelException oExp = new javawrapper.ModelException();
        oExp.sMessage = "There was a problem writing the file " + sFileName +
            ".";
        oExp.sCallingFunction = "JAVA";
        javawrapper.ErrorGUI oHandler = new javawrapper.ErrorGUI(m_jFrame);
        oHandler.WriteErrorMessage(oExp);
      }
      finally {
        try {
          if (jOut != null) {
            jOut.close();
          }
        }
        catch (java.io.IOException oErr) {
          ; //do nothing - not much can be done at this point
        }
      }
    }
  }
}

/**
 * This encapsulates any kind of a chart dataset with a string to match it to
 * its chart window.
 * <p>Copyright: Copyright (c) Charles D. Canham 2003</p>
 * <p>Company: Institute of Ecosystem Studies</p>
 * @author Lora E. Murphy
 * @version 1.0
 *
 * <br>Edit history:
 * <br>------------------
 * <br>April 28, 2004: Submitted in beta version (LEM)
 */
class ModelDataset {
  private Object m_oDataset; /**<Dataset being encapsulated*/
  private String m_sName; /**<Associated string*/

  /**
   * Constructor
   * @param oDataset The dataset
   * @param sName String name
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public ModelDataset(Object oDataset, String sName) {
    m_oDataset = oDataset;
    m_sName = sName;
  }

  /**
   * Gets the dataset.
   * @return The dataset.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public Object GetDataset() {
    return m_oDataset;
  }

  /**
   * Gets the string for the dataset.
   * @return Dataset string.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public String GetName() {
    return m_sName;
  }
}

/**
 * Responds to the buttons in the overview table windows.
 * <p>Copyright: Copyright (c) Charles D. Canham 2003</p>
 * <p>Company: Institute of Ecosystem Studies</p>
 * @author Lora E. Murphy
 * @version 1.0
 */
class TableUpdater
    implements java.awt.event.ActionListener {
  /**The content panel containing the tables and controls*/
  JPanel m_jContentPanel;

  /**The file manager that's managing the table data*/
  ShortOutputFileManager m_oManager;

  /**The subplot for this table*/
  int m_iSubplot;

  /**
   * Constructor.
   * @param jContentPanel Content panel of the window containing the
   * buttons and tables.
   * @param oManager The file manager that's managing the table's data.
   * @param iSubplot The subplot that this table is displaying, or 0 if it's
   * the whole plot's data.
   */
  TableUpdater(JPanel jContentPanel, ShortOutputFileManager oManager,
               int iSubplot) {
    m_jContentPanel = jContentPanel;
    m_oManager = oManager;
    m_iSubplot = iSubplot;
  }

  /**
   * Responds to the button events.
   * @param oEvent Event that triggered this object.
   */
  public void actionPerformed(java.awt.event.ActionEvent oEvent) {
    String sAction = oEvent.getActionCommand();
    if (sAction.equals("StepForward")) {

      //Advance the table one step
      int iCurrentTimestep, iNumberTimesteps;

      //Find the total number of timesteps
      JLabel jLabel = (JLabel) DataGrapher.FindNamedComponent(m_jContentPanel,
          "number_timesteps");
      if (jLabel == null) {
        return;
      }
      iNumberTimesteps = new Integer(jLabel.getText()).intValue();

      //Find the timestep it's on right now
      jLabel = (JLabel) DataGrapher.FindNamedComponent(m_jContentPanel,
          "current_timestep");
      if (jLabel == null) {
        return;
      }
      iCurrentTimestep = new Integer(jLabel.getText()).intValue();

      //If we're already at the end, skip out
      if (iCurrentTimestep == iNumberTimesteps) {
        return;
      }

      //Update the tables
      iCurrentTimestep++;
      m_oManager.UpdateTable(m_jContentPanel, iCurrentTimestep, m_iSubplot);
    }
    else if (sAction.equals("StepBack")) {

      //Take the table one step back
      int iCurrentTimestep;

      //Find the timestep it's on right now
      JLabel jLabel = (JLabel) DataGrapher.FindNamedComponent(m_jContentPanel,
          "current_timestep");
      if (jLabel == null) {
        return;
      }
      iCurrentTimestep = new Integer(jLabel.getText()).intValue();

      //If we're already at the end, skip out
      if (iCurrentTimestep == 0) {
        return;
      }

      //Update the tables
      iCurrentTimestep--;
      m_oManager.UpdateTable(m_jContentPanel, iCurrentTimestep, m_iSubplot);
    }
  }

}
