package javawrapper;

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import java.io.File;
import javax.help.*;
import java.net.URL;

/**
 * Main application window.
 *
 * The main application window has two main functions; creating and editing
 * parameter files for running the model, and viewing the output generated.
 * To run these functions, it has a separate manager object for each.  The
 * rest of the components in this window are for GUI display.
 * <p>Copyright: Copyright (c) Charles D. Canham 2003</p>
 * <p>Company: Institute of Ecosystem Studies</p>
 * @author Lora E. Murphy
 * @version 1.1
 * @todo Window icon
 *
 * <br>Edit history:
 * <br>------------------
 * <br>April 28, 2004: Submitted in beta version (LEM)
 * <br>May 4, 2004: Added the Harvest window (LEM)
 * <br>July 30, 2004:  Added the ability to close output files (LEM)
 * <br>November 8, 2004:  Added pause run and stop run functionality (LEM)
 * <br>November 12, 2004:  Added Help->About (LEM)
 * <br>November 17, 2004:  Added the run control buttons panel (LEM)
 * <br>November 18, 2004:  Added real-time data visualization (LEM)
 * <br>March 31, 2005:  Added batch file choice and upgraded version number to
 * 1.1 (not that I've been very good about this) (LEM)
 * <br>October 13, 2005:  Added scroll bars (LEM)
 */

public class MainWindow
    extends JFrame
    implements ActionListener {

  /**Object managing the interface between this application and the C++
   * model core*/
  private Interface m_oInterface = null;

  /**Object that allows us to update the GUI while the core model is
   * running*/
  private Timer m_oTimer;

  /**Manager controlling all functions related to creating and editing
   * parameter files and running the model*/
  protected GUIManager m_oDataManager;

  /**Manager controlling all functions related to output data visualization*/
  protected DataVisualizerManager m_oDataVisualizerManager;

  /**We'll use this so that our file choosers can open to the same directory
     that they were in last*/
  protected File m_oFileDirectory = new File("");

  /**The HelpSet object for the help file - class is from JavaHelp*/
  protected HelpSet m_oHelpSet;

  /**JavaHelp class for managing help displays*/
  public HelpBroker m_oHelpBroker;
  private int m_iState; /**<file loaded, no file loaded, running, etc.*/

  /**Desktop pane - allows the use of JInternalFrames*/
  private JDesktopPane m_oDesktop;
  private JScrollPane m_jScroller;

  //*********************************
   // Menu bar
   //*********************************
    //Contents of "File" menu
    protected JMenuItem m_jMenuFileNew = new JMenuItem("New parameter file",
        KeyEvent.VK_N);
  protected JMenuItem m_jMenuFileBatchNew = new JMenuItem("New batch file",
      KeyEvent.VK_B);
  protected JMenuItem m_jMenuFileOpen = new JMenuItem("Open file",
      KeyEvent.VK_O);
  protected JMenuItem m_jMenuFileSave = new JMenuItem("Save parameter file",
      KeyEvent.VK_S);
  protected JMenuItem m_jMenuFileOpenData = new JMenuItem("Open output file",
      KeyEvent.VK_R);
  protected JMenuItem m_jMenuFileCloseData = new JMenuItem(
      "Close output file", KeyEvent.VK_C);
  protected JMenuItem m_jMenuFileExit = new JMenuItem("Exit", KeyEvent.VK_X);

  //Contents of "Edit" menu
  protected JMenuItem m_jMenuEditTree = new JMenuItem("Tree setup",
      KeyEvent.VK_T);
  protected JMenuItem m_jMenuEditGrid = new JMenuItem("Grid layer setup",
      KeyEvent.VK_G);
  protected JMenuItem m_jMenuEditFlow = new JMenuItem("Model flow",
      KeyEvent.VK_M);
  protected JMenuItem m_jMenuEditParameters = new JMenuItem("Parameters",
      KeyEvent.VK_P);
  protected JMenuItem m_jMenuEditDisturbance = new JMenuItem("Episodic events",
      KeyEvent.VK_D);
  protected JMenuItem m_jMenuEditOutput = new JMenuItem("Output options",
      KeyEvent.VK_O);

  //Contents of "Model" menu
  protected JMenuItem m_jMenuModelRun = new JMenuItem("Run", KeyEvent.VK_R);
  protected JMenuItem m_jMenuModelRunBatch = new JMenuItem("Run Batch...");
  protected JMenuItem m_jMenuModelPause = new JMenuItem("Pause", KeyEvent.VK_P);
  protected JMenuItem m_jMenuModelStop = new JMenuItem("Stop run",
      KeyEvent.VK_S);

  //Contents of "Help" menu
  protected JMenuItem m_jMenuHelpContents = new JMenuItem("Contents",
      KeyEvent.VK_C);
  protected JMenuItem m_jMenuHelpAbout = new JMenuItem("About", KeyEvent.VK_A);

  /**Holds the chart choices for the currently open output files*/
  protected JComboBox m_jChartChoicesComboBox = new JComboBox();

  /**Holds the list of currently open output files*/
  protected JComboBox m_jFileChoicesComboBox = new JComboBox();

  /**Field in the status bar holding the parameter file*/
  protected JTextField m_jParameterFileField = new JTextField();

  /**Field holding the model status*/
  protected JTextField m_jModelStatusField = new JTextField();

  /**Field holding model messages*/
  protected JTextField m_jMessagesField = new JTextField();

  /**Button for running model*/
  protected JButton m_jModelRunButton = new JButton(new ModelIcon(15, 15,
      ModelIcon.RIGHT_TRIANGLE));

  /**Button for stopping model*/
  protected JButton m_jModelStopButton = new JButton(new ModelIcon(15, 15,
      ModelIcon.RECTANGLE));

  /**Button for pausing model*/
  protected JButton m_jModelPauseButton = new JButton(new ModelIcon(15, 15,
      ModelIcon.PAUSE));

  /**Button for stepping model*/
  protected JButton m_jModelStepForwardButton = new JButton(new ModelIcon(15,
      15,
      ModelIcon.STEP_FORWARD));

  /**Button for loading run's output*/
  protected JButton m_jModelLoadOutputButton = new JButton(
      "View this run's output");

  /**Flag for whether or not the user is doing real-time data visualization
   * of the current run*/
  private boolean m_bViewingRunOutput = false;

  /**Flag for whether, during real-time data visualization, a new run
   * timestep should be triggered after data updates.  If the user has run the
   * model instead of pausing it or stepping forward one timestep, this
   * is set to true.  Otherwise, it is false and no new timestep will be
   * triggered after the charts have been refreshed.*/
  private boolean m_bKeepRunning = false;

  /**
   * Constructor.  Draws the window.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public MainWindow() {
    enableEvents(AWTEvent.WINDOW_EVENT_MASK);
    CreateGUI();
    try {
      m_oDataManager = new GUIManager(this);
      SetModelState(MainWindowStateSetter.NO_PAR_FILE);
      m_oDataVisualizerManager = new DataVisualizerManager(m_oDesktop,
          m_jScroller);
      LoadHelp();

      m_jMenuHelpAbout.addActionListener(new CSH.DisplayHelpFromSource(
          m_oHelpBroker));
      CSH.setHelpIDString(m_jMenuHelpAbout, "license.license");
    }
    catch (ModelException oErr) {
      ErrorGUI oHandler = new ErrorGUI(this);
      oHandler.WriteErrorMessage(oErr);
    }
  }

  /**
   * Loads the SORTIE help file so that it can be opened from buttons and
   * menu commands.
   * @throws ModelException if the help is not found.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  private void LoadHelp() throws ModelException {
    // Find the HelpSet file and create the HelpSet object.  The SORTIE help
    //jar file should be included into the project as a required library so that
    //it will be added to the classpath
    String sHelpSetFileName = "help/sortiehelp.hs"; //path within the JAR file
    try {
      URL hsURL = HelpSet.findHelpSet(null, sHelpSetFileName);
      m_oHelpSet = new HelpSet(null, hsURL);
    }
    catch (Exception ee) {
      // Say what the exception really is
      ModelException oErr = new ModelException();
      oErr.iErrorCode = ErrorGUI.CANT_FIND_OBJECT;
      oErr.sCallingFunction = "JAVA";
      oErr.sMessage = "I can't find the SORTIE help file.";
      throw (oErr);
    }
    // Create a HelpBroker object:
    m_oHelpBroker = m_oHelpSet.createHelpBroker();
    //Hook up the help menu item
    m_jMenuHelpContents.addActionListener(new CSH.DisplayHelpFromSource(
        m_oHelpBroker));
    CSH.setHelpIDString(m_jMenuHelpContents, "intro");

    //Set the window-level context-sensitive help
    m_oHelpBroker.enableHelpKey(this.getRootPane(), "intro", null);
  }

  /**
   * Gets the current file directory.  This is where file chooser windows
   * should default to.
   * @return The current file directory.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public File GetCurrentFileDirectory() {
    return m_oFileDirectory;
  }

  /**
   * Sets the current file directory.  This is where file chooser windows
   * should default to.
   * @param oDirectory The current file directory.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void SetCurrentFileDirectory(File oDirectory) {
    m_oFileDirectory = oDirectory;
  }

  /**
   * SendMessage accepts a message from the interface.
   * @param oMsg Message to accept.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void SendMessage(ModelMessage oMsg) {
    m_jMessagesField.setText(oMsg.sMessage);
  }

  /**
   * Component initialization and GUI construction.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   * <br>November 17, 2004:  Added run control buttons panel (LEM)
   */
  private void CreateGUI() {

    Dimension screenSize = Toolkit.getDefaultToolkit().getScreenSize();
    int iInset = 50;
    setBounds(iInset, iInset,
              screenSize.width - iInset * 2,
              screenSize.height - iInset * 2);
    this.setTitle("SORTIE-ND Version 6.06");

    //****************************************************************
     //Set up menu bar
     //****************************************************************
      JMenuBar jMenuBar = new JMenuBar();
      //Declare "File menu"
    JMenu jMenuFile = new JMenu();
    //Declare "Edit" menu
    JMenu jMenuEdit = new JMenu();
    //"Model" menu
    JMenu jMenuModel = new JMenu();
    //"Help" menu
    JMenu jMenuHelp = new JMenu();

    //File options text
    jMenuFile.setBackground(SystemColor.control);
    jMenuFile.setFont(new ModelFont());
    jMenuFile.setText("File");
    jMenuFile.setMnemonic(KeyEvent.VK_F);

    //Add file options
    //File New
    m_jMenuFileNew.setBackground(SystemColor.control);
    m_jMenuFileNew.setFont(new ModelFont());
    m_jMenuFileNew.setAccelerator(KeyStroke.getKeyStroke(
        KeyEvent.VK_N, ActionEvent.CTRL_MASK));
    m_jMenuFileNew.setActionCommand("FileNew");
    m_jMenuFileNew.addActionListener(this);

    //File New Batch
    m_jMenuFileBatchNew.setBackground(SystemColor.control);
    m_jMenuFileBatchNew.setFont(new ModelFont());
    m_jMenuFileBatchNew.setActionCommand("FileNewBatch");
    m_jMenuFileBatchNew.addActionListener(this);

    //File Open
    m_jMenuFileOpen.setBackground(SystemColor.control);
    m_jMenuFileOpen.setFont(new ModelFont());
    m_jMenuFileOpen.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_O,
        ActionEvent.CTRL_MASK));
    m_jMenuFileOpen.setActionCommand("FileOpen");
    m_jMenuFileOpen.addActionListener(this);

    //File Open Output
    m_jMenuFileOpenData.setBackground(SystemColor.control);
    m_jMenuFileOpenData.setFont(new ModelFont());
    m_jMenuFileOpenData.setForeground(Color.black);
    m_jMenuFileOpenData.setActionCommand("FileOpenData");
    m_jMenuFileOpenData.addActionListener(this);

    //File Close Output
    m_jMenuFileCloseData.setBackground(SystemColor.control);
    m_jMenuFileCloseData.setFont(new ModelFont());
    m_jMenuFileCloseData.setForeground(Color.black);
    m_jMenuFileCloseData.setActionCommand("FileCloseData");
    m_jMenuFileCloseData.addActionListener(this);

    //File Save
    m_jMenuFileSave.setBackground(SystemColor.control);
    m_jMenuFileSave.setFont(new ModelFont());
    m_jMenuFileSave.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_S,
        ActionEvent.CTRL_MASK));
    m_jMenuFileSave.setActionCommand("FileSave");
    m_jMenuFileSave.addActionListener(this);

    //File Exit
    m_jMenuFileExit.setBackground(SystemColor.control);
    m_jMenuFileExit.setFont(new ModelFont());
    m_jMenuFileExit.setActionCommand("FileExit");
    m_jMenuFileExit.addActionListener(this);

    //Edit options text
    jMenuEdit.setBackground(SystemColor.control);
    jMenuEdit.setFont(new ModelFont());
    jMenuEdit.setFocusPainted(false);
    jMenuEdit.setText("Edit");
    jMenuEdit.setMnemonic(KeyEvent.VK_E);

    //Edit Tree
    m_jMenuEditTree.setBackground(SystemColor.control);
    m_jMenuEditTree.setFont(new ModelFont());
    m_jMenuEditTree.setActionCommand("EditTree");
    m_jMenuEditTree.addActionListener(this);

    //Edit Grid
    m_jMenuEditGrid.setBackground(SystemColor.control);
    m_jMenuEditGrid.setFont(new ModelFont());
    m_jMenuEditGrid.setActionCommand("EditGrid");
    m_jMenuEditGrid.addActionListener(this);

    //Edit Model Flow
    m_jMenuEditFlow.setBackground(SystemColor.control);
    m_jMenuEditFlow.setFont(new ModelFont());
    m_jMenuEditFlow.setActionCommand("EditFlow");
    m_jMenuEditFlow.addActionListener(this);

    //Edit Model Parameters
    m_jMenuEditParameters.setBackground(SystemColor.control);
    m_jMenuEditParameters.setFont(new ModelFont());
    m_jMenuEditParameters.setActionCommand("EditParameters");
    m_jMenuEditParameters.addActionListener(this);

    //Edit Harvest
    m_jMenuEditDisturbance.setBackground(SystemColor.control);
    m_jMenuEditDisturbance.setFont(new ModelFont());
    m_jMenuEditDisturbance.setActionCommand("EditDisturbance");
    m_jMenuEditDisturbance.addActionListener(this);

    //Edit Output optiosn
    m_jMenuEditOutput.setBackground(SystemColor.control);
    m_jMenuEditOutput.setFont(new ModelFont());
    m_jMenuEditOutput.setActionCommand("EditOutput");
    m_jMenuEditOutput.addActionListener(this);

    //Model options text
    jMenuModel.setBackground(SystemColor.control);
    jMenuModel.setFont(new ModelFont());
    jMenuModel.setText("Model");
    jMenuModel.setMnemonic(KeyEvent.VK_M);

    //Model Run
    m_jMenuModelRun.setBackground(SystemColor.control);
    m_jMenuModelRun.setFont(new ModelFont());
    m_jMenuModelRun.setActionCommand("ModelRun");
    m_jMenuModelRun.addActionListener(this);

    //Model Run Batch
    m_jMenuModelRunBatch.setBackground(SystemColor.control);
    m_jMenuModelRunBatch.setFont(new ModelFont());
    m_jMenuModelRunBatch.setActionCommand("ModelRunBatch");
    m_jMenuModelRunBatch.addActionListener(this);

    //Model Pause
    m_jMenuModelPause.setBackground(SystemColor.control);
    m_jMenuModelPause.setFont(new ModelFont());
    m_jMenuModelPause.setActionCommand("ModelPause");
    m_jMenuModelPause.addActionListener(this);

    //Model Stop Run
    m_jMenuModelStop.setBackground(SystemColor.control);
    m_jMenuModelStop.setFont(new ModelFont());
    m_jMenuModelStop.setActionCommand("ModelStop");
    m_jMenuModelStop.addActionListener(this);

    //Help options
    jMenuHelp.setBackground(SystemColor.control);
    jMenuHelp.setFont(new ModelFont());
    jMenuHelp.setText("Help");
    jMenuHelp.setMnemonic(KeyEvent.VK_H);

    m_jMenuHelpAbout.setBackground(SystemColor.control);
    m_jMenuHelpAbout.setFont(new ModelFont());

    m_jMenuHelpContents.setBackground(SystemColor.control);
    m_jMenuHelpContents.setFont(new ModelFont());

    //Build the menu bar
    jMenuBar.setBackground(SystemColor.control);
    jMenuBar.setFont(new ModelFont());
    jMenuFile.add(m_jMenuFileNew);
    jMenuFile.add(m_jMenuFileBatchNew);
    jMenuFile.addSeparator();
    jMenuFile.add(m_jMenuFileOpen);
    jMenuFile.add(m_jMenuFileSave);
    jMenuFile.addSeparator();
    jMenuFile.add(m_jMenuFileOpenData);
    jMenuFile.add(m_jMenuFileCloseData);
    jMenuFile.addSeparator();
    jMenuFile.add(m_jMenuFileExit);

    jMenuHelp.add(m_jMenuHelpAbout);
    jMenuHelp.add(m_jMenuHelpContents);

    //Build the edit menu
    jMenuEdit.add(m_jMenuEditParameters);
    jMenuEdit.addSeparator();
    jMenuEdit.add(m_jMenuEditDisturbance);
    jMenuEdit.addSeparator();
    jMenuEdit.add(m_jMenuEditTree);
    jMenuEdit.add(m_jMenuEditGrid);
    jMenuEdit.add(m_jMenuEditFlow);
    jMenuEdit.addSeparator();
    jMenuEdit.add(m_jMenuEditOutput);

    //Build the model menu
    jMenuModel.add(m_jMenuModelRun);
    jMenuModel.add(m_jMenuModelRunBatch);
    jMenuModel.add(m_jMenuModelPause);
    jMenuModel.add(m_jMenuModelStop);

    jMenuBar.add(jMenuFile);
    jMenuBar.add(jMenuEdit);
    jMenuBar.add(jMenuModel);
    jMenuBar.add(jMenuHelp);

    this.setJMenuBar(jMenuBar);

    m_oDesktop = new JDesktopPane();
    m_jScroller = new JScrollPane(m_oDesktop);

    //****************************************************************
     //Build the status bar
     //****************************************************************
    m_jParameterFileField.setBackground(SystemColor.control);
    m_jParameterFileField.setFont(new ModelFont());
    m_jParameterFileField.setAlignmentX( (float) 0.5);
    m_jParameterFileField.setDisabledTextColor(Color.black);
    m_jParameterFileField.setEditable(false);
    m_jParameterFileField.setText("No parameter file loaded");
    m_jModelStatusField.setBackground(SystemColor.control);
    m_jModelStatusField.setFont(new ModelFont());
    m_jModelStatusField.setDisabledTextColor(Color.black);
    m_jModelStatusField.setEditable(false);
    m_jModelStatusField.setText("Not ready to run");
    m_jMessagesField.setBackground(SystemColor.control);
    m_jMessagesField.setEnabled(false);
    m_jMessagesField.setFont(new ModelFont());
    m_jMessagesField.setDisabledTextColor(Color.black);
    m_jMessagesField.setSelectionStart(0);
    m_jMessagesField.setText("");

    //****************************************************************
     //Build the main panel
     //****************************************************************
    JPanel jMainPanel = new JPanel(new BorderLayout());
    jMainPanel.setMinimumSize(new Dimension(18, 25));
    jMainPanel.add(m_jParameterFileField, BorderLayout.WEST);
    jMainPanel.add(m_jMessagesField, BorderLayout.CENTER);
    jMainPanel.add(m_jModelStatusField, BorderLayout.EAST);

    //Add the desktop
    m_oDesktop.setBackground(SystemColor.control);
    m_oDesktop.setDragMode(JDesktopPane.OUTLINE_DRAG_MODE);

    //****************************************************************
     //Build the data visualizer panel
     //****************************************************************
    JPanel jDataVisualizerPanel = new JPanel(new FlowLayout());
    jDataVisualizerPanel.setName("data_visualizer_panel");
    jDataVisualizerPanel.setBackground(SystemColor.control);
    jDataVisualizerPanel.setBorder(BorderFactory.createMatteBorder(0, 0, 1, 0,
        Color.BLACK));

    JButton jDrawChartButton = new JButton("Draw");
    jDrawChartButton.setFont(new ModelFont());
    jDrawChartButton.setActionCommand("DrawChart");
    jDrawChartButton.addActionListener(this);

    m_jChartChoicesComboBox.setFont(new ModelFont());
    m_jChartChoicesComboBox.addItem("---Load data file---");
    m_jFileChoicesComboBox.addItem("---Load data file---");

    //Label for output files
    JLabel jOutputFilesLabel = new JLabel("Open output files:");
    jOutputFilesLabel.setFont(new ModelFont());

    //Label for chart choices
    JLabel jChartChoicesLabel = new JLabel("Display choices for this file:");
    jChartChoicesLabel.setFont(new ModelFont());

    m_jFileChoicesComboBox.setFont(new ModelFont());
    m_jFileChoicesComboBox.addActionListener(this);

    //Size things
    int iScreenWidth = getBounds().width;
    iScreenWidth -= jDrawChartButton.getPreferredSize().getWidth();
    iScreenWidth -= jOutputFilesLabel.getPreferredSize().getWidth();
    iScreenWidth -= jChartChoicesLabel.getPreferredSize().getWidth();
    iScreenWidth -= 50;
    //File choices gets 2/3 of what's left, chart choices the rest
    m_jFileChoicesComboBox.setPreferredSize(new Dimension( (int) (iScreenWidth *
        0.66),
        m_jFileChoicesComboBox.getPreferredSize().height));
    m_jChartChoicesComboBox.setPreferredSize(new Dimension( (int) (iScreenWidth *
        0.33),
        m_jChartChoicesComboBox.getPreferredSize().height));

    //Set the renderers AFTER we've set the preferred size
    m_jFileChoicesComboBox.setRenderer(new ComboBoxRenderer(
        m_jFileChoicesComboBox));
    m_jChartChoicesComboBox.setRenderer(new ComboBoxRenderer(
        m_jChartChoicesComboBox));

    jDataVisualizerPanel.add(jOutputFilesLabel);
    jDataVisualizerPanel.add(m_jFileChoicesComboBox, null);
    jDataVisualizerPanel.add(jChartChoicesLabel);
    jDataVisualizerPanel.add(m_jChartChoicesComboBox, null);
    jDataVisualizerPanel.add(jDrawChartButton);

    //****************************************************************
     //Build the run controls button panel
     //****************************************************************
    JPanel jButtonPanel = new JPanel(new FlowLayout(FlowLayout.LEFT));
    jButtonPanel.setName("run_button_panel");
    m_jModelRunButton.setActionCommand("ModelRun");
    m_jModelRunButton.addActionListener(this);
    m_jModelRunButton.setToolTipText("Run model");
    m_jModelRunButton.setDisabledIcon(new ModelIcon(15, 15,
        ModelIcon.RIGHT_TRIANGLE, Color.GRAY));
    m_jModelStopButton.setActionCommand("ModelStop");
    m_jModelStopButton.addActionListener(this);
    m_jModelStopButton.setToolTipText("Quit model run");
    m_jModelStopButton.setDisabledIcon(new ModelIcon(15, 15,
        ModelIcon.RECTANGLE, Color.GRAY));
    m_jModelPauseButton.setActionCommand("ModelPause");
    m_jModelPauseButton.addActionListener(this);
    m_jModelPauseButton.setToolTipText("Pause model run");
    m_jModelPauseButton.setDisabledIcon(new ModelIcon(15, 15, ModelIcon.PAUSE,
        Color.GRAY));
    m_jModelStepForwardButton.setActionCommand("ModelStepForward");
    m_jModelStepForwardButton.addActionListener(this);
    m_jModelStepForwardButton.setToolTipText("Run model one timestep");
    m_jModelStepForwardButton.setDisabledIcon(new ModelIcon(15, 15,
        ModelIcon.STEP_FORWARD, Color.GRAY));
    m_jModelLoadOutputButton.setActionCommand("OpenRunOutput");
    m_jModelLoadOutputButton.addActionListener(this);
    m_jModelLoadOutputButton.setFont(new ModelFont());
    m_jModelLoadOutputButton.setToolTipText("View the data for this run");
    //Make all the buttons the same height
    int iButtonHeight = (int) m_jModelLoadOutputButton.getPreferredSize().
        getHeight();
    m_jModelRunButton.setPreferredSize(new Dimension( (int) m_jModelRunButton.
        getPreferredSize().getWidth(), iButtonHeight));
    m_jModelStopButton.setPreferredSize(new Dimension( (int) m_jModelStopButton.
        getPreferredSize().getWidth(), iButtonHeight));
    m_jModelPauseButton.setPreferredSize(new Dimension( (int)
        m_jModelPauseButton.getPreferredSize().getWidth(), iButtonHeight));
    m_jModelStepForwardButton.setPreferredSize(new Dimension( (int)
        m_jModelStepForwardButton.getPreferredSize().getWidth(), iButtonHeight));
    jButtonPanel.add(m_jModelStopButton);
    jButtonPanel.add(m_jModelPauseButton);
    jButtonPanel.add(m_jModelRunButton);
    jButtonPanel.add(m_jModelStepForwardButton);
    jButtonPanel.add(m_jModelLoadOutputButton);
    jButtonPanel.setBorder(BorderFactory.createMatteBorder(0, 0, 1, 0,
        java.awt.Color.BLACK));

    //Package together the data visualizer and buttons panels
    JPanel jTempPanel = new JPanel(new BorderLayout());
    jTempPanel.setName("north_panel");
    jTempPanel.add(jButtonPanel, BorderLayout.NORTH);
    jTempPanel.add(jDataVisualizerPanel, BorderLayout.CENTER);

    //Add it to the main window
    getContentPane().setLayout(new BorderLayout());
    getContentPane().add(jTempPanel, BorderLayout.NORTH);
    getContentPane().add(m_jScroller, BorderLayout.CENTER);
    getContentPane().add(jMainPanel, BorderLayout.SOUTH);
  }

  /**
   * Draws a chart upon selection in the data visualizer panel.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  void DoDrawChart() {
    try {

      //Get the selected string from the file combo box
      String sFileChoice = (String) m_jFileChoicesComboBox.getSelectedItem();
      if (null == sFileChoice || sFileChoice.startsWith("---")) {
        throw (new ModelException(ErrorGUI.BAD_COMMAND, "JAVA",
                                  "Please select a file for which to draw a chart."));
      }

      //Get the selected string from the chart combo box
      String sChartChoice = (String) m_jChartChoicesComboBox.getSelectedItem();
      if (null == sChartChoice || sChartChoice.startsWith("---")) {
        throw (new ModelException(ErrorGUI.BAD_COMMAND, "JAVA",
                                  "Please select a chart to draw."));
      }
      this.setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));

      //If the run is currently running (thus signalling real-time data
      //visualization), and the user didn't pause the run before opening this
      //chart, and this is the first chart to be opened, we want to pause the
      //run and then re-start so the flags will be set to keep the charts
      //updated
      if (m_bViewingRunOutput == true &&
          m_iState == MainWindowStateSetter.RUNNING
          && m_oDataVisualizerManager.AreCurrentRunChartsOpen() == false) {
        DoModelPause(false);
        DoModelRun(0);
      }
      m_oDataVisualizerManager.DrawChart(sFileChoice, sChartChoice);

    }
    catch (ModelException oErr) {
      ErrorGUI oHandler = new ErrorGUI(this);
      oHandler.WriteErrorMessage(oErr);
    }
    finally {
      this.setCursor(Cursor.getPredefinedCursor(Cursor.DEFAULT_CURSOR));
    }
  }

  /**
   * Gets the GUI manager.
   * @return The GUI manager.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public GUIManager GetDataManager() {
    return m_oDataManager;
  }

  /**
   * File | Exit action performed.  This prompts a save if the user has
   * changed the parameter file, and causes the data visualizer manager to
   * perform cleanup operations.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  private void DoFileExit() {

    m_oDataVisualizerManager.CleanUp();

    //If we have a currently running model, kill it
    if (m_oInterface != null && m_oInterface.IsDone() == false) {
      m_oInterface.KillProcess();
    }

    //See if they'd like to save any existing files
    //Trigger a save if an existing file was being edited
    if (m_oDataManager.HasParameterFileBeenModified()) {
      int iReturnVal = new JOptionPane().showConfirmDialog(this,
          "Do you wish to save a copy of the current file?", "Model",
          JOptionPane.YES_NO_OPTION);

      if (iReturnVal == JOptionPane.YES_OPTION) {
        DoFileSave();
        if (m_oDataManager.HasParameterFileBeenModified()) { //they may have cancelled out
          return;
        }
      }
    }

    System.exit(0);
  }

  /**
   * File | New Parameter File action performed
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  private void DoFileNew() {
    try {
      m_oDataManager.CreateNewParameterFile();
    }
    catch (ModelException oErr) {
      ErrorGUI oHandler = new ErrorGUI(this);
      oHandler.WriteErrorMessage(oErr);
    }
  }

  /**
   * File | New Batch File action performed
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  private void DoFileBatchNew() {
    BatchSetup oWindow = new BatchSetup(this, m_oDataManager);
    oWindow.pack();
    oWindow.setLocationRelativeTo(null);
    oWindow.setVisible(true);
  }

  /**
   * Responds to the click of the button for viewing the current run's output.
   * If there is no parameter file loaded, this tells the user that it can't
   * do the requested operation.  If there's no output to load, this tells the
   * user that.  Otherwise, this will load all currently created output files
   * for the current run.
   * @throws ModelException if there is a problem opening the output.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>November 18, 2004:  Created (LEM)
   */
  private void DoViewRunOutput() throws ModelException {

    try {
      //Check the state - if there's no parameter file loaded, just tell the
      //user there's nothing to do
      if (m_iState == MainWindowStateSetter.NO_PAR_FILE) {
        JOptionPane.showMessageDialog(this,
                                      "You must load a parameter file and run " +
                                      "it to view its data.", "Model",
                                      JOptionPane.ERROR_MESSAGE);
        return;
      }

      //Get the names of the output files to load
      String sDetailedOutputFile = m_oDataManager.GetOutputBehaviors().
          GetDetailedOutputFileName();
      String sShortOutFile = m_oDataManager.GetOutputBehaviors().
          GetShortOutputFileName();

      //If there are no output files set up, warn the user and exit
      if ( (sDetailedOutputFile == null || sDetailedOutputFile.length() == 0) &&
          (sShortOutFile == null || sShortOutFile.length() == 0)) {
        //If the run's already going, tell them they have to stop it
        if (m_iState == MainWindowStateSetter.RUNNING ||
            m_iState == MainWindowStateSetter.PAUSED) {

          JOptionPane.showMessageDialog(this, "The data viewing reads from " +
                                        "output files.\nYou must stop the run " +
                                        "and add output files to your " +
                                        "parameter file to use this option.",
                                        "Model", JOptionPane.ERROR_MESSAGE);
          return;
        }
        else {
          JOptionPane.showMessageDialog(this,
                                        "The data viewing reads from output " +
                                        "files.  You must set up output files to " +
                                        "use this option.", "Model",
                                        JOptionPane.ERROR_MESSAGE);
          return;
        }
      }

      //If the run hasn't started, tell them to run a timestep and try again
      if (m_iState == MainWindowStateSetter.PAR_FILE_LOADED) {
        JOptionPane.showMessageDialog(this, "Run the run for at least one " +
                                      "timestep in order to create output to " +
                                      "view.\nIt is safest to open charts when " +
                                      "the model is paused.", "Model",
                                      JOptionPane.INFORMATION_MESSAGE);
        return;
      }

      //If the state is running, tell them to pause before opening charts (but
      //this time don't return)
      if (m_iState == MainWindowStateSetter.RUNNING) {
        JOptionPane.showMessageDialog(this, "It is recommended that you pause " +
                                      "the model before opening new charts.",
                                      "Model", JOptionPane.INFORMATION_MESSAGE);
      }

      //Add the output files to the data visualization manager and set our
      //flag
      m_bViewingRunOutput = true;
      if (sDetailedOutputFile != null && sDetailedOutputFile.length() > 0) {
        m_oDataVisualizerManager.AddCurrentRunFile(sDetailedOutputFile);
        //Add this file to the file choices combo box
        if (m_jFileChoicesComboBox.getItemAt(0).equals("---Load data file---")) {
          m_jFileChoicesComboBox.removeItemAt(0);
        }
        m_jFileChoicesComboBox.insertItemAt(sDetailedOutputFile, 0);
        m_jFileChoicesComboBox.setSelectedIndex(0);

      }
      if (sShortOutFile != null && sShortOutFile.length() > 0) {
        m_oDataVisualizerManager.AddCurrentRunFile(sShortOutFile);
        //Add this file to the file choices combo box
        if (m_jFileChoicesComboBox.getItemAt(0).equals("---Load data file---")) {
          m_jFileChoicesComboBox.removeItemAt(0);
        }
        m_jFileChoicesComboBox.insertItemAt(sShortOutFile, 0);
        m_jFileChoicesComboBox.setSelectedIndex(0);
      }
    }
    catch (ModelException oErr) {
      //If there is a problem, reset the flag and rethrow the error
      m_bViewingRunOutput = false;
      throw (oErr);
    }

  }

  /**
   * File | Open Parameter File action performed.  Allows the user to input a
   * parameter file of either the old or new type.  Any selected file is
   * passed to the GUI manager for processing.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   * <br>November 2, 2004:  Added a message to the message bar (LEM)
   * <br>June 7, 2005:  Removed old parameter file support (LEM)
   */
  private void DoFileOpen() {
    try {
      boolean bFileRead = false;
      JFileChooser chooser = new JFileChooser();

      //Set up our file if needed
      if (m_oFileDirectory.getName().length() == 0) {
        m_oFileDirectory = chooser.getCurrentDirectory();
      }

      chooser.setCurrentDirectory(m_oFileDirectory);

      chooser.setFileFilter(new FileOpenFilter());
      int returnVal = chooser.showOpenDialog(this);
      if (returnVal == JFileChooser.APPROVE_OPTION) {
        //User chose a file - what kind?
        File oFile = chooser.getSelectedFile();
        String sFileName = oFile.getAbsolutePath();
        String sDisplayFile = oFile.getName(); //name we'll display to user

        this.setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));

        if (sFileName.endsWith(".xml")) {

          //It's a plaintext XML file
          bFileRead = m_oDataManager.InputXMLFile(sFileName, this);
        }
        else if (sFileName.endsWith(".gz.tar")) {

          //It's a detailed output file - display for the user the filenames to
          //choose from
          String[] p_sTarFiles = Tarball.GetTarballEntries(sFileName);
          if (p_sTarFiles != null) {
            File oDetailedOutputFile = null;
            try {
              String sTarName = (String) JOptionPane.showInputDialog(
                  this, //component in which to display
                  "Choose the file to add", //user message
                  "SORTIE", //window title
                  JOptionPane.INFORMATION_MESSAGE, //message type
                  null, //icon
                  p_sTarFiles, //list of possibilities to display
                  p_sTarFiles[0]); //initial selection

              if (sTarName == null || sTarName.length() == 0) {
                return;
              }

              String sFile = Tarball.ExtractTarballFile(sFileName, sTarName);
              if (sFile != null) {
                m_oDataManager.InputXMLFile(sFile, this);
                //Erase the file - it's a temp
                oDetailedOutputFile = new File(sFile);
                bFileRead = true;
                sDisplayFile = sFile;
              }
            }
            finally {
              if (oDetailedOutputFile != null) {
                //Make sure the file gets deleted, even in the event of
                //problems
                oDetailedOutputFile.delete();
              }
            }

          }
          else {
            //Invalid file - throw an error
            ModelException oErr = new ModelException();
            oErr.sMessage = "This file is empty.";
            oErr.sCallingFunction = "JAVA";
            throw (oErr);
          }
        }
        else if (sFileName.endsWith(".txt")) {
          //Assume this is a tree map file

          //Make sure a parameter file has been loaded
          if (m_iState == MainWindowStateSetter.PAR_FILE_LOADED) {
            m_oDataManager.InputTreeMap(sFileName);
            m_jMessagesField.setText("Added tree map " + sFileName);
          }
          else {
            throw (new ModelException(ErrorGUI.BAD_COMMAND, "JAVA",
                                      "You must load a parameter file before loading a file of this type."));
          }
        }
        else if (sFileName.endsWith(".hvr")) {

          //Make sure a parameter file has been loaded
          if (m_iState == MainWindowStateSetter.PAR_FILE_LOADED) {
            m_oDataManager.InputHarvestRegime(sFileName);
            m_jMessagesField.setText("Added harvest regime " + sFileName);
          }
          else {
            throw (new ModelException(ErrorGUI.BAD_COMMAND, "JAVA",
                                      "You must load a parameter file before loading a file of this type."));
          }
        }
        else {
          //This file is not a recognized type
          throw (new ModelException(ErrorGUI.BAD_FILE, "JAVA", "This is not a " +
                                    "recognized file type."));
        }

        //Grab the directory so we can go there again
        m_oFileDirectory = oFile.getParentFile();

        if (bFileRead) {

          //Add a message to the message bar
          m_jMessagesField.setText("Opened file " + sDisplayFile);

          //Display the last 40 characters
          String sParFileName = m_oDataManager.GetParameterFileName();
          if (sParFileName.length() > 40) {
            sParFileName = "..." +
                sParFileName.substring(sParFileName.length() - 40);
          }
          m_jParameterFileField.setText(sParFileName);

          //If we're coming from a no-file-loaded state, switch to file loaded
          if (m_iState == MainWindowStateSetter.NO_PAR_FILE) {
            SetModelState(MainWindowStateSetter.PAR_FILE_LOADED);
          }

          //Reset all real-time output visualization flags
          m_bKeepRunning = false;
          m_bViewingRunOutput = false;
        }
      }
    }
    catch (ModelException oErr) {
      ErrorGUI oHandler = new ErrorGUI(this);
      oHandler.WriteErrorMessage(oErr);
    }
    finally {
      this.setCursor(Cursor.getPredefinedCursor(Cursor.DEFAULT_CURSOR));
    }
  }

  /**
   * Model | Run Batch action performed.  Allows the user to input a
   * batch file and run it.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>March 31, 2005:  Created (LEM)
   */
  private void DoModelRunBatch() {
    try {
      JFileChooser jChooser = new JFileChooser();

      //Set up our file if needed
      if (m_oFileDirectory.getName().length() == 0) {
        m_oFileDirectory = jChooser.getCurrentDirectory();
      }

      jChooser.setCurrentDirectory(m_oFileDirectory);

      jChooser.setFileFilter(new XMLFileFilter());
      int returnVal = jChooser.showOpenDialog(this);
      if (returnVal == JFileChooser.APPROVE_OPTION) {

        File oFile = jChooser.getSelectedFile();
        String sFileName = oFile.getAbsolutePath();

        //Grab the directory so we can go there again
        m_oFileDirectory = oFile.getParentFile();
        //Display the last 40 characters
        String sParFileName = m_oDataManager.GetParameterFileName();
        if (sParFileName.length() > 40) {
          sParFileName = "..." +
              sParFileName.substring(sParFileName.length() - 40);
        }
        m_jParameterFileField.setText(sParFileName);

        //Reset all real-time output visualization flags
        m_bKeepRunning = false;
        m_bViewingRunOutput = false;

        //If there's a run in progress, do nothing
        if (m_oInterface != null && m_oInterface.IsDone() == false &&
            m_oInterface.IsPaused() == false) {
          return;
        }

        if (m_oInterface == null) {
          m_oInterface = null;
          m_oInterface = new Interface();
          //Close all open charts
          m_oDataVisualizerManager.CloseCurrentRunCharts(this);
        }
        try {

          this.setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));
          SetModelState(MainWindowStateSetter.RUNNING);
          //Disable output visualization and pausing
          m_jModelLoadOutputButton.setEnabled(false);
          m_jModelPauseButton.setEnabled(false);
          m_jMenuModelPause.setEnabled(false);

          //Create a timer.  This will put the model run in another thread,
          //which allows the GUI to update during the run, meaning that the
          //run's messages will be displayed.  This timer will trigger once
          //every second.
          m_oTimer = new Timer(1000, new RunListener(this));

          m_oInterface.Run(this, sFileName, 0);
          m_oTimer.start();
        }
        finally {
          this.setCursor(Cursor.getPredefinedCursor(Cursor.DEFAULT_CURSOR));
        }
      }
    }
    finally {
      this.setCursor(Cursor.getPredefinedCursor(Cursor.DEFAULT_CURSOR));
    }
  }

  /**
   * Opens an output file.  Any file that was chosen by the user gets passed
   * to the Data Visualization Manager for processing.
   * @throws ModelException if anything goes wrong with the data visualization.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  private void DoFileOpenData() throws ModelException {
    try {
      JFileChooser jChooser = new JFileChooser();
      jChooser.setFileFilter(new datavisualizer.DataFileFilter());

      //Set up our file if needed
      if (m_oFileDirectory.getName().length() == 0) {
        m_oFileDirectory = jChooser.getCurrentDirectory();
      }

      jChooser.setCurrentDirectory(m_oFileDirectory);

      int iReturnVal = jChooser.showOpenDialog(this);
      if (iReturnVal == JFileChooser.APPROVE_OPTION) {

        this.setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));

        //Pass this file to the output manager
        File jFile = jChooser.getSelectedFile();
        String sFileName = jFile.getAbsolutePath();
        m_oDataVisualizerManager.AddFile(sFileName);

        m_oFileDirectory = jFile.getParentFile();

        //Add this file to the file choices combo box
        if (m_jFileChoicesComboBox.getItemAt(0).equals("---Load data file---")) {
          m_jFileChoicesComboBox.removeItemAt(0);
        }
        m_jFileChoicesComboBox.insertItemAt(sFileName, 0);
        m_jFileChoicesComboBox.setSelectedIndex(0);
      }
    }
    finally {
      this.setCursor(Cursor.getPredefinedCursor(Cursor.DEFAULT_CURSOR));
    }
  }

  /**
   * Closes an open output file.  All open windows related to this file are
   * closed as well.
   * <br>Edit history:
   * <br>------------------
   * <br>July 30, 2004: Created (LEM)
   */
  protected void DoFileCloseData() {

    //Construct a window listing the files to close
    //List of files, with an "all" option if there are at least two
    String[] p_sFilesToClose = null;
    String sName;
    int i,
        iNumFiles = m_jFileChoicesComboBox.getItemCount();

    for (i = 0; i < iNumFiles; i++) {
      sName = (String) m_jFileChoicesComboBox.getItemAt(i);
      if (sName.startsWith("--")) {
        iNumFiles--;
      }
    }

    if (iNumFiles == 0) {
      return;
    }

    p_sFilesToClose = new String[iNumFiles];
    for (i = 0; i < iNumFiles; i++) {
      sName = (String) m_jFileChoicesComboBox.getItemAt(i);
      if (!sName.startsWith("--")) {
        p_sFilesToClose[i] = (String) m_jFileChoicesComboBox.getItemAt(i);
      }
    }

    //Find out what the user wants to close
    String sFileName = (String) JOptionPane.showInputDialog(
        this, //component in which to display
        "Choose the file to close", //user message
        "SORTIE", //window title
        JOptionPane.INFORMATION_MESSAGE, //message type
        null, //icon
        p_sFilesToClose, //list of possibilities to display
        p_sFilesToClose[0]); //initial selection

    if (sFileName == null || sFileName.length() == 0) {
      return;
    }

    CloseOneDataFile(sFileName);

  }

  /**
   * Closes a single output file and manages the file list.
   * @param sFileName String File to close.
   */
  protected void CloseOneDataFile(String sFileName) {

    //Ask the data visualization manager to close the file
    m_oDataVisualizerManager.CloseFile(sFileName);

    int i,
        iNumFiles = m_jFileChoicesComboBox.getItemCount();

    //Remove the choice from the combo file
    for (i = 0; i < iNumFiles; i++) {
      if (sFileName.equals( (String) m_jFileChoicesComboBox.getItemAt(i))) {
        m_jFileChoicesComboBox.removeItemAt(i);
        break;
      }
    }
    if (m_jFileChoicesComboBox.getItemCount() == 0) {
      m_jFileChoicesComboBox.insertItemAt("---Load data file---", 0);
      m_jFileChoicesComboBox.setSelectedIndex(0);
    }

    //If we were doing real-time visualization, check to see if this closes
    //all files for the current run; if so, turn off real-time visualization
    if (m_bViewingRunOutput) {

      //If there are no files open, turn off real-time visualization
      if (m_jFileChoicesComboBox.getItemCount() == 1 &&
          ( (String) m_jFileChoicesComboBox.getItemAt(0)).startsWith("---")) {
        m_bKeepRunning = true; //tell the run to keep going from now on
        m_bViewingRunOutput = false;
        return;
      }

      //Get the current run output files and check to see if any of them
      //are still open
      boolean bOpen = false;
      String sDetailedOutputFile = m_oDataManager.GetOutputBehaviors().
          GetDetailedOutputFileName();
      String sShortOutFile = m_oDataManager.GetOutputBehaviors().
          GetShortOutputFileName();
      if (sDetailedOutputFile == null) {
        sDetailedOutputFile = "";
      }
      if (sShortOutFile == null) {
        sShortOutFile = "";
      }
      iNumFiles = m_jFileChoicesComboBox.getItemCount();
      for (i = 0; i < iNumFiles; i++) {
        String sName = (String) m_jFileChoicesComboBox.getItemAt(i);
        if (sDetailedOutputFile.equals(sName) || sShortOutFile.equals(sName)) {
          bOpen = true;
          break;
        }
      }
      if (bOpen == false) {
        //No files from the current run are open - turn off real-time
        //visualization
        m_bKeepRunning = true; //tell the run to keep going from now on
        m_bViewingRunOutput = false;
      }
    }
  }

  /**
   * File | Save Parameter File action performed.  This is a request to save
   * a parameter file.  The request is passed to the GUI manager for
   * processing.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   * <br>November 22, 2005:  Made it so that the save dialog would display the
   * name of the last parameter file
   */
  protected void DoFileSave() {
    try {

      JFileChooser jChooser = new JFileChooser();

      //Set up our file if needed
      if (m_oFileDirectory.getName().length() == 0) {
        m_oFileDirectory = jChooser.getCurrentDirectory();
      }

      jChooser.setCurrentDirectory(m_oFileDirectory);

      jChooser.setSelectedFile(new File(m_oDataManager.GetParameterFileName()));
      jChooser.setFileFilter(new XMLFileFilter());

      int returnVal = jChooser.showSaveDialog(this);
      if (returnVal == JFileChooser.APPROVE_OPTION) {

        //User chose a file - does it already exist?
        File oFile = jChooser.getSelectedFile();
        if (oFile.exists()) {
          JOptionPane pane = new JOptionPane();
          returnVal = pane.showConfirmDialog(this,
                                             "Do you wish to overwrite the existing file?",
                                             "Model",
                                             JOptionPane.YES_NO_OPTION);
        }
        if (!oFile.exists() || returnVal == JOptionPane.YES_OPTION) {
          String sFileName = oFile.getAbsolutePath();
          if (!sFileName.endsWith(".xml")) {
            sFileName = sFileName.concat(".xml");
          }
          //Grab the directory so we can go there again
          m_oFileDirectory = oFile.getParentFile();

          if (m_oDataManager.WriteParameterFile(sFileName)) {

            m_jMessagesField.setText("Successfully saved file " + sFileName);

            String sParFileName = m_oDataManager.GetParameterFileName();
            if (sParFileName.length() > 40) {
              sParFileName = "..." +
                  sParFileName.substring(sParFileName.length() - 40);
            }
            m_jParameterFileField.setText(sParFileName);
          }
        }
      }
    }
    catch (ModelException oErr) {
      ErrorGUI oHandler = new ErrorGUI(this);
      oHandler.WriteErrorMessage(oErr);
    }
  }

  /**
   * Edit | Edit model flow action performed
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */

  private void DoEditModelFlow() {
    m_oDataManager.DisplayModelFlowWindow();
  }

  /**
   * Edit | Run parameters data action performed
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */

  private void DoEditParameters() {
    m_oDataManager.DisplayRunParameters();
  }

  /**
   * Edit | Edit output options action performed
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */

  private void DoEditOutput() {
    m_oDataManager.DisplayOutputWindow();
  }

  /**
   * Edit | Episodic events action performed
   * @throws ModelException passing through from called methods.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>May 4, 2004: Created (LEM)
   */

  private void DoEditEpisodicEvents() throws ModelException {
    JDialog jWindow = m_oDataManager.DisplayEpisodicEventsWindow();
    jWindow.pack();

    //Resize if too big
    if (jWindow.getSize().height > this.getSize().height ||
        jWindow.getSize().width > this.getSize().width) {

      int iWidth = java.lang.Math.min(jWindow.getSize().width,
                                      this.getSize().width - 10);
      int iHeight = java.lang.Math.min(jWindow.getSize().height,
                                       this.getSize().height - 10);

      jWindow.setBounds(jWindow.getBounds().x, jWindow.getBounds().y,
                        iWidth, iHeight);
    }
    jWindow.setLocationRelativeTo(null);
    jWindow.setVisible(true);
  }

  /**
   * Sets the current window state.
   * @param iState State to set to.  Should be one of the choices in
   * MainWindowStateSetter.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void SetModelState(int iState) {
    m_iState = iState;
    MainWindowStateSetter.GoToState(iState, this);
  }

  /**
   * Gets the current window state.
   * @return int Current state.  This matches one of the choices in
   * MainWindowStateSetter.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public int GetModelState() {
    return m_iState;
  }

  /**
   * Model | Pause action performed.  Pauses the model.
   * @throws ModelException passing through from called methods.
   * @param bAlert Whether or not to alert the user to the pause.
   */
  private void DoModelPause(boolean bAlert) throws ModelException {
    if (m_oInterface != null && m_oInterface.IsDone() == false) {
      m_bKeepRunning = false;
      m_oInterface.Pause();
      SetModelState(MainWindowStateSetter.PAUSED);
      if (bAlert) {
        JOptionPane.showMessageDialog(this,
                                      "The model will pause at the end of the current timestep.\n  Watch the messages panel at the bottom of the screen.",
                                      "Model", JOptionPane.PLAIN_MESSAGE);
      }
    }
  }

  /**
   * Model | Stop Run action performed.  Stops the currently executing run.
   * @throws ModelException passed through from called functions.
   */
  private void DoModelStop() throws ModelException {
    if (m_oInterface != null) {
      if (m_iState == MainWindowStateSetter.RUNNING) {
        JOptionPane.showMessageDialog(this,
                                      "The model will stop at the end of the current timestep.");
      }
      m_oInterface.Stop();
      //Get any messages from the model run and display them
      String sMessage = m_oInterface.GetMessage();
      if (sMessage != null && sMessage.length() > 0) {
        m_jMessagesField.setText(sMessage);
      }
    }
  }

  /**
   * Model | Run action performed.  This runs the model.  The
   * request is passed off to the Interface, which handles the communications
   * with the C++ core.
   *
   * If there is an existing Interface object whose run has been paused, then
   * this will pass it the run command and let it continue.  If the model is
   * already running, this does nothing.
   *
   * @param iNumStepsToRun Number of timesteps to run the model.  Set to 0 if
   * the model should run without interruption.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   * <br>November 8, 2004:  Added run after pause feature (LEM)
   */
  private void DoModelRun(int iNumStepsToRun) {
    //If there's a run in progress, do nothing
    if (m_oInterface != null && m_oInterface.IsDone() == false &&
        m_oInterface.IsPaused() == false) {
      return;
    }

    //If the number of timesteps is 0 and we're doing real-time data
    //visualization, set the flag that keeps the run going after data
    //updates
    if (m_bViewingRunOutput && iNumStepsToRun == 0) {
      m_bKeepRunning = true;
    }
    else {
      m_bKeepRunning = false;
    }

    //First check to see if there's a run that's paused - don't interrupt if
    //there is
    boolean bWasPaused = false;
    if (m_oInterface != null) {
      bWasPaused = m_oInterface.IsPaused();

    }
    if (m_oInterface == null || bWasPaused == false) {
      m_oInterface = null;
      m_oInterface = new Interface();
      //Close all open charts
      m_oDataVisualizerManager.CloseCurrentRunCharts(this);
    }
    try {
      if (m_iState == MainWindowStateSetter.NO_PAR_FILE) {
        //Error message
        ModelException oErr = new ModelException();
        oErr.iErrorCode = ErrorGUI.MODEL_NOT_READY;
        oErr.sMessage = "The model is not ready to run.";
        oErr.sCallingFunction = "JAVA";
        throw (oErr);
      }
      //Get a parameter file name to load
      this.setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));
      final String sFileName = m_oDataManager.PrepToRun();
      SetModelState(MainWindowStateSetter.RUNNING);

      //Create a timer.  This will put the model run in another thread,
      //which allows the GUI to update during the run, meaning that the
      //run's messages will be displayed.  This timer will trigger once
      //every second.
      if (bWasPaused == false) {
        m_oTimer = new Timer(1000, new RunListener(this));
      }

      //If we're viewing output, ignore the actual number of timesteps and run
      //for one timestep.  The timer code will catch the pause, update the
      //charts, and run it some more if necessary.
      if (m_bViewingRunOutput) {
        m_oInterface.Run(this, sFileName, 1);
      }
      else {
        m_oInterface.Run(this, sFileName, iNumStepsToRun);
      }

      if (bWasPaused == false) {
        m_oTimer.start();
      }
    }
    catch (ModelException oErr) {
      //If there's a process, kill it and set the interface back to null
      if (m_oInterface != null) {
        m_oInterface.KillProcess();
        m_oInterface = null;
      }
      ErrorGUI oHandler = new ErrorGUI(this);
      oHandler.WriteErrorMessage(oErr);
      SetModelState(MainWindowStateSetter.PAR_FILE_LOADED);
    }
    finally {
      this.setCursor(Cursor.getPredefinedCursor(Cursor.DEFAULT_CURSOR));
    }
  }

  /**
   * Changes the cursor to the default.  This is a separate
   * function so it can be called from the private inner class created in
   * DoModelRun().
   */
  protected void SetCursorToDefault() {
    this.setCursor(Cursor.getPredefinedCursor(Cursor.DEFAULT_CURSOR));
  }

  /**
   * Manages window events.
   * @param oEvent Event triggered.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void actionPerformed(ActionEvent oEvent) {
    try {
      String sCommandString = oEvent.getActionCommand();
      if (sCommandString.equals("FileExit")) {
        DoFileExit();
      }
      else if (sCommandString.equals("FileOpen")) {
        DoFileOpen();
      }
      else if (sCommandString.equals("FileNew")) {
        DoFileNew();
      }
      else if (sCommandString.equals("FileOpenData")) {
        DoFileOpenData();
      }
      else if (sCommandString.equals("FileCloseData")) {
        DoFileCloseData();
      }
      else if (sCommandString.equals("FileSave")) {
        DoFileSave();
      }
      else if (sCommandString.equals("EditTree")) {
        DoEditTree();
      }
      else if (sCommandString.equals("EditGrid")) {
        DoEditGrid();
      }
      else if (sCommandString.equals("FileNewBatch")) {
        DoFileBatchNew();
      }
      else if (sCommandString.equals("ModelRun")) {
        DoModelRun(0);
      }
      else if (sCommandString.equals("ModelRunBatch")) {
        DoModelRunBatch();
      }
      else if (sCommandString.equals("ModelStepForward")) {
        DoModelRun(1);
      }
      else if (sCommandString.equals("ModelPause")) {
        DoModelPause(true);
      }
      else if (sCommandString.equals("ModelStop")) {
        DoModelStop();
      }
      else if (sCommandString.equals("EditFlow")) {
        DoEditModelFlow();
      }
      else if (sCommandString.equals("EditOutput")) {
        DoEditOutput();
      }
      else if (sCommandString.equals("EditParameters")) {
        DoEditParameters();
      }
      else if (sCommandString.equals("EditDisturbance")) {
        DoEditEpisodicEvents();
      }
      else if (sCommandString.equals("DrawChart")) {
        DoDrawChart();
      }
      else if (sCommandString.equals("OpenRunOutput")) {
        DoViewRunOutput();
      }
      else if (oEvent.getSource().equals(m_jFileChoicesComboBox)) {
        UpdateChartChoices();
      }
    }
    catch (ModelException oErr) {
      ErrorGUI oHandler = new ErrorGUI(this);
      oHandler.WriteErrorMessage(oErr);
    }
    catch (Error oErr) {
      ModelException oExp = new ModelException();
      oExp.iErrorCode = ErrorGUI.UNKNOWN;
      String sMessage = oErr.getMessage();
      if (oErr.getClass() != null) {
        sMessage += ", " + oErr.getClass();
      }
      if (oErr.getCause() != null) {
        sMessage += ", " + oErr.getCause().getMessage();
      }
      oExp.sMessage = sMessage;
      ErrorGUI oHandler = new ErrorGUI(this);
      oHandler.WriteErrorMessage(oExp);
    }
    catch (Exception oErr) {
      ModelException oExp = new ModelException();
      oExp.iErrorCode = ErrorGUI.UNKNOWN;
      oExp.sMessage = oErr.getMessage() + oErr.getClass();
      ErrorGUI oHandler = new ErrorGUI(this);
      oHandler.WriteErrorMessage(oExp);
    }
  }

  /**
   * Displays the Tree editing window.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  private void DoEditTree() {
    m_oDataManager.DisplayTreeSetupWindow();
  }

  /**
   * Displays the Grid editing window.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  private void DoEditGrid() {
    m_oDataManager.DisplayGridSetupWindow();
  }

  /**
   * Updates the chart choices in the data visualization panel according to the
   * file which is selected in the file list.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void UpdateChartChoices() {
    try {
      //Get the data options for this file to display
      String sFileName = (String) m_jFileChoicesComboBox.getSelectedItem();
      m_jChartChoicesComboBox.removeAllItems();

      if (sFileName == null || sFileName.length() == 0 ||
          sFileName.startsWith("---")) {
        m_jChartChoicesComboBox.addItem("---Load data file---");
        return;
      }
      String[] p_sOptions = m_oDataVisualizerManager.GetChartOptionsForFile(
          sFileName);

      for (int i = 0; i < p_sOptions.length; i++) {
        m_jChartChoicesComboBox.addItem(p_sOptions[i]);
      }
    }
    catch (ModelException e) {
      ErrorGUI oHandler = new ErrorGUI(this);
      oHandler.WriteErrorMessage(e);
    }
  }

//Overridden so we can exit when window is closed
  protected void processWindowEvent(WindowEvent e) {
    super.processWindowEvent(e);
    if (e.getID() == WindowEvent.WINDOW_CLOSING) {
      DoFileExit();
    }
  }

  /**
   * Runs the window's timer code while the C++ core is running.  This code
   * will execute at specified intervals to accept messages from the C++
   * core.
   * <p>Copyright: Copyright (c) Charles D. Canham 2003</p>
   * <p>Company: Institute of Ecosystem Studies</p>
   * @author Lora E. Murphy
   * @version 1.0
   */
  class RunListener
      implements ActionListener {
    MainWindow m_oWindow;
    boolean m_bUpdatedCharts = false;

    public RunListener(MainWindow oWindow) {
      m_oWindow = oWindow;
    }

    //This is the code that will trigger every interval
    public void actionPerformed(ActionEvent evt) {
      try {

        //Get any messages from the model run and display them
        String sMessage = m_oInterface.GetMessage();
        if (sMessage != null && sMessage.length() > 0) {
          //If the message contains the word "paused" assume that the run
          //is paused
          if (sMessage.indexOf("paused") > -1) {
            SetModelState(MainWindowStateSetter.PAUSED);
            m_bUpdatedCharts = false;
          }
          m_jMessagesField.setText(sMessage);
        }

        //If the run is over, reset everything and notify the user
        if (m_oInterface.IsDone()) {
          m_bKeepRunning = false;
          //        if (m_bViewingRunOutput == true) {
          //          m_oDataVisualizerManager.UpdateCurrentRunCharts();
          //        }
          SetCursorToDefault();
          SetModelState(MainWindowStateSetter.PAR_FILE_LOADED);
          m_oTimer.stop();
          if (m_oInterface.IsOK()) {
            JOptionPane.showMessageDialog(m_oWindow, "Run complete.", "SORTIE",
                                          JOptionPane.INFORMATION_MESSAGE);
          }
          return;
        }

        //If we are doing realtime data visualization and the model is
        //paused, update the charts
        if (m_iState == MainWindowStateSetter.PAUSED && m_bViewingRunOutput
            && m_bUpdatedCharts == false) {
          m_oDataVisualizerManager.UpdateCurrentRunCharts();
          m_bUpdatedCharts = true;
        }

        //If we are doing realtime data visualization, the model is paused,
        //and the user wants to keep running timesteps, run the next one
        if (m_iState == MainWindowStateSetter.PAUSED && m_bKeepRunning) {
          if (m_bViewingRunOutput) {
            DoModelRun(1);
            //DoModelRun will reset this since the number of timesteps was 1 -
            //so set it back
            m_bKeepRunning = true;
          }
          else {
            //This situation means that data visualization was just turned OFF,
            //by closing all the output files.  So trigger a regular run
            DoModelRun(0);
          }
        }

      }
      catch (ModelException oErr) {
        ErrorGUI oHandler = new ErrorGUI(m_oWindow);
        oHandler.WriteErrorMessage(oErr);
      }

    }
  }

}

/**
 * This limits values displayed in the file choices combo box to 50 characters
 * (the last 50, so filenames are displayed).
 * <p>Copyright: Copyright (c) Charles D. Canham 2003</p>
 * <p>Company: Institute of Ecosystem Studies</p>
 * @author Lora E. Murphy
 * @version 1.0
 *
 * <br>Edit history:
 * <br>------------------
 * <br>April 28, 2004: Submitted in beta version (LEM)
 * <br>December 15, 2004:  Added tool-tip text and smart sizing (LEM)
 */
class ComboBoxRenderer
    extends JLabel
    implements ListCellRenderer {

  int m_iMaxChars = 50; /**<Max chars to display*/

  /**
   * Constructor.
   * @param oBox JComboBox Combo box to display
   */
  public ComboBoxRenderer(JComboBox oBox) {
    setOpaque(true);
    setHorizontalAlignment(CENTER);
    setVerticalAlignment(CENTER);

    //Figure out the max chars to display.
    //How wide is 10 chars?
    JLabel jTempLabel = new JLabel("XXXXXXXXXX");
    jTempLabel.setFont(new ModelFont());
    m_iMaxChars = (int) ( (10 * oBox.getPreferredSize().getWidth()) /
                         jTempLabel.getPreferredSize().getWidth());
  }

  /*
   * This method displays the file choice, with a max of 50 chars.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public Component getListCellRendererComponent(
      JList list,
      Object value,
      int index,
      boolean isSelected,
      boolean cellHasFocus) {

    if (isSelected) {
      setBackground(list.getSelectionBackground());
      setForeground(list.getSelectionForeground());
    }
    else {
      setBackground(list.getBackground());
      setForeground(list.getForeground());
    }

    //Get the string to display
    String sText = (String) value;
    String sTextToDisplay = sText;

    setToolTipText(sText);

    //Truncate to max chars
    if (sText != null) {
      if (sText.length() > m_iMaxChars) {
        //Take the LAST 50
        sText = sText.substring(sText.length() - m_iMaxChars, sText.length());
        sTextToDisplay = "..." + sText;
      }
    }

    setText(sTextToDisplay);

    return this;
  }
}
