package javawrapper;

/**
 * Manages mortality behaviors and data.
 * <p>Copyright: Copyright (c) 2003 Charles D. Canham</p>
 * <p>Company: Institute of Ecosystem Studies</p>
 * @author Lora E. Murphy
 * @version 1.0
 *
 * <br>Edit history:
 * <br>------------------
 * <br>April 28, 2004: Submitted in beta version (LEM)
 * <br>October 4, 2004:  Added Puerto Rico NCI mortality (LEM)
 * <br>February 9, 2005:  Added the growth-and-resource mortality (LEM)
 * <br>February 23, 2005:  Added competition mortality (LEM)
 * <br>March 8, 2005:  Made weibull snag size classes species-specific (LEM)
 * <br>March 16, 2005:  Upgraded NCI mort to v. 2.0 (LEM)
 * <br>March 17, 2005:  Added the density self thinning behavior (LEM)
 * <br>May 6, 2005:  Added logistic bi-level mortality behavior (LEM)
 * <br>October 31, 2005:  Added stochastic bi-level mortality behavior (LEM)
 * <br>January 26, 2006:  Added height-GLI weibull mortality behavior (LEM)
 */

public class MortalityBehaviors
    extends BehaviorTypeBase {

  /**Mortality at zero growth for each species*/
  protected ModelVector mp_fMortAtZeroGrowth = new ModelVector(
      "Mortality at Zero Growth",
      "mo_mortAtZeroGrowth", "mo_mazgVal",
      0, ModelVector.FLOAT);

  /**Light-dependent mortality for each species*/
  protected ModelVector mp_fLightDependentMortality = new ModelVector(
      "Light-Dependent Mortality",
      "mo_lightDependentMortality",
      "mo_ldmVal", 0,
      ModelVector.FLOAT);

  /**Senescence mortality alpha for each species*/
  protected ModelVector mp_fRandomMortalityAlpha = new ModelVector(
      "Senescence Mortality Alpha",
      "mo_randomMortAlpha",
      "mo_rmaVal", 0,
      ModelVector.FLOAT);

  /**Senescence mortality beta for each species*/
  protected ModelVector mp_fRandomMortalityBeta = new ModelVector(
      "Senescence Mortality Beta",
      "mo_randomMortBeta", "mo_rmbVal",
      0, ModelVector.FLOAT);

  /**Random adult mortality for each species*/
  protected ModelVector mp_fAdultRandomMortality = new ModelVector(
      "Adult Background Mortality Rate", "mo_nonSizeDepMort", "mo_nsdmVal", 0,
      ModelVector.FLOAT);

  /**Random juvenile mortality for each species*/
  protected ModelVector mp_fJuvenileRandomMortality = new ModelVector(
      "Juvenile Background Mortality Rate", "mo_randomJuvenileMortality",
      "mo_rjmVal", 0, ModelVector.FLOAT);

  /**DBH at onset of senescence for each species*/
  protected ModelVector mp_fDbhAtOnsetOfSenescence = new ModelVector(
      "DBH at Onset of Senescence, in cm", "mo_dbhAtOnsetOfSenescence",
      "mo_daoosVal", 0, ModelVector.FLOAT);

  /**Juvenile self-thinning slope for each species*/
  protected ModelVector mp_fJuvenileSelfThinningSlope = new ModelVector(
      "Juvenile Self-Thinning Slope", "mo_juvSelfThinSlope", "mo_jstsVal", 0,
      ModelVector.FLOAT);

  /**Juvenile self-thinning intercept for each species*/
  protected ModelVector mp_fJuvenileSelfThinningIntercept = new ModelVector(
      "Juvenile Self-Thinning Intercept", "mo_juvSelfThinIntercept",
      "mo_jstiVal", 0, ModelVector.FLOAT);

  /**Adult self-thinning slope for each species*/
  protected ModelVector mp_fAdultSelfThinningSlope = new ModelVector(
      "Adult Self-Thinning Slope", "mo_adultSelfThinSlope",
      "mo_astsVal", 0, ModelVector.FLOAT);

  /**Adult self-thinning intercept for each species*/
  protected ModelVector mp_fAdultSelfThinningIntercept = new ModelVector(
      "Adult Self-Thinning Intercept", "mo_adultSelfThinIntercept",
      "mo_astiVal", 0, ModelVector.FLOAT);

  /**Adult self-thinning DBH for each species*/
  protected ModelVector mp_fAdultSelfThinningMaxDbh = new ModelVector(
      "Maximum DBH for Adult Self-Thinning", "mo_adultSelfThinMaxDbh",
      "mo_astmdVal", 0, ModelVector.FLOAT);

  /**Weibull snag mortality - snag size class 1 "a" parameter*/
  protected ModelVector mp_fSnag1WeibullA = new ModelVector(
      "Weibull Annual \"a\" Parameter for Snag Size Class 1 Mortality",
      "mo_snag1WeibullA", "mo_s1waVal", 0, ModelVector.FLOAT);

  /**Weibull snag mortality - snag size class 1 "a" parameter*/
  protected ModelVector mp_fSnag2WeibullA = new ModelVector(
      "Weibull Annual \"a\" Parameter for Snag Size Class 2 Mortality",
      "mo_snag2WeibullA", "mo_s2waVal", 0, ModelVector.FLOAT);

  /**Weibull snag mortality - snag size class 3 "a" parameter*/
  protected ModelVector mp_fSnag3WeibullA = new ModelVector(
      "Weibull Annual \"a\" Parameter for Snag Size Class 3 Mortality",
      "mo_snag3WeibullA", "mo_s3waVal", 0, ModelVector.FLOAT);

  /**Weibull snag mortality - snag size class 1 "b" parameter*/
  protected ModelVector mp_fSnag1WeibullB = new ModelVector(
      "Weibull Annual \"b\" Parameter for Snag Size Class 1 Mortality",
      "mo_snag1WeibullB", "mo_s1wbVal", 0, ModelVector.FLOAT);

  /**Weibull snag mortality - snag size class 2 "b" parameter*/
  protected ModelVector mp_fSnag2WeibullB = new ModelVector(
      "Weibull Annual \"b\" Parameter for Snag Size Class 2 Mortality",
      "mo_snag2WeibullB", "mo_s2wbVal", 0, ModelVector.FLOAT);

  /**Weibull snag mortality - snag size class 3 "b" parameter*/
  protected ModelVector mp_fSnag3WeibullB = new ModelVector(
      "Weibull Annual \"b\" Parameter for Snag Size Class 3 Mortality",
      "mo_snag3WeibullB", "mo_s3wbVal", 0, ModelVector.FLOAT);

  /**Weibull snag mortality - snag size class 1 upper DBH value*/
  protected ModelVector mp_fSnagSizeClass1Dbh = new ModelVector(
      "Weibull Upper DBH of Snag Size Class 1",
      "mo_snagSizeClass1DBH", "mo_sc1dVal", 0, ModelVector.FLOAT);

  /**Weibull snag mortality - snag size class 2 upper DBH value*/
  protected ModelVector mp_fSnagSizeClass2Dbh = new ModelVector(
      "Weibull Upper DBH of Snag Size Class 2",
      "mo_snagSizeClass2DBH", "mo_sc2dVal", 0, ModelVector.FLOAT);

  /**Weibull snag mortality - snag size class 1 upper DBH value
   * @deprecated This value now has a species-specific version and this one
   * will eventually be eliminated.*/
  protected ModelFloat m_fSnagSizeClass1DbhDep = new ModelFloat(0,
      "Upper DBH of Snag Size Class 1", "mo_snagSizeClass1");

  /**Weibull snag mortality - snag size class 2 upper DBH value
   * @deprecated This value now has a species-specific version and this one
   * will eventually be eliminated.*/
  protected ModelFloat m_fSnagSizeClass2DbhDep = new ModelFloat(0,
      "Upper DBH of Snag Size Class 2", "mo_snagSizeClass2");

  /**DBH at asymptotic maximum mortality - for senescence*/
  protected ModelInt m_iDbhAtAsymptoticMaximumMortality = new ModelInt(0,
      "DBH of Maximum Senescence Mortality Rate, as an integer in cm",
      "mo_dbhAtAsympMaxMort");

  /**NCI Mortality - Maximum crowding radius*/
  protected ModelVector mp_fNCIMaxCrowdingRadius = new ModelVector(
      "NCI Max Radius of Crowding Neighbors, in m",
      "mo_prnciMaxCrowdingRadius", "mo_pmcrVal", 0, ModelVector.FLOAT);

  /**NCI Mortality - Neighbor DBH effect (alpha) parameter*/
  protected ModelVector mp_fNCINeighDBHEffect = new ModelVector(
      "NCI Neighbor DBH Effect (alpha)", "mo_prnciNeighDBHEff",
      "mo_pndeVal", 0, ModelVector.FLOAT);

  /**NCI Mortality - Neighbor distance effect (beta) parameter*/
  protected ModelVector mp_fNCINeighDistanceEffect = new ModelVector(
      "NCI Neighbor Distance Effect (beta)", "mo_prnciNeighDistEff",
      "mo_pndseVal", 0, ModelVector.FLOAT);

  /**NCI minimum DBH for crowding neighbors, for each species; all species
   * required*/
  protected ModelVector mp_fNCIMinNeighborDBH = new ModelVector(
      "NCI Minimum Neighbor DBH, in cm", "mo_nciMinNeighborDBH",
      "mo_nmndVal", 0, ModelVector.FLOAT, true);

  /**NCI Mortality - Shading coefficient (m)*/
  protected ModelVector mp_fNCIShadingEffectCoefficient = new ModelVector(
      "NCI Shading Effect Coefficient (m)", "mo_nciShadingCoefficient",
      "mo_nscVal", 0, ModelVector.FLOAT);

  /**NCI Mortality - Shading exponent (n)*/
  protected ModelVector mp_fNCIShadingEffectExponent = new ModelVector(
      "NCI Shading Effect Exponent (n)", "mo_nciShadingExponent",
      "mo_nseVal", 0, ModelVector.FLOAT);

  /**NCI Mortality - Size sensitivity to NCI parameter (gamma)*/
  protected ModelVector mp_fNCISizeSensToNCI = new ModelVector(
      "NCI Size Sensitivity to NCI (gamma)", "mo_prnciSizeSensNCI",
      "mo_pssnVal", 0, ModelVector.FLOAT);

  /**NCI Mortality - Maximum probability of survival*/
  protected ModelVector mp_fNCIMaxProbSurvival = new ModelVector(
      "NCI Max Survival Probability (0-1)", "mo_prnciMaxPotentialSurvival",
      "mo_pmpsVal", 0, ModelVector.FLOAT);

  /**NCI Mortality - Size effect mode (X0)*/
  protected ModelVector mp_fNCISizeEffectMode = new ModelVector(
      "NCI Size Effect Mode, in cm", "mo_prnciSizeEffectMode",
      "mo_psemVal", 0, ModelVector.FLOAT);

  /**NCI Mortality - Size effect variance (Xb)*/
  protected ModelVector mp_fNCISizeEffectVariance = new ModelVector(
      "NCI Size Effect Variance, in cm", "mo_prnciSizeEffectVar",
      "mo_psevVal", 0, ModelVector.FLOAT);

  /**NCI Mortality - Crowding Effect Slope (C)*/
  protected ModelVector mp_fNCISlope = new ModelVector(
      "NCI Crowding Effect Slope (C)", "mo_prnciNCISlope", "mo_pnslVal", 0,
      ModelVector.FLOAT);

  /**NCI Mortality - Crowding Effect Steepness (D)*/
  protected ModelVector mp_fNCISteepness = new ModelVector(
      "NCI Crowding Effect Steepness (D)", "mo_prnciNCISteepness", "mo_pnstVal",
      0, ModelVector.FLOAT);

  /**NCI Mortality - Storm effect - medium damage*/
  protected ModelVector mp_fNCIStormEffectMed = new ModelVector(
      "NCI Storm Effect - Medium Damage (0-1)", "mo_prnciStormEffMedDmg",
      "mo_psemdVal", 0, ModelVector.FLOAT);

  /**NCI Mortality - Storm effect - full damage*/
  protected ModelVector mp_fNCIStormEffectFull = new ModelVector(
      "NCI Storm Effect - Complete Damage (0-1)",
      "mo_prnciStormEffFullDmg", "mo_psefdVal", 0, ModelVector.FLOAT);

  /**NCI Mortality - Neighbor storm effect - medium damage*/
  protected ModelVector mp_fNCINeighStormEffMed = new ModelVector(
      "NCI Neighbor Storm Damage (eta) - Medium (0-1)",
      "mo_prnciNeighStormEffMedDmg", "mo_pnsemdVal", 0, ModelVector.FLOAT, true);

  /**NCI Mortality - Neighbor storm effect - full damage*/
  protected ModelVector mp_fNCINeighStormEffFull = new ModelVector(
      "NCI Neighbor Storm Damage (eta) - Complete (0-1)",
      "mo_prnciNeighStormEffFullDmg", "mo_pnsefdVal", 0, ModelVector.FLOAT, true);

  /**Resource mortality - scaling factor (rho)*/
  protected ModelVector mp_fResMortScalingFactor = new ModelVector(
      "Growth/Resource - Scaling Factor (rho)",
      "mo_resMortScalingFactor", "mo_rmsfVal", 0, ModelVector.FLOAT);

  /**Resource mortality - function mode (mu)*/
  protected ModelVector mp_fResMortFunctionMode = new ModelVector(
      "Growth/Resource - Function Mode (mu)",
      "mo_resMortMode", "mo_rmmVal", 0, ModelVector.FLOAT);

  /**Resource mortality - growth increase in survival (delta)*/
  protected ModelVector mp_fResMortGrowthIncSurv = new ModelVector(
      "Growth/Resource - Survival Increase with Growth (delta)",
      "mo_resMortGrowthIncSurv", "mo_rmgisVal", 0, ModelVector.FLOAT);

  /**Resource mortality - low growth function shape (sigma)*/
  protected ModelVector mp_fResMortLoGrowthShape = new ModelVector(
      "Growth/Resource - Low-Growth Survival Parameter (sigma)",
      "mo_resMortLoGrowthShape", "mo_rmlgsVal", 0, ModelVector.FLOAT);

  /**Competition mortality - Shape parameter (Z)*/
  protected ModelVector mp_fCompMortShape = new ModelVector(
      "Competition Mortality Shape Parameter (Z)",
      "mo_CompMort", "mo_cmVal", 0, ModelVector.FLOAT);

  /**Competition mortality - maximum parameter (max)*/
  protected ModelVector mp_fCompMortMax = new ModelVector(
      "Competition Mortality Maximum Parameter (max)",
      "mo_CompMortMax", "mo_cmmVal", 0, ModelVector.FLOAT);

  /**Density self-thinning mortality asymptote (A)*/
  protected ModelVector mp_fDensSelfThinAsymptote = new ModelVector(
      "Density Self-Thinning Asymptote (A)",
      "mo_selfThinAsymptote", "mo_staVal", 0, ModelVector.FLOAT);

  /**Density self-thinning mortality diameter effect (C)*/
  protected ModelVector mp_fDensSelfThinDiamEffect = new ModelVector(
      "Density Self-Thinning Diameter Effect (C)",
      "mo_selfThinDiamEffect", "mo_stdieVal", 0, ModelVector.FLOAT);

  /**Density self-thinning mortality density effect (S)*/
  protected ModelVector mp_fDensSelfThinDensityEffect = new ModelVector(
      "Density Self-Thinning Density Effect (S)",
      "mo_selfThinDensityEffect", "mo_stdeeVal", 0, ModelVector.FLOAT);

  /**Density self-thinning neighborhood radius*/
  protected ModelVector mp_fDensSelfThinNeighRadius = new ModelVector(
      "Density Self-Thinning Neighborhood Radius, in m",
      "mo_selfThinRadius", "mo_strVal", 0, ModelVector.FLOAT);

  /**Density self-thinning minimum density for mortality*/
  protected ModelVector mp_fDensSelfThinMinDensity = new ModelVector(
      "Density Self-Thinning Minimum Density for Mortality (#/ha)",
      "mo_minDensityForMort", "mo_mdfmVal", 0, ModelVector.FLOAT);

  /**Logistic bi-level mortality - low-light "a"*/
  protected ModelVector mp_fLogBiLevLoLiteA = new ModelVector(
      "Logistic Bi-Level - Low-Light \"a\"",
      "mo_logBilevLoLiteA", "mo_lbllaVal", 0, ModelVector.FLOAT);

  /**Logistic bi-level mortality - low-light "b"*/
  protected ModelVector mp_fLogBiLevLoLiteB = new ModelVector(
      "Logistic Bi-Level - Low-Light \"b\"",
      "mo_logBilevLoLiteB", "mo_lbllbVal", 0, ModelVector.FLOAT);

  /**Logistic bi-level mortality - high-light "a"*/
  protected ModelVector mp_fLogBiLevHiLiteA = new ModelVector(
      "Logistic Bi-Level - High-Light \"a\"",
      "mo_logBilevHiLiteA", "mo_lbhlaVal", 0, ModelVector.FLOAT);

  /**Logistic bi-level mortality - high-light "b"*/
  protected ModelVector mp_fLogBiLevHiLiteB = new ModelVector(
      "Logistic Bi-Level - High-Light \"b\"",
      "mo_logBilevHiLiteB", "mo_lbhlbVal", 0, ModelVector.FLOAT);

  /** Logistic bi-level mortality - threshold for high-light mortality*/
  protected ModelVector mp_fLogBiLevHiLiteThreshold = new ModelVector(
      "Logistic Bi-Level - High-Light Mortality Threshold (0-100)",
      "mo_logBilevHiLiteThreshold", "mo_lbhltVal", 0, ModelVector.FLOAT);

  /**Stochastic bi-level mortality - low-light probability of mortality*/
  protected ModelVector mp_fStochBiLevLoLiteMortProb = new ModelVector(
     "Stochastic Bi-Level - Low-Light Mortality Probability (0-1)",
     "mo_stochBilevLoLiteMortProb", "mo_sbllmpVal", 0, ModelVector.FLOAT);

  /**Stochastic bi-level mortality - high-light probability of mortality*/
  protected ModelVector mp_fStochBiLevHiLiteMortProb = new ModelVector(
     "Stochastic Bi-Level - High-Light Mortality Probability (0-1)",
     "mo_stochBilevHiLiteMortProb", "mo_sbhlmpVal", 0, ModelVector.FLOAT);

  /** Stochastic bi-level mortality - threshold for high-light mortality*/
  protected ModelVector mp_fStochBiLevHiLiteThreshold = new ModelVector(
      "Stochastic Bi-Level - High-Light Mortality Threshold (0-100)",
      "mo_stochBilevHiLiteThreshold", "mo_sbhltVal", 0, ModelVector.FLOAT);

  /** Height-GLI weibull mortality - a */
  protected ModelVector mp_fHeightGLIWeibA = new ModelVector(
      "Height-GLI Weibull - a", "mo_heightGLIWeibA", "mo_hgwaVal", 0,
      ModelVector.FLOAT);

  /** Height-GLI weibull mortality - b */
  protected ModelVector mp_fHeightGLIWeibB = new ModelVector(
      "Height-GLI Weibull - b", "mo_heightGLIWeibB", "mo_hgwbVal", 0,
      ModelVector.FLOAT);

  /** Height-GLI weibull mortality - c */
  protected ModelVector mp_fHeightGLIWeibC = new ModelVector(
      "Height-GLI Weibull - c", "mo_heightGLIWeibC", "mo_hgwcVal", 0,
      ModelVector.FLOAT);

  /** Height-GLI weibull mortality - d */
  protected ModelVector mp_fHeightGLIWeibD = new ModelVector(
      "Height-GLI Weibull - d", "mo_heightGLIWeibD", "mo_hgwdVal", 0,
      ModelVector.FLOAT);

  /** Height-GLI weibull mortality - max mortality */
  protected ModelVector mp_fHeightGLIWeibMaxMort = new ModelVector(
      "Height-GLI Weibull - Max Mortality (0 - 1)", "mo_heightGLIWeibMaxMort",
      "mo_hgwmmVal", 0, ModelVector.FLOAT);


  /**NCI - NCI DBH divisor*/
  protected ModelFloat m_fNCIDbhDivisor = new ModelFloat(1,
      "NCI DBH Divisor (q)", "mo_nciDbhDivisor");

  /**NCI - Whether or not to include snags in NCI calculations - LEM 05/12/05*/
  protected ModelEnum m_iIncludeSnagsInNCI =
      new ModelEnum(new int[] {0, 1}
                    ,
                    new String[] {"false", "true"}
                    ,
                    "Include Snags in NCI Calculations",
                    "mo_nciIncludeSnagsInNCI");

  /**
   * Constructor.
   * @param oManager GUIManager object.
   * @throws ModelException passed through from called functions.  Should never
   * be thrown.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public MortalityBehaviors(GUIManager oManager) throws ModelException {
    super(oManager, "Mortality", "mortality");

    m_sHelpID = "data.mortality_behaviors";

    //Set up child behaviors
    mp_oChildBehaviors = new Behavior[15];

    //*****************
     //GMF mortality
     //*****************
    mp_oChildBehaviors[0] = new Behavior("gmfmort", "GMF Mortality",
                                         "gmfmortality", (float) 1);
    //Data
    mp_oChildBehaviors[0].AddRequiredData(mp_fMortAtZeroGrowth);
    mp_oChildBehaviors[0].AddRequiredData(mp_fLightDependentMortality);
    //Data members
    mp_oChildBehaviors[0].mp_oNewTreeDataMembers.add(new DataMember("Dead flag",
        "dead", DataMember.BOOLEAN));

    //*****************
     //BC mortality
     //*****************
    mp_oChildBehaviors[1] = new Behavior("bcmort", "BC Mortality",
                                         "bcmortality",
                                         (float) 1);
    //Data
    mp_oChildBehaviors[1].AddRequiredData(mp_fMortAtZeroGrowth);
    mp_oChildBehaviors[1].AddRequiredData(mp_fLightDependentMortality);
    //Data members
    mp_oChildBehaviors[1].mp_oNewTreeDataMembers.add(new DataMember("Dead flag",
        "dead", DataMember.BOOLEAN));

    //*****************
     //Adult self-thinning
     //*****************
    mp_oChildBehaviors[2] = new Behavior("adultselfthin", "Adult Self Thinning",
                                         "adultselfthin", (float) 1);
    //Data
    mp_oChildBehaviors[2].AddRequiredData(mp_fAdultSelfThinningSlope);
    mp_oChildBehaviors[2].AddRequiredData(mp_fAdultSelfThinningIntercept);
    mp_oChildBehaviors[2].AddRequiredData(mp_fAdultSelfThinningMaxDbh);
    //Disallow for seedlings
    mp_oChildBehaviors[2].SetCanApplyTo(TreePopulation.SEEDLING, false);
    mp_oChildBehaviors[2].SetCanApplyTo(TreePopulation.SAPLING, false);
    //Data members
    mp_oChildBehaviors[2].mp_oNewTreeDataMembers.add(new DataMember("Dead flag",
        "dead", DataMember.BOOLEAN));

    //*****************
     //Juvenile self-thinning
     //*****************
    mp_oChildBehaviors[3] = new Behavior("juvselfthin",
                                         "Juvenile Self Thinning",
                                         "juvselfthin", (float) 1);
    //Data
    mp_oChildBehaviors[3].AddRequiredData(mp_fJuvenileSelfThinningSlope);
    mp_oChildBehaviors[3].AddRequiredData(mp_fJuvenileSelfThinningIntercept);
    //Disallow for adults
    mp_oChildBehaviors[3].SetCanApplyTo(TreePopulation.ADULT, false);
    //Data members
    mp_oChildBehaviors[3].mp_oNewTreeDataMembers.add(new DataMember("Dead flag",
        "dead", DataMember.BOOLEAN));

    //*****************
     //Senescense
     //*****************
    mp_oChildBehaviors[4] = new Behavior("senescence", "Senescence",
                                         "senescence", (float) 1);
    //Data
    mp_oChildBehaviors[4].AddRequiredData(mp_fRandomMortalityAlpha);
    mp_oChildBehaviors[4].AddRequiredData(mp_fRandomMortalityBeta);
    mp_oChildBehaviors[4].AddRequiredData(mp_fDbhAtOnsetOfSenescence);
    mp_oChildBehaviors[4].AddRequiredData(m_iDbhAtAsymptoticMaximumMortality);
    //Disallow for seedlings
    mp_oChildBehaviors[4].SetCanApplyTo(TreePopulation.SEEDLING, false);
    //Data members
    mp_oChildBehaviors[4].mp_oNewTreeDataMembers.add(new DataMember("Dead flag",
        "dead", DataMember.BOOLEAN));

    //*****************
     //Adult stochastic mortality
     //*****************
    mp_oChildBehaviors[5] = new Behavior("adultstochmort",
                                         "Adult Stochastic Mortality",
                                         "adultstochasticmort", (float) 1);
    //Data
    mp_oChildBehaviors[5].AddRequiredData(mp_fAdultRandomMortality);
    //Data members
    mp_oChildBehaviors[5].mp_oNewTreeDataMembers.add(new DataMember("Dead flag",
        "dead", DataMember.BOOLEAN));

    //*****************
     //Juvenile stochastic mortality
     //*****************
    mp_oChildBehaviors[6] = new Behavior("juvstochmort",
                                         "Juvenile Stochastic Mortality",
                                         "juvstochasticmort", (float) 1);
    //Data
    mp_oChildBehaviors[6].AddRequiredData(mp_fJuvenileRandomMortality);
    //Data members
    mp_oChildBehaviors[6].mp_oNewTreeDataMembers.add(new DataMember("Dead flag",
        "dead", DataMember.BOOLEAN));

    //*****************
     //Weibull snag mortality
     //*****************
    mp_oChildBehaviors[7] = new Behavior("weibull snag mortality",
                                         "Weibull Snag Mortality",
                                         "weibull snag mortality", (float) 1.1);
    //Set to only apply for snags
    mp_oChildBehaviors[7].SetCanApplyTo(TreePopulation.SEEDLING, false);
    mp_oChildBehaviors[7].SetCanApplyTo(TreePopulation.SAPLING, false);
    mp_oChildBehaviors[7].SetCanApplyTo(TreePopulation.ADULT, false);
    mp_oChildBehaviors[7].SetCanApplyTo(TreePopulation.SNAG, true);
    //Data
    mp_oChildBehaviors[7].AddRequiredData(mp_fSnag1WeibullA);
    mp_oChildBehaviors[7].AddRequiredData(mp_fSnag2WeibullA);
    mp_oChildBehaviors[7].AddRequiredData(mp_fSnag3WeibullA);
    mp_oChildBehaviors[7].AddRequiredData(mp_fSnag1WeibullB);
    mp_oChildBehaviors[7].AddRequiredData(mp_fSnag2WeibullB);
    mp_oChildBehaviors[7].AddRequiredData(mp_fSnag3WeibullB);
    mp_oChildBehaviors[7].AddRequiredData(mp_fSnagSizeClass1Dbh);
    mp_oChildBehaviors[7].AddRequiredData(mp_fSnagSizeClass2Dbh);
    //Data members
    mp_oChildBehaviors[7].mp_oNewTreeDataMembers.add(new DataMember("Dead flag",
        "dead", DataMember.BOOLEAN));

    //*****************
     //NCI mortality
     //*****************
    mp_oChildBehaviors[8] = new Behavior("NCI Mortality",
                                         "NCI Mortality",
                                         "NCI Mortality",
                                         (float) 2.0);
    mp_oChildBehaviors[8].AddRequiredData(mp_fNCIMaxCrowdingRadius);
    mp_oChildBehaviors[8].AddRequiredData(mp_fNCINeighDBHEffect);
    mp_oChildBehaviors[8].AddRequiredData(mp_fNCINeighDistanceEffect);
    mp_oChildBehaviors[8].AddRequiredData(mp_fNCISizeSensToNCI);
    mp_oChildBehaviors[8].AddRequiredData(mp_fNCIMaxProbSurvival);
    mp_oChildBehaviors[8].AddRequiredData(mp_fNCISizeEffectMode);
    mp_oChildBehaviors[8].AddRequiredData(mp_fNCISizeEffectVariance);
    mp_oChildBehaviors[8].AddRequiredData(mp_fNCISlope);
    mp_oChildBehaviors[8].AddRequiredData(mp_fNCISteepness);
    mp_oChildBehaviors[8].AddRequiredData(mp_fNCIStormEffectMed);
    mp_oChildBehaviors[8].AddRequiredData(mp_fNCIStormEffectFull);
    mp_oChildBehaviors[8].AddRequiredData(mp_fNCINeighStormEffMed);
    mp_oChildBehaviors[8].AddRequiredData(mp_fNCINeighStormEffFull);
    mp_oChildBehaviors[8].AddRequiredData(mp_fNCIMinNeighborDBH);
    mp_oChildBehaviors[8].AddRequiredData(mp_fNCIShadingEffectCoefficient);
    mp_oChildBehaviors[8].AddRequiredData(mp_fNCIShadingEffectExponent);
    mp_oChildBehaviors[8].AddRequiredData(m_fNCIDbhDivisor);
    mp_oChildBehaviors[8].AddRequiredData(m_iIncludeSnagsInNCI);
    //Disallow for seedlings
    mp_oChildBehaviors[8].SetCanApplyTo(TreePopulation.SEEDLING, false);
    //Data members
    mp_oChildBehaviors[8].mp_oNewTreeDataMembers.add(new DataMember("Dead flag",
        "dead", DataMember.BOOLEAN));


    //*****************
     //Growth and resource based mortality
     //*****************
    mp_oChildBehaviors[9] = new Behavior("resource mortality",
                                         "Growth and Resource-Based Mortality",
                                         "Growth resource mortality",
                                         (float) 1);
    mp_oChildBehaviors[9].AddRequiredData(mp_fResMortFunctionMode);
    mp_oChildBehaviors[9].AddRequiredData(mp_fResMortGrowthIncSurv);
    mp_oChildBehaviors[9].AddRequiredData(mp_fResMortLoGrowthShape);
    mp_oChildBehaviors[9].AddRequiredData(mp_fResMortScalingFactor);
    //Data members
    mp_oChildBehaviors[9].mp_oNewTreeDataMembers.add(new DataMember("Dead flag",
        "dead", DataMember.BOOLEAN));


    //*****************
     //Competition mortality
     //*****************
    mp_oChildBehaviors[10] = new Behavior("competition mortality",
                                          "Competition Mortality",
                                          "competitionmortality",
                                          (float) 1);
    mp_oChildBehaviors[10].AddRequiredData(mp_fCompMortShape);
    mp_oChildBehaviors[10].AddRequiredData(mp_fCompMortMax);
    //Can't apply to seedlings
    mp_oChildBehaviors[10].SetCanApplyTo(TreePopulation.SEEDLING, false);
    //Data members
    mp_oChildBehaviors[10].mp_oNewTreeDataMembers.add(new DataMember("Dead flag",
        "dead", DataMember.BOOLEAN));

    //*****************
     //Density self-thinning - LEM 03-17-2005
     //*****************
    mp_oChildBehaviors[11] = new Behavior("density self-thinning mortality",
                                          "Density Self-Thinning Mortality",
                                          "densityselfthinning",
                                          (float) 1);
    mp_oChildBehaviors[11].AddRequiredData(mp_fDensSelfThinAsymptote);
    mp_oChildBehaviors[11].AddRequiredData(mp_fDensSelfThinDensityEffect);
    mp_oChildBehaviors[11].AddRequiredData(mp_fDensSelfThinDiamEffect);
    mp_oChildBehaviors[11].AddRequiredData(mp_fDensSelfThinMinDensity);
    mp_oChildBehaviors[11].AddRequiredData(mp_fDensSelfThinNeighRadius);
    //Can't apply to adults
    mp_oChildBehaviors[11].SetCanApplyTo(TreePopulation.ADULT, false);
    //Data members
    mp_oChildBehaviors[11].mp_oNewTreeDataMembers.add(new DataMember("Dead flag",
        "dead", DataMember.BOOLEAN));

    //*****************
     //Logistic bi-level - LEM 05-06-2005
     //*****************
    mp_oChildBehaviors[12] = new Behavior("logistic bi-level mortality",
                                          "Logistic Bi-Level Mortality",
                                          "Logistic Bi-Level Mortality",
                                          (float) 1);
    mp_oChildBehaviors[12].AddRequiredData(mp_fLogBiLevHiLiteThreshold);
    mp_oChildBehaviors[12].AddRequiredData(mp_fLogBiLevLoLiteA);
    mp_oChildBehaviors[12].AddRequiredData(mp_fLogBiLevLoLiteB);
    mp_oChildBehaviors[12].AddRequiredData(mp_fLogBiLevHiLiteA);
    mp_oChildBehaviors[12].AddRequiredData(mp_fLogBiLevHiLiteB);
    //Data members
    mp_oChildBehaviors[12].mp_oNewTreeDataMembers.add(new DataMember("Dead flag",
        "dead", DataMember.BOOLEAN));

    //*****************
     //Stochastic bi-level - LEM 10-31-2005
     //*****************
    mp_oChildBehaviors[13] = new Behavior("stochastic bi-level mortality",
                                          "Stochastic Bi-Level Mortality",
                                          "Stochastic Bi-Level Mortality",
                                          (float) 1);
    mp_oChildBehaviors[13].AddRequiredData(mp_fStochBiLevHiLiteThreshold);
    mp_oChildBehaviors[13].AddRequiredData(mp_fStochBiLevLoLiteMortProb);
    mp_oChildBehaviors[13].AddRequiredData(mp_fStochBiLevHiLiteMortProb);
    //Data members
    mp_oChildBehaviors[13].mp_oNewTreeDataMembers.add(new DataMember("Dead flag",
        "dead", DataMember.BOOLEAN));

    //*****************
     //Height-GLI Weibull mortality - LEM 01-26-2006
     //*****************
    mp_oChildBehaviors[14] = new Behavior("height-gli weibull mortality",
                                          "Height-GLI Weibull Mortality",
                                          "Height GLI Weibull Mortality",
                                          (float) 1);
    mp_oChildBehaviors[14].AddRequiredData(mp_fHeightGLIWeibMaxMort);
    mp_oChildBehaviors[14].AddRequiredData(mp_fHeightGLIWeibA);
    mp_oChildBehaviors[14].AddRequiredData(mp_fHeightGLIWeibB);
    mp_oChildBehaviors[14].AddRequiredData(mp_fHeightGLIWeibC);
    mp_oChildBehaviors[14].AddRequiredData(mp_fHeightGLIWeibD);
    //Data members
    mp_oChildBehaviors[14].mp_oNewTreeDataMembers.add(new DataMember("Dead flag",
        "dead", DataMember.BOOLEAN));

    mp_oAllData.add(mp_fJuvenileRandomMortality);
    mp_oAllData.add(mp_fAdultRandomMortality);
    mp_oAllData.add(mp_fMortAtZeroGrowth);
    mp_oAllData.add(mp_fLightDependentMortality);
    mp_oAllData.add(mp_fJuvenileSelfThinningSlope);
    mp_oAllData.add(mp_fJuvenileSelfThinningIntercept);
    mp_oAllData.add(mp_fAdultSelfThinningSlope);
    mp_oAllData.add(mp_fAdultSelfThinningIntercept);
    mp_oAllData.add(mp_fAdultSelfThinningMaxDbh);
    mp_oAllData.add(mp_fRandomMortalityAlpha);
    mp_oAllData.add(mp_fRandomMortalityBeta);
    mp_oAllData.add(mp_fDbhAtOnsetOfSenescence);
    mp_oAllData.add(m_iDbhAtAsymptoticMaximumMortality);
    mp_oAllData.add(mp_fSnag1WeibullA);
    mp_oAllData.add(mp_fSnag2WeibullA);
    mp_oAllData.add(mp_fSnag3WeibullA);
    mp_oAllData.add(mp_fSnag1WeibullB);
    mp_oAllData.add(mp_fSnag2WeibullB);
    mp_oAllData.add(mp_fSnag3WeibullB);
    mp_oAllData.add(mp_fSnagSizeClass1Dbh);
    mp_oAllData.add(mp_fSnagSizeClass2Dbh);
    mp_oAllData.add(m_fSnagSizeClass1DbhDep);
    mp_oAllData.add(m_fSnagSizeClass2DbhDep);
    mp_oAllData.add(mp_fNCIMaxCrowdingRadius);
    mp_oAllData.add(mp_fNCINeighDBHEffect);
    mp_oAllData.add(mp_fNCINeighDistanceEffect);
    mp_oAllData.add(mp_fNCISizeSensToNCI);
    mp_oAllData.add(mp_fNCIMaxProbSurvival);
    mp_oAllData.add(mp_fNCISizeEffectMode);
    mp_oAllData.add(mp_fNCISizeEffectVariance);
    mp_oAllData.add(mp_fNCISlope);
    mp_oAllData.add(mp_fNCISteepness);
    mp_oAllData.add(mp_fNCIStormEffectMed);
    mp_oAllData.add(mp_fNCIStormEffectFull);
    mp_oAllData.add(mp_fNCINeighStormEffMed);
    mp_oAllData.add(mp_fNCINeighStormEffFull);
    mp_oAllData.add(mp_fNCIMinNeighborDBH);
    mp_oAllData.add(mp_fNCIShadingEffectCoefficient);
    mp_oAllData.add(mp_fNCIShadingEffectExponent);
    mp_oAllData.add(m_fNCIDbhDivisor);
    mp_oAllData.add(m_iIncludeSnagsInNCI);
    mp_oAllData.add(mp_fResMortFunctionMode);
    mp_oAllData.add(mp_fResMortGrowthIncSurv);
    mp_oAllData.add(mp_fResMortLoGrowthShape);
    mp_oAllData.add(mp_fResMortScalingFactor);
    mp_oAllData.add(mp_fCompMortShape);
    mp_oAllData.add(mp_fCompMortMax);
    mp_oAllData.add(mp_fDensSelfThinAsymptote);
    mp_oAllData.add(mp_fDensSelfThinDensityEffect);
    mp_oAllData.add(mp_fDensSelfThinDiamEffect);
    mp_oAllData.add(mp_fDensSelfThinMinDensity);
    mp_oAllData.add(mp_fDensSelfThinNeighRadius);
    mp_oAllData.add(mp_fLogBiLevHiLiteThreshold);
    mp_oAllData.add(mp_fLogBiLevLoLiteA);
    mp_oAllData.add(mp_fLogBiLevLoLiteB);
    mp_oAllData.add(mp_fLogBiLevHiLiteA);
    mp_oAllData.add(mp_fLogBiLevHiLiteB);
    mp_oAllData.add(mp_fStochBiLevHiLiteThreshold);
    mp_oAllData.add(mp_fStochBiLevLoLiteMortProb);
    mp_oAllData.add(mp_fStochBiLevHiLiteMortProb);
    mp_oAllData.add(mp_fHeightGLIWeibMaxMort);
    mp_oAllData.add(mp_fHeightGLIWeibA);
    mp_oAllData.add(mp_fHeightGLIWeibB);
    mp_oAllData.add(mp_fHeightGLIWeibC);
    mp_oAllData.add(mp_fHeightGLIWeibD);
  }

  /**
   * Adds NCI lambda values.
   * @param oPop Tree population
   * @throws ModelException Not thrown.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   * <br>October 4, 2004:  Added NCI stuff (LEM)
   * <br>March 16, 2005:  Added NCI array setup to aid backwards compatibility
   * (LEM)
   */
  public void DoSetup(TreePopulation oPop) throws ModelException {
    Behavior oBeh = GetBehaviorByKey("NCI Mortality");
    int iNumSpecies = oPop.GetNumberOfSpecies(), i, j;
    boolean bAlreadyLambdas = false;

    /*    for (i = 0; i < iNumSpecies; i++) {
          String sSpName = oPop.GetSpeciesNameFromCode(i);
          ModelVector oVector = new ModelVector(
              sSpName.replace('_', ' ') + " NCI Lambda Neighbors",
              "mo_prnci" + sSpName + "NeighborLambda", "mo_plVal",
              0, ModelVector.FLOAT);
          mp_oAllData.add(oVector);
          oBeh.AddRequiredData(oVector);
        } */

    //See if we already had lambdas
    for (i = 0; i < mp_oAllData.size(); i++) {
      ModelData oData = (ModelData) mp_oAllData.elementAt(i);
      String sKey = oData.GetDescriptor().toLowerCase();
      if (sKey.indexOf("lambda neighbor") > -1) {
        bAlreadyLambdas = true;
        break;
      }
    }

    if (bAlreadyLambdas == false) {

      for (i = 0; i < iNumSpecies; i++) {
        String sSpName = oPop.GetSpeciesNameFromCode(i);
        ModelVector oVector = new ModelVector(
            sSpName.replace('_', ' ') + " NCI Lambda Neighbors",
            "mo_prnci" + sSpName + "NeighborLambda", "mo_plVal",
            0, ModelVector.FLOAT);

        mp_oAllData.add(oVector);
        oBeh.AddRequiredData(oVector);
      }
    }
    else {
      //We already had lambdas.  What we want to do is preserve any for species
      //that didn't change, because they may have values.

      //Create an array of the lambda arrays, placed in the new species
      //order
      ModelData[] p_oNCI = new ModelData[iNumSpecies];
      int iCode;

      for (i = 0; i < iNumSpecies; i++) {
        p_oNCI[i] = null;
      }

      //Get the existing lambdas and put them where they go in our arrays -
      //any for species that were removed will be left behind
      for (i = 0; i < mp_oAllData.size(); i++) {
        ModelData oData = (ModelData) mp_oAllData.elementAt(i);
        String sKey = oData.GetDescriptor().toLowerCase();
        if (sKey.indexOf("lambda neighbor") > -1) {
          //get the species
          String sSpecies;
          sKey = oData.GetDescriptor();
          sSpecies = sKey.substring(0, sKey.indexOf(" NCI")).replace(' ', '_');
          iCode = oPop.GetSpeciesCodeFromName(sSpecies);
          if (iCode > -1) {
            p_oNCI[iCode] = oData;
          }
        }
      }

      //Now remove all lambdas from the required data and from all behaviors
      for (i = 0; i < mp_oAllData.size(); i++) {
        ModelData oData = (ModelData) mp_oAllData.elementAt(i);
        String sKey = oData.GetDescriptor().toLowerCase();
        if (sKey.indexOf("lambda neighbor") > -1) {
          mp_oAllData.remove(i);
          i--;
        }
      }
      for (i = 0; i < mp_oChildBehaviors.length; i++) {
        for (j = 0; j < mp_oChildBehaviors[i].mp_oRequiredData.size(); j++) {
          ModelData oData = (ModelData) mp_oChildBehaviors[i].
              mp_oRequiredData.elementAt(j);
          String sKey = oData.GetDescriptor().toLowerCase();
          if (sKey.indexOf("lambda neighbor") > -1) {
            mp_oChildBehaviors[i].mp_oRequiredData.remove(j);
            j--;
          }
        }
      }

      //Now add back the lambdas, creating new ones for any new species
      for (i = 0; i < p_oNCI.length; i++) {
        if (p_oNCI[i] == null) {
          String sSpName = oPop.GetSpeciesNameFromCode(i);
          p_oNCI[i] = new ModelVector(
              sSpName.replace('_', ' ') + " NCI Lambda Neighbors",
              "mo_prnci" + sSpName + "NeighborLambda", "mo_plVal",
              0, ModelVector.FLOAT);
        }
        mp_oAllData.add(p_oNCI[i]);
        oBeh.AddRequiredData(p_oNCI[i]);
      }
    }

    //Now set new NCI arrays to 0 to allow easy backwards compatibility -
    //remove this section later.  Setting them to 0 allows a simple save to
    //write a new, compatible file.
    //LEM 03-16-05
    if (mp_fNCIMinNeighborDBH.GetValue().size() == 0) {
      for (i = 0; i < iNumSpecies; i++) {
        mp_fNCIMinNeighborDBH.GetValue().add(new Float(0.0));
      }
    }
    if (mp_fNCIShadingEffectCoefficient.GetValue().size() == 0) {
      for (i = 0; i < iNumSpecies; i++) {
        mp_fNCIShadingEffectCoefficient.GetValue().add(new Float(0.0));
      }
    }
    if (mp_fNCIShadingEffectExponent.GetValue().size() == 0) {
      for (i = 0; i < iNumSpecies; i++) {
        mp_fNCIShadingEffectExponent.GetValue().add(new Float(0.0));
      }
    }

    //If the damage parameters have not already been set up, set all their
    //values to a default of 1.  Otherwise leave them alone.
    if (mp_fNCINeighStormEffMed.GetValue().size() == 0) {
      for (i = 0; i < iNumSpecies; i++) {
        mp_fNCINeighStormEffMed.GetValue().add(new Float(1.0));
      }
    }

    if (mp_fNCINeighStormEffFull.GetValue().size() == 0) {
      for (i = 0; i < iNumSpecies; i++) {
        mp_fNCINeighStormEffFull.GetValue().add(new Float(1.0));
      }
    }

    if (mp_fNCIStormEffectMed.GetValue().size() == 0) {
      for (i = 0; i < iNumSpecies; i++) {
        mp_fNCIStormEffectMed.GetValue().add(new Float(1.0));
      }
    }

    if (mp_fNCIStormEffectFull.GetValue().size() == 0) {
      for (i = 0; i < iNumSpecies; i++) {
        mp_fNCIStormEffectFull.GetValue().add(new Float(1.0));
      }
    }
  }

  /**
   * Overridden to provide backwards compatibility for Weibull snag mortality
   * @param sXMLTag XML tag of data object whose value is to be set.
   * @param sXMLParentTag The immediate parent tag that sXMLTag is within.
   * @param oAttributes Attributes of the object.  Ignored, but may be needed
   * by overriding objects.
   * @param oData Data value, either String or type appropriate to the data
   * type
   * @return true if the value was set successfully; false if the value could
   * not be found.  (This would not be an error, because I need a way to
   * cycle through the objects until one of the objects comes up with a
   * match.)
   * @throws ModelException if the value could not be assigned to the data
   * object.
   */
  public boolean SetSingleValueByXMLTag(String sXMLTag, String sXMLParentTag,
                                        org.xml.sax.Attributes oAttributes,
                                        Object oData) throws
      ModelException {
    boolean bFlag = super.SetSingleValueByXMLTag(sXMLTag, sXMLParentTag,
                                                 oAttributes, oData);
    if (bFlag == true) {
      //A value was assigned - see if it was one of the weibull snag size
      //classes
      if (sXMLTag.equals(m_fSnagSizeClass1DbhDep.GetXMLTag())) {
        //Cross-set this value into the species-specific version
        Float fVal = null;
        if (oData instanceof String) {
          fVal = new Float( (String) oData);
        }
        else {
          fVal = (Float) oData;
        }
        int iNumSpecies = m_oManager.GetTreePopulation().GetNumberOfSpecies();
        Float[] p_fVals = new Float[iNumSpecies];
        for (int i = 0; i < p_fVals.length; i++) {
          p_fVals[i] = fVal;
        }
        SetVectorValues(mp_fSnagSizeClass1Dbh, p_fVals);
      }
      else if (sXMLTag.equals(m_fSnagSizeClass2DbhDep.GetXMLTag())) {
        //Cross-set this value into the species-specific version
        Float fVal = null;
        if (oData instanceof String) {
          fVal = new Float( (String) oData);
        }
        else {
          fVal = (Float) oData;
        }
        int iNumSpecies = m_oManager.GetTreePopulation().GetNumberOfSpecies();
        Float[] p_fVals = new Float[iNumSpecies];
        for (int i = 0; i < p_fVals.length; i++) {
          p_fVals[i] = fVal;
        }
        SetVectorValues(mp_fSnagSizeClass2Dbh, p_fVals);
      }
    }
    return bFlag;
  }

  /**
   * Validates the data prior to writing it to a parameter file or some such.
   * @param oPop TreePopulation
   * @throws ModelException if:
   * <ul>
   * <li>Any of the following are not expressed as a proportion:
   * <ul>
   * <li>mp_fAdultRandomMortality
   * <li>mp_fJuvenileRandomMortality
   * <li>mp_fMortAtZeroGrowth
   * </ul></li>
   * <li>If adult self-thinning is used and any mp_fAdultSelfThinningMaxDbh
   * value is less than zero.</li>
   * <li>Any  value in mp_fDbhAtOnsetOfSenescence is less than zero.</li>
   * <li>Any vector is not sized correctly to equal total number of
   * species.</li>
   * <li>GMF mortality is enabled and the timestep length is not five
   * years.</li>
   * <li>Any value in light dependent mortality is zero.</li>
   * <li>Competition mortality is enabled and the timestep length is not one
   * year.</li>
   * <li>Competition mortality is enabled and NCI growth is not.</li>
   * <li>NCI Mortality is enabled and any of the following is true:
   * <ul>
   * <li>NCI Max. radius of neighbor effects is <= 0 for any species</li>
   * <li>NCI Max growth is <= 0 for any species</li>
   * <li>NCI size effect mode for any species is <= 0</li>
   * <li>NCI Size effect variance = 0 for any species</li>
   * <li>NCI Eta for any species for any damage category is not between 0 and
   * 1</li>
   * <li>NCI Storm Effect parameters for any species for any damage category
   * are not between 0 and 1</li>
   * <li>NCI DBH divisor is <= 0</li>
   * <li>Minimun neighbor DBH value is < 0</li>
   * </ul></li>
   * <li>Density self-thinning is enabled, and any of the following are true:
   * <ul>
   * <li>The timestep length is greater than 1.</li>
   * <li>The neighborhood radius is not greater than 0.</li>
   * <li>The minimum density for mortality is less than 0.</li>
   * <li>The self-thinning asymptote is not greater than 0.</li>
   * <li>The density effect is not greater than 0.</li>
   * </ul></li>
   * <li>Logistic bi-level is enabled, and the threshold for high-light
   * mortality is not between 0 and 100.</li>
   * <li>Stochastic bi-level is enabled, and the threshold for high-light
   * mortality is not between 0 and 100.</li>
   * <li>Stochastic bi-level is enabled, and values for high- and low-light
   * mortality probability are not between 0 and 1.</li>
   * <li>Stochastic bi-level is enabled, and Storm Light is not.</li>
   * <li>Weibull snag mortality is enabled, and either a is 0 or a is negative
   * and b is not a whole number for any size class.</li>
   * <li>Height-GLI Weibull mortality is enabled, and max mortality is not a
   * proportion.</li>
   * </ul>
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   * <br>April 30, 2004: Fixed some validation bugs (LEM)
   * <br>March 17, 2005:  Added density self-thinning validation (LEM)
   * <br>May 6, 2005:  Added logistic bi-level validation (LEM)
   * <br>October 31, 2005:  Added stochastic bi-level validation (LEM)
   * <br>January 26, 2006:  Added Height-GlI weibull validation (LEM)
   */
  public void ValidateData(TreePopulation oPop) throws ModelException {
    Behavior oBeh;
    int iNumSpecies = oPop.GetNumberOfSpecies(),
        i;

    //If no behaviors are enabled, exit
    if (!AnyBehaviorsEnabled()) {
      return;
    }

    //For each behavior, make sure that the appropriate data is good
    //GMF mortality
    oBeh = GetBehaviorByKey("gmfmort");
    if (oBeh.m_bIsEnabled) {
      MakeSureRightSize(mp_fMortAtZeroGrowth, iNumSpecies);
      MakeSureRightSize(mp_fLightDependentMortality, iNumSpecies);

      if (m_oManager.GetPlot().GetNumberOfYearsPerTimestep() != 5) {
        ModelException oErr = new ModelException();
        oErr.iErrorCode = ErrorGUI.BAD_DATA;
        oErr.sMessage =
            "GMF Mortality cannot be used unless the timestep length is five years.";
        oErr.sCallingFunction = "JAVA";
        throw (oErr);
      }

      MakeSureAllPositive(mp_fLightDependentMortality,
                          oBeh.GetWhichSpeciesUsed(oPop));
    }

    //BC mortality
    oBeh = GetBehaviorByKey("bcmort");
    if (oBeh.m_bIsEnabled) {
      MakeSureRightSize(mp_fMortAtZeroGrowth, iNumSpecies);
      MakeSureRightSize(mp_fLightDependentMortality, iNumSpecies);
      MakeSureAllPositive(mp_fLightDependentMortality,
                          oBeh.GetWhichSpeciesUsed(oPop));
    }

    //Senescence
    oBeh = GetBehaviorByKey("senescence");
    if (oBeh.m_bIsEnabled) {
      MakeSureRightSize(mp_fDbhAtOnsetOfSenescence, iNumSpecies);
      MakeSureRightSize(mp_fRandomMortalityAlpha, iNumSpecies);
      MakeSureRightSize(mp_fRandomMortalityBeta, iNumSpecies);
      MakeSurePositive(m_iDbhAtAsymptoticMaximumMortality);
      MakeSureAllPositive(mp_fDbhAtOnsetOfSenescence,
                          oBeh.GetWhichSpeciesUsed(oPop));

    }

    //Adult self-thinning
    oBeh = GetBehaviorByKey("adultselfthin");
    if (oBeh.m_bIsEnabled) {
      MakeSureRightSize(mp_fAdultSelfThinningIntercept, iNumSpecies);
      MakeSureRightSize(mp_fAdultSelfThinningMaxDbh, iNumSpecies);
      MakeSureRightSize(mp_fAdultSelfThinningSlope, iNumSpecies);
      MakeSureAllPositive(mp_fAdultSelfThinningMaxDbh,
                          oBeh.GetWhichSpeciesUsed(oPop));
    }

    //Juvenile self-thinning
    oBeh = GetBehaviorByKey("juvselfthin");
    if (oBeh.m_bIsEnabled) {
      MakeSureRightSize(mp_fJuvenileSelfThinningIntercept, iNumSpecies);
      MakeSureRightSize(mp_fJuvenileSelfThinningSlope, iNumSpecies);
    }

    //Adult stochastic mortality
    oBeh = GetBehaviorByKey("adultstochmort");
    if (oBeh.m_bIsEnabled) {
      MakeSureRightSize(mp_fAdultRandomMortality, iNumSpecies);
      MakeSureAllAreProportions(mp_fAdultRandomMortality,
                                oBeh.GetWhichSpeciesUsed(oPop));
    }

    //Juvenile stochastic mortality
    oBeh = GetBehaviorByKey("juvstochmort");
    if (oBeh.m_bIsEnabled) {
      MakeSureRightSize(mp_fJuvenileRandomMortality, iNumSpecies);
      MakeSureAllAreProportions(mp_fJuvenileRandomMortality,
                                oBeh.GetWhichSpeciesUsed(oPop));
    }

    //Weibull snag mortality
    oBeh = GetBehaviorByKey("weibull snag mortality");
    if (oBeh.m_bIsEnabled) {
      boolean[] p_bAppliesTo = oBeh.GetWhichSpeciesUsed(oPop);
      MakeSureAllNonZero(mp_fSnag1WeibullA, p_bAppliesTo);
      MakeSureAllNonZero(mp_fSnag2WeibullA, p_bAppliesTo);
      MakeSureAllNonZero(mp_fSnag3WeibullA, p_bAppliesTo);

      //Make sure something's entered

      int iCount = 0;
      for (i = 0; i < p_bAppliesTo.length; i++) {
        if (p_bAppliesTo[i] == true) iCount++;
      }

      if(mp_fSnag1WeibullA.GetValue().size() < iCount) {
        throw(new ModelException(ErrorGUI.BAD_DATA, "JAVA",
             m_sName + ":  There is the wrong number of values in " +
          mp_fSnag1WeibullA.GetDescriptor() +
          ".  Expected: " + iCount + "; actual: " + mp_fSnag1WeibullA.GetValue().size() +
          "."));
      }
      if(mp_fSnag2WeibullA.GetValue().size() < iCount) {
        throw(new ModelException(ErrorGUI.BAD_DATA, "JAVA",
             m_sName + ":  There is the wrong number of values in " +
          mp_fSnag2WeibullA.GetDescriptor() +
          ".  Expected: " + iCount + "; actual: " + mp_fSnag2WeibullA.GetValue().size() +
          "."));
      }
      if(mp_fSnag3WeibullA.GetValue().size() < iCount) {
        throw(new ModelException(ErrorGUI.BAD_DATA, "JAVA",
             m_sName + ":  There is the wrong number of values in " +
          mp_fSnag3WeibullA.GetDescriptor() +
          ".  Expected: " + iCount + "; actual: " + mp_fSnag3WeibullA.GetValue().size() +
          "."));
      }
      if(mp_fSnag1WeibullB.GetValue().size() < iCount) {
        throw(new ModelException(ErrorGUI.BAD_DATA, "JAVA",
             m_sName + ":  There is the wrong number of values in " +
          mp_fSnag1WeibullB.GetDescriptor() +
          ".  Expected: " + iCount + "; actual: " + mp_fSnag1WeibullB.GetValue().size() +
          "."));
      }
      if(mp_fSnag2WeibullB.GetValue().size() < iCount) {
        throw(new ModelException(ErrorGUI.BAD_DATA, "JAVA",
             m_sName + ":  There is the wrong number of values in " +
          mp_fSnag2WeibullB.GetDescriptor() +
          ".  Expected: " + iCount + "; actual: " + mp_fSnag2WeibullB.GetValue().size() +
          "."));
      }
      if(mp_fSnag3WeibullB.GetValue().size() < iCount) {
        throw(new ModelException(ErrorGUI.BAD_DATA, "JAVA",
             m_sName + ":  There is the wrong number of values in " +
          mp_fSnag3WeibullB.GetDescriptor() +
          ".  Expected: " + iCount + "; actual: " + mp_fSnag3WeibullB.GetValue().size() +
          "."));
      }

      //OK, if a is negative, b has to be a whole number
      Number oA, oB;
      for (i = 0; i < p_bAppliesTo.length; i++) {
        if (p_bAppliesTo[i]) {
          oA = (Number) mp_fSnag1WeibullA.GetValue().elementAt(i);
          if (oA.floatValue() < 0) {
            oB = (Number) mp_fSnag1WeibullB.GetValue().elementAt(i);
            if (oB.floatValue() <= 0 || oB.floatValue() - oB.intValue() > 0) {
              throw(new ModelException(ErrorGUI.BAD_DATA, "JAVA",
              "Values for " + mp_fSnag1WeibullA.GetDescriptor() + "\n must be greater than zero if that species's value for \n"
              + mp_fSnag1WeibullB.GetDescriptor() + "\n is not a whole number."));
            }
          }
          oA = (Number) mp_fSnag2WeibullA.GetValue().elementAt(i);
          if (oA.floatValue() < 0) {
            oB = (Number) mp_fSnag2WeibullB.GetValue().elementAt(i);
            if (oB.floatValue() <= 0 || oB.floatValue() - oB.intValue() > 0) {
              throw(new ModelException(ErrorGUI.BAD_DATA, "JAVA",
              "Values for " + mp_fSnag2WeibullA.GetDescriptor() + "\n must be greater than zero if that species's value for \n"
              + mp_fSnag2WeibullB.GetDescriptor() + "\n is not a whole number."));
            }
          }
          oA = (Number) mp_fSnag3WeibullA.GetValue().elementAt(i);
          if (oA.floatValue() < 0) {
            oB = (Number) mp_fSnag3WeibullB.GetValue().elementAt(i);
            if (oB.floatValue() <= 0 || oB.floatValue() - oB.intValue() > 0) {
              throw(new ModelException(ErrorGUI.BAD_DATA, "JAVA",
              "Values for " + mp_fSnag3WeibullA.GetDescriptor() + "\n must be greater than zero if that species's value for \n"
              + mp_fSnag3WeibullB.GetDescriptor() + "\n is not a whole number."));
            }
          }
        }
      }

    }

    //NCI Mortality
    oBeh = GetBehaviorByKey("NCI Mortality");
    if (oBeh.m_bIsEnabled) {

      boolean[] p_bAppliesTo = oBeh.GetWhichSpeciesUsed(oPop);
      MakeSureAllPositive(mp_fNCIMaxCrowdingRadius, p_bAppliesTo);
      MakeSureAllAreProportions(mp_fNCIMaxProbSurvival, p_bAppliesTo);
      MakeSureAllNonZero(mp_fNCISizeEffectMode, p_bAppliesTo);
      MakeSureAllNonZero(mp_fNCISizeEffectVariance, p_bAppliesTo);
      MakeSureAllAreProportions(mp_fNCINeighStormEffFull, p_bAppliesTo);
      MakeSureAllAreProportions(mp_fNCINeighStormEffMed, p_bAppliesTo);
      MakeSureAllAreProportions(mp_fNCIStormEffectFull, p_bAppliesTo);
      MakeSureAllAreProportions(mp_fNCIStormEffectMed, p_bAppliesTo);
      MakeSureAllNonNegative(mp_fNCIMinNeighborDBH); //this is for all species

      //Count the number of lambdas
      int iNumLambdas = 0;
      for (i = 0; i < oBeh.GetNumberOfRequiredDataObjects(); i++) {
        ModelData oData = (ModelData) oBeh.GetRequiredData(i);
        if (oData.GetDescriptor().toLowerCase().indexOf("nci lambda") > -1) {
          ModelVector oVectorTest = (ModelVector) oData;
          MakeSureRightSize(oVectorTest, iNumSpecies);
          MakeSureAllAreProportions(oVectorTest, p_bAppliesTo);
          iNumLambdas++;
        }
      }

      if (iNumSpecies > iNumLambdas) {
        ModelException oErr = new ModelException();
        oErr.iErrorCode = ErrorGUI.BAD_DATA;
        oErr.sCallingFunction = "JAVA";
        oErr.sMessage = "There is the wrong number of NCI lambda neighbors.";
      }
    }

    //Competition mortality
    oBeh = GetBehaviorByKey("competition mortality");
    if (oBeh.m_bIsEnabled) {
      if (m_oManager.GetPlot().GetNumberOfYearsPerTimestep() != 1) {
        ModelException oErr = new ModelException();
        oErr.iErrorCode = ErrorGUI.BAD_DATA;
        oErr.sMessage =
            oBeh.GetDescriptor() +
            " cannot be used unless the timestep length is one year.";
        oErr.sCallingFunction = "JAVA";
        throw (oErr);
      }
      //Make sure an NCI growth behavior is enabled
      GrowthBehaviors oGrowth = m_oManager.GetGrowthBehaviors();
      if (oGrowth.GetBehaviorByKey("nci").m_bIsEnabled == false &&
          oGrowth.GetBehaviorByKey("nci diam only").m_bIsEnabled == false) {
        throw (new ModelException(ErrorGUI.CANT_FIND_OBJECT, "JAVA",
                                  oBeh.GetDescriptor() +
                                  " must be used with an NCI growth behavior."));
      }
    }

    //Density self-thinning - LEM 03-17-05
    oBeh = GetBehaviorByKey("density self-thinning mortality");
    boolean[] p_bAppliesTo = oBeh.GetWhichSpeciesUsed(oPop);

    //Make sure the timestep length is 1 year
    if (oBeh.m_bIsEnabled) {
      if (m_oManager.GetPlot().GetNumberOfYearsPerTimestep() != 1) {
        ModelException oErr = new ModelException();
        oErr.iErrorCode = ErrorGUI.BAD_DATA;
        oErr.sMessage =
            oBeh.GetDescriptor() +
            " cannot be used unless the timestep length is one year.";
        oErr.sCallingFunction = "JAVA";
        throw (oErr);
      }

      MakeSureAllPositive(mp_fDensSelfThinNeighRadius, p_bAppliesTo);
      MakeSureAllNonNegative(mp_fDensSelfThinMinDensity, p_bAppliesTo);
      MakeSureAllPositive(mp_fDensSelfThinAsymptote, p_bAppliesTo);
      MakeSureAllPositive(mp_fDensSelfThinDensityEffect, p_bAppliesTo);
    }

    //LEM 05-06-05
    if (GetBehaviorByKey("logistic bi-level mortality").m_bIsEnabled) {
      p_bAppliesTo = GetBehaviorByKey("logistic bi-level mortality").
          GetWhichSpeciesUsed(oPop);
      MakeSureAllAreBounded(mp_fLogBiLevHiLiteThreshold, p_bAppliesTo, 0, 100);
    }

    //LEM 10-31-05
    if (GetBehaviorByKey("stochastic bi-level mortality").m_bIsEnabled) {
      p_bAppliesTo = GetBehaviorByKey("stochastic bi-level mortality").
          GetWhichSpeciesUsed(oPop);
      MakeSureAllAreBounded(mp_fStochBiLevHiLiteThreshold, p_bAppliesTo, 0, 100);
      MakeSureAllAreProportions(mp_fStochBiLevHiLiteMortProb, p_bAppliesTo);
      MakeSureAllAreProportions(mp_fStochBiLevLoLiteMortProb, p_bAppliesTo);
      if (m_oManager.GetLightBehaviors().GetBehaviorByKey("storm light").m_bIsEnabled == false) {
        throw(new ModelException(ErrorGUI.CANT_FIND_OBJECT, "JAVA", "If the \"" +
          GetBehaviorByKey("stochastic bi-level mortality").GetDescriptor() +
          "\" behavior is enabled, you must also enable the \"" +
          m_oManager.GetLightBehaviors().GetBehaviorByKey("storm light").GetDescriptor() +
          "\" behavior."));
      }
    }

    //LEM 01-26-06
    if (GetBehaviorByKey("height-gli weibull mortality").m_bIsEnabled) {
      p_bAppliesTo = GetBehaviorByKey("height-gli weibull mortality").
          GetWhichSpeciesUsed(oPop);
      MakeSureAllAreProportions(mp_fHeightGLIWeibMaxMort, p_bAppliesTo);
    }
  }

  /**
   * Finds a behavior by its XML tag.  For backwards compatibility, we'll
   * watch out for the old name of the NCI Mortality behavior for a while.
   * @param sXMLTag String The XML tag for which to find a behavior.
   * @return Behavior Behavior for the XML tag, or NULL if none of the
   * behaviors has that tag.
   */
  public Behavior GetBehaviorByXMLTag(String sXMLTag) {
    if (sXMLTag.equals("puerto rico nci mortality")) {
      return GetBehaviorByKey("NCI Mortality");
    }
    return super.GetBehaviorByXMLTag(sXMLTag);
  }

  /**
   * Updates the lambda neighbor names when a species name is changed.
   * @param sOldSpecies String Old name of the species, with underscores
   * instead of spaces (like the species names would come from the tree
   * population)
   * @param sNewSpecies String New name of the species, with underscores
   * instead of spaces (like the species names would come from the tree
   * population)

   */
  public void ChangeOfSpeciesName (String sOldSpecies, String sNewSpecies) {
    int i;

    sOldSpecies = sOldSpecies.replace('_', ' ');

    //Find the lambdas in our list
    for (i = 0; i < mp_oAllData.size(); i++) {
      ModelData oData = (ModelData) mp_oAllData.elementAt(i);
      String sKey = oData.GetDescriptor();
      if (sKey.indexOf(sOldSpecies) > -1) {
        ModelVector oVector = (ModelVector) oData;
        oVector.m_sDescriptor = sNewSpecies.replace('_', ' ') + " NCI Lambda Neighbors";
        oVector.m_sXMLTag = "mo_prnci" + sNewSpecies + "NeighborLambda";
      }
    }
  }
}
